//
//  Mapmaker.java
// Started on June 16th, 2001
// Copyright (c) Virtu Halttunen
//
// July 2nd, 2001: 	Added drag-painting
// March 29th, 2002: 	New Map format
// July 2nd, 2002: 	GUI tweaks: accelerators, some checks, normal saving
// July 18th, 2005:	Room graph management, other GUI tweaks
// July 20th, 2005:	Map object management
// July 21st, 2005:	New XML-based map format
// August 18th, 2005:	Can now add new object sets from XML files
// September 14th, 2005: Finally can add floor layers and walls by dragging. About time... Lazy developer

//import TheGame.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;


public class MapMaker extends JFrame implements MouseListener, ActionListener, MouseMotionListener, KeyListener
{
    
    public static final String MISSION_FILE_PATH_KEY = "MissionFilePath";
    public static final String BASE_RESOURCE_PATH_KEY = "BaseResourcePath";
        
    public static int SELECTED = 1;
    
    public static int EMPTY = -1;
    public static int FLOOR = 0;
    public static int WALL = 1;
    public static int DOOR = 2;
    public static int OBJECT = 3;
    
    public static final int SELECT = 0;
    public static final int PAINT_MAP = 1;
    public static final int MANAGE_ROOMS = 2;
    public static final int MANAGE_OBJECTS = 3;
    
    private static final String PAINT_DETAILS_KEY = "Paint details";
    private static final String ROOM_DETAILS_KEY = "Room details";
    private static final String OBJECT_DETAILS_KEY = "Object details";

    private JMenuBar menubar;
    private JMenuItem newMap;
    private JMenuItem save;
    private JMenuItem saveAs;
    private JMenuItem open;
    private JMenuItem openOld;
    private Map map;
    private MapPanel mapPanel;
    private OverviewMapPanel overview;
    private CellSelectionPanel floorCellSelector;
    private CellSelectionPanel wallCellSelector;
    private DoorSelectionPanel doorCellSelector;
    private CellSelectionPanel objectCellSelector;
    private JTabbedPane selectors;
    private JTextField nameField;
    private JButton selectionButton;
    private JButton mapButton;
    private JButton roomButton;
    private JButton objectsButton;
    
    private JButton addCellsButton;
    
    private JPanel detailViewsPanel;		// For displaying the details for different modes
    private CardLayout detailViewSwitcher;
    // Paint mode details
    private Box paintDetailsPanel;
    private JCheckBox showLightsBox;
    private JCheckBox applyLightBox;
    private JCheckBox applyCellBox;
    private JComboBox lightLevels;
    private ImagePanel leftIcon;
    private ImagePanel rightIcon;
    // Room mode details
    private Box roomDetailsPanel;
    private JTextField roomNameField;
    private JTextField roomStartXField;
    private JTextField roomStartYField;
    private JTextField roomEndXField;
    private JTextField roomEndYField;
    private JTextField roomHotSpotXField;
    private JTextField roomHotSpotYField;
    private JButton addRoomButton;
    private JButton removeRoomButton;
    // Map object details
    private Box objectDetailsPanel;
    private JTextField objectNameField;
    private JTextField objectStrengthField;
    private JTextField objectCostToMoveField;
    private JButton addObjectButton;
    private JButton removeObjectButton;
    
    
    private int cellSize = 48;
    private int leftSelectedType = 0;
    private int rightSelectedType = 0;
    private int leftSelectedValue = 0;
    private int rightSelectedValue = 0;
    private boolean painting = false;
    private Point paintFrom = null;
    private Point paintTo = null;
    private boolean edited = false;
    private Room currentRoom = null;
    private Vector mapObjectCellData = null;
    private MapObjectCell [] selectedObjectCells = null;
    private MapObject currentObject = null;
    private int mode = PAINT_MAP;
    
    private File mapFile = null;
    private Properties preferences;
    private String baseResourcePath;
    private String missionDirectoryPath;

    public MapMaker() 
    {
	super("MapMaker");
	if( !loadPreferences() )
	{
	    JOptionPane.showMessageDialog( this, "Please run The Game before using MapEditor" );
	    System.exit( 0 );
	}
	map = new Map( 50, 50 );
	map.addFloorSet( "ftiles.jpg" );
	map.addFloorSet( "doorbase.png" );
	map.addWallSet( "walltiles.png" );
	map.addDoorSet( "doors.png" );
	mapObjectCellData = new Vector();	
	//map.addObjectSet( "objects.png" );
	try
	{
	    loadMapObjectCells( "DefaultObjects.xml" );
	}
	catch( Exception e )
	{
	    System.out.println( "Error loading default objects!" );
	    e.printStackTrace();
	    System.exit( 1 );
	}	
	
	mapPanel = new MapPanel( map, 48, 15, 10 );
	mapPanel.setAutoscrolls( true );
	mapPanel.addMouseListener( this );
	mapPanel.addMouseMotionListener( this );
	overview = new OverviewMapPanel( map );
	overview.addMouseListener( this );
	
	Box bottom = new Box( BoxLayout.X_AXIS );
	Box bottomLeft = Box.createVerticalBox();
	bottom.add( bottomLeft );
	
	Box buttonsBox = Box.createHorizontalBox();
	ButtonGroup actionButtons = new ButtonGroup();
	ImageIcon selectIcon = new ImageIcon( getClass().getResource( "ArrowIcon.png" ) );
	selectionButton = new JButton( selectIcon );
	selectionButton.setToolTipText( "Select" );
	selectionButton.setModel( new JToggleButton.ToggleButtonModel() );
	selectionButton.addActionListener( this );
	actionButtons.add( selectionButton );
	buttonsBox.add( selectionButton );
	ImageIcon mapIcon = new ImageIcon( getClass().getResource( "PaintMapIcon.png" ) );
	mapButton = new JButton( mapIcon );
	mapButton.setToolTipText( "Paint map" );
	mapButton.setModel( new JToggleButton.ToggleButtonModel() );
	mapButton.addActionListener( this );
	actionButtons.add( mapButton );
	buttonsBox.add( mapButton );
	ImageIcon roomIcon = new ImageIcon( getClass().getResource( "RoomsIcon.png" ) );
	roomButton = new JButton( roomIcon );
	roomButton.setToolTipText( "Manage rooms" );
	roomButton.setModel( new JToggleButton.ToggleButtonModel() );
	roomButton.addActionListener( this );
	actionButtons.add( roomButton );
	buttonsBox.add( roomButton );
	ImageIcon objectsIcon = new ImageIcon( getClass().getResource( "ObjectsIcon.png" ) );
	objectsButton = new JButton( objectsIcon );
	objectsButton.setToolTipText( "Manage objects" );
	objectsButton.setModel( new JToggleButton.ToggleButtonModel() );
	objectsButton.addActionListener( this );
	actionButtons.add( objectsButton );
	buttonsBox.add( objectsButton );
	
	mapButton.setSelected( true );
	buttonsBox.add( Box.createHorizontalGlue() );
	bottomLeft.add( buttonsBox );
	bottomLeft.add( Box.createVerticalGlue() );
	
	Box nameBox = Box.createHorizontalBox();
	nameBox.add( new JLabel( "Map name" ) );
	nameField = new JTextField( "Test map", 30 );
	nameField.setMinimumSize( nameField.getPreferredSize() );
	nameField.setMaximumSize( nameField.getPreferredSize() );
	nameBox.add( nameField );
	bottomLeft.add( nameBox );
	bottomLeft.add( Box.createVerticalStrut( 16 ) );
	
	// Detail panels
	detailViewsPanel = new JPanel();
	detailViewSwitcher = new CardLayout();
	detailViewsPanel.setLayout( detailViewSwitcher );
	
	/// Cell details
	paintDetailsPanel = Box.createVerticalBox();
	Box lights = new Box( BoxLayout.X_AXIS );
	lights.add( Box.createHorizontalGlue() );
	lights.add( new JLabel( "Darkness" ) );
	lightLevels = new JComboBox();
	lightLevels.setAlignmentX( 0.5f );
	for( int i=0; i < 32; i++ )
	{
	    lightLevels.addItem( String.valueOf(i) );
	}
	lights.add( lightLevels );
	lights.add( Box.createHorizontalGlue() );
	paintDetailsPanel.add( lights );
	showLightsBox = new JCheckBox( "Show lights", true );
	showLightsBox.addActionListener( this );
	paintDetailsPanel.add( showLightsBox );
	Box icons = new Box( BoxLayout.X_AXIS );
	icons.add( Box.createHorizontalGlue() );
	leftIcon = new ImagePanel( (Image)(mapPanel.getFloorCells().elementAt(0)), 0, Color.lightGray );
	leftIcon.setSize( cellSize, cellSize );
	icons.add( leftIcon );
	icons.add( Box.createHorizontalStrut( 10 ) );
	rightIcon = new ImagePanel( (Image)(mapPanel.getWallCells().elementAt(0)), 0, Color.lightGray );
	rightIcon.setSize( cellSize, cellSize );
	icons.add( rightIcon );
	icons.add( Box.createHorizontalGlue() );

	paintDetailsPanel.add( icons );
	
	Box cellDetailsBox = new Box( BoxLayout.Y_AXIS );
	applyLightBox = new JCheckBox( "Apply light", true );
	cellDetailsBox.add( applyLightBox );
	applyCellBox = new JCheckBox( "Apply cell", true );
	cellDetailsBox.add( applyCellBox );
	paintDetailsPanel.add( cellDetailsBox );
	
	detailViewsPanel.add( PAINT_DETAILS_KEY, paintDetailsPanel );
	
	/// Room details
	roomDetailsPanel = Box.createVerticalBox();
	Box roomNamePanel = Box.createHorizontalBox();
	roomNamePanel.add( new JLabel( "Room name" ) );
	roomNamePanel.add( Box.createHorizontalStrut( 10 ) );
	roomNamePanel.add( Box.createHorizontalGlue() );
	roomNameField = new JTextField();
	roomNamePanel.add( roomNameField );
	roomDetailsPanel.add( roomNamePanel );
	Box roomStartPanel = Box.createHorizontalBox();
	roomStartPanel.add( new JLabel( "Top left" ) );
	roomStartPanel.add( Box.createHorizontalGlue() );
	roomStartXField = new JTextField( "-1" , 4 );
	roomStartXField.setHorizontalAlignment( SwingConstants.RIGHT );
	roomStartPanel.add( roomStartXField );
	roomStartPanel.add( Box.createHorizontalStrut( 10 ) );
	roomStartYField  = new JTextField( "-1" , 4 );
	roomStartYField.setHorizontalAlignment( SwingConstants.RIGHT );
	roomStartPanel.add( roomStartYField );
	roomDetailsPanel.add( roomStartPanel );
	Box roomEndPanel = Box.createHorizontalBox();
	roomEndPanel.add( new JLabel( "Bottom right" ) );
	roomEndPanel.add( Box.createHorizontalGlue() );
	roomEndXField = new JTextField( "-1" , 4 );
	roomEndXField.setHorizontalAlignment( SwingConstants.RIGHT );
	roomEndPanel.add( roomEndXField );
	roomEndPanel.add( Box.createHorizontalStrut( 10 ) );
	roomEndYField  = new JTextField( "-1" , 4 );
	roomEndYField.setHorizontalAlignment( SwingConstants.RIGHT );
	roomEndPanel.add( roomEndYField );
	roomDetailsPanel.add( roomEndPanel );
	Box roomHotSpotPanel = Box.createHorizontalBox();
	roomHotSpotPanel.add( new JLabel( "Hotspot" ) );
	roomHotSpotPanel.add( Box.createHorizontalGlue() );
	roomHotSpotXField = new JTextField( "-1" , 4 );
	roomHotSpotXField.setHorizontalAlignment( SwingConstants.RIGHT );
	roomHotSpotPanel.add( roomHotSpotXField );
	roomHotSpotPanel.add( Box.createHorizontalStrut( 10 ) );
	roomHotSpotYField  = new JTextField( "-1" , 4 );
	roomHotSpotYField.setHorizontalAlignment( SwingConstants.RIGHT );
	roomHotSpotPanel.add( roomHotSpotYField );
	roomDetailsPanel.add( roomHotSpotPanel );
	Box roomButtonsPanel = Box.createHorizontalBox();
	ImageIcon plusIcon = new ImageIcon( ResourceManager.loadImage( "PlusSign.png" ) );
	addRoomButton = new JButton( plusIcon );
	addRoomButton.addActionListener( this );
	addRoomButton.setToolTipText( "Add room" );
	roomButtonsPanel.add( addRoomButton );
	ImageIcon minusIcon = new ImageIcon( ResourceManager.loadImage( "MinusSign.png" ) );
	removeRoomButton = new JButton( minusIcon );
	removeRoomButton.addActionListener( this );
	removeRoomButton.setEnabled( false );
	roomButtonsPanel.add( removeRoomButton );
	roomButtonsPanel.add( Box.createHorizontalGlue() );
	roomDetailsPanel.add( roomButtonsPanel );
	
	roomDetailsPanel.add( Box.createVerticalGlue() );
	detailViewsPanel.add( ROOM_DETAILS_KEY, roomDetailsPanel );
	
	/// Object details
	objectDetailsPanel = Box.createVerticalBox();
	Box objectNamePanel = Box.createHorizontalBox();
	objectNamePanel.add( new JLabel( "Object name" ) );
	objectNamePanel.add( Box.createHorizontalStrut( 10 ) );
	objectNamePanel.add( Box.createHorizontalGlue() );
	objectNameField = new JTextField();
	objectNamePanel.add( objectNameField );
	objectDetailsPanel.add( objectNamePanel );
	Box objectStrengthPanel = Box.createHorizontalBox();
	objectStrengthPanel.add( new JLabel( "Strength" ) );
	objectStrengthPanel.add( Box.createHorizontalGlue() );
	objectStrengthField = new JTextField( "1", 5 );
	objectStrengthField.setHorizontalAlignment( SwingConstants.RIGHT );
	objectStrengthPanel.add( objectStrengthField );
	objectDetailsPanel.add( objectStrengthPanel );
	Box objectCostToMovePanel = Box.createHorizontalBox();
	objectCostToMovePanel.add( new JLabel( "Cost to move" ) );
	objectCostToMovePanel.add( Box.createHorizontalGlue() );
	objectCostToMoveField = new JTextField( "-1", 5 );
	objectCostToMoveField.setHorizontalAlignment( SwingConstants.RIGHT );
	objectCostToMovePanel.add( objectCostToMoveField );
	objectDetailsPanel.add( objectCostToMovePanel );
	Box objectButtonsPanel = Box.createHorizontalBox();
	addObjectButton = new JButton( plusIcon );
	addObjectButton.addActionListener( this );
	addObjectButton.setToolTipText( "Add object" );
	addObjectButton.setEnabled( false );
	objectButtonsPanel.add( addObjectButton );
	removeObjectButton = new JButton( minusIcon );
	removeObjectButton.addActionListener( this );
	removeObjectButton.setToolTipText( "Remove object" );
	removeObjectButton.setEnabled( false );
	objectButtonsPanel.add( removeObjectButton );
	objectButtonsPanel.add( Box.createHorizontalGlue() );
	objectDetailsPanel.add( objectButtonsPanel );
	detailViewsPanel.add( OBJECT_DETAILS_KEY, objectDetailsPanel );
	 	
	// Details end
	bottom.add( Box.createHorizontalStrut( 10 ) );
	bottom.add( Box.createHorizontalGlue() );
	bottom.add( detailViewsPanel );
	
	Box left = new Box( BoxLayout.Y_AXIS );
	JScrollPane mapScroller = new JScrollPane( mapPanel );
	left.add( mapScroller );
	left.add( bottom );
	
	Box right = new Box( BoxLayout.Y_AXIS );
	right.add( overview );
	
	selectors = new JTabbedPane(); // Needs to be final for the inner class, see below
	floorCellSelector = new CellSelectionPanel( mapPanel.getFloorCells(), 48, 4, 6, 4 );
	floorCellSelector.addMouseListener( this );
	selectors.add( "Floors", new JScrollPane( floorCellSelector ) );
	wallCellSelector = new CellSelectionPanel( mapPanel.getWallCells(), 48, 4, 6, 4 );
	wallCellSelector.addMouseListener( this );
	selectors.add( "Walls", new JScrollPane( wallCellSelector ) );
	doorCellSelector = new DoorSelectionPanel( mapPanel.getDoorCells(), 48, 4, 6, 4 );
	doorCellSelector.addMouseListener( this );
	selectors.add( "Doors", new JScrollPane( doorCellSelector ) );
	objectCellSelector = new CellSelectionPanel( mapPanel.getObjectCells(), 48, 4, 6, 4 );
	objectCellSelector.addMouseListener( this );
	selectors.add( "Objects", new JScrollPane( objectCellSelector ) );
	right.add( selectors );
	
	Box addButtonBox = Box.createHorizontalBox();
	addCellsButton = new JButton( plusIcon );
	addCellsButton.addActionListener( this );
	addButtonBox.add( addCellsButton );
	addButtonBox.add( Box.createHorizontalGlue() );
	right.add( addButtonBox );
	
	Box mainBox = new Box( BoxLayout.X_AXIS );
	mainBox.add( left );
	mainBox.add( right );
	getContentPane().add( mainBox );
	
	int shortCutMask = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
	menubar = new JMenuBar();
	JMenu file = new JMenu( "File" );
        newMap = new JMenuItem( "New... " );
        newMap.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_N, shortCutMask ) );
        newMap.addActionListener( this );
        file.add( newMap );
        open = new JMenuItem( "Open... " );
        open.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_O, shortCutMask ) );
	open.addActionListener( this );
	file.add( open );
        openOld = new JMenuItem( "Open old... " );
        openOld.addActionListener( this );
        file.add( openOld );
        save = new JMenuItem( "Save" );
        save.addActionListener( this );
        save.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_S, shortCutMask ) );
        file.add( save );
        saveAs = new JMenuItem( "Save as... " );
	saveAs.addActionListener( this );
	file.add( saveAs );
        menubar.add( file );
	setJMenuBar( menubar );
	
	mapPanel.addKeyListener( this );
	addKeyListener( this );
	
	pack();
	
	// The "proper" and thread-safe way to stop the various fields from growing too much. Ugly as sin
	SwingUtilities.invokeLater( new Runnable() {
	    public void run() {
		
		// Stop the selectors panel from growing
		Dimension selectorsSize = new Dimension( selectors.getWidth(), selectors.getHeight() );
		selectors.setMinimumSize( selectorsSize );
		selectors.setMaximumSize( selectorsSize );
		Dimension fieldSize = roomStartXField.getPreferredSize();
		roomStartXField.setMaximumSize( fieldSize );
		roomStartYField.setMaximumSize( fieldSize );
		roomEndXField.setMaximumSize( fieldSize );
		roomEndYField.setMaximumSize( fieldSize );
		roomHotSpotXField.setMaximumSize( fieldSize );
		roomHotSpotYField.setMaximumSize( fieldSize );
		objectStrengthField.setMaximumSize( fieldSize );
		objectCostToMoveField.setMaximumSize( fieldSize );
		
		fieldSize.width = fieldSize.width*4;
		roomNameField.setMaximumSize( fieldSize );
		roomNameField.setPreferredSize( fieldSize );
		objectNameField.setMaximumSize( fieldSize );
		objectNameField.setPreferredSize( fieldSize );
		
		roomNameField.invalidate();
		objectNameField.invalidate();
		validate();
		
		setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
		setVisible(true);
	    }
	} );
    }

    public static int convertDoorToDirection( int doorIndex )
    {
        if( doorIndex == 0 ) return Map.SOUTH; 
        if( doorIndex == 1 ) return Map.NORTH; 
        if( doorIndex == 2 ) return Map.EAST; 
        if( doorIndex == 3 ) return Map.WEST;
        return 0;
    }
    
    public void displayObjectInfo( MapObject mapObject )
    {
	if( mapObject == null )
	{
	    addObjectButton.setEnabled( false );
	    removeObjectButton.setEnabled( false );
	    objectStrengthField.setText( "-1" );
	    objectCostToMoveField.setText( "-1" );
	}
	else
	{
	    addObjectButton.setEnabled( true );
	    addObjectButton.setToolTipText( "Edit object" );
	    removeObjectButton.setEnabled( true );
	    objectNameField.setText( mapObject.getName() );
	    objectStrengthField.setText( String.valueOf( mapObject.getStrength() ) );
	    objectCostToMoveField.setText( String.valueOf( mapObject.getCostToMove() ) );
	}
    }
    
    public void displayRoomInfo( Room room )
    {
	if( room == null )
	{
	    roomStartXField.setText( "-1" );
	    roomStartYField.setText( "-1" );
	    roomEndXField.setText( "-1" );
	    roomEndYField.setText( "-1" );	
	    roomHotSpotXField.setText( "-1" );
	    roomHotSpotYField.setText( "-1" );
	    roomNameField.setText( "" );
	    removeRoomButton.setEnabled( false );
	    addRoomButton.setToolTipText( "Add room" );
	}
	else
	{
	    roomStartXField.setText( String.valueOf( room.getTopLeft().x ) );
	    roomStartYField.setText( String.valueOf( room.getTopLeft().y ) );
	    roomEndXField.setText( String.valueOf( room.getBottomRight().x ) );
	    roomEndYField.setText( String.valueOf( room.getBottomRight().y ) );	
	    roomHotSpotXField.setText( String.valueOf( room.getHotSpot().x ) );
	    roomHotSpotYField.setText( String.valueOf( room.getHotSpot().y ) );
	    roomNameField.setText( room.getName() );
	    removeRoomButton.setEnabled( true );
	    addRoomButton.setToolTipText( "Edit room" );
	}
    }
    
    public void keyPressed( KeyEvent e )
    {
    	System.out.println( "Key pressed" );
	int key = e.getKeyCode();
	if( ( key == KeyEvent.VK_BACK_SPACE || key == KeyEvent.VK_DELETE ) && ( mode == PAINT_MAP || mode == SELECT ) )
	{
	    System.out.println( "Deleting selection" );
	    map.deleteSelected();
	    mapPanel.repaint();
	    overview.repaint();
	}
	if( ( mode == SELECT || mode == PAINT_MAP ) && ( key == KeyEvent.VK_UP || key == KeyEvent.VK_DOWN || key == KeyEvent.VK_LEFT || key == KeyEvent.VK_RIGHT ) )
	{
	    int direction = Map.NORTH;
	    switch( key )
	    {
		case KeyEvent.VK_DOWN:
		    direction = Map.SOUTH;
		    break;
		case KeyEvent.VK_LEFT:
		    direction = Map.WEST;
		    break;
		case KeyEvent.VK_RIGHT:
		    direction = Map.EAST;
		    break;
	    }
	    int shortCutMask = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
	    if( ( e.getModifiers() & shortCutMask ) == shortCutMask )
	    {
		map.nudgeSelection( direction );
		mapPanel.repaint();
		overview.repaint();
	    }
	}
	if( e.getKeyChar() == '+' && e.isMetaDown() )
	{
	    System.out.println( "Enlarge" );
	    int newSize = mapPanel.getCellSize()+8;
	    if( newSize > 48 )
	    {
		newSize = 48;
	    }
	    mapPanel.setCellSize( newSize );
	    validate();
	}
	if( e.getKeyChar() == '-' && e.isMetaDown() )
	{
	    System.out.println( "Shrink" );
	    int newSize = mapPanel.getCellSize()-8;
	    if( newSize < 8 )
	    {
		newSize = 8;
	    }
	    mapPanel.setCellSize( newSize );
	    validate();
	}
    }
    
    public void keyReleased( KeyEvent e )
    {
	System.out.println( "Key released" );
    }
    
    public void keyTyped( KeyEvent e )
    {
    }
    
    private void loadMapObjectCells( String fileName ) throws Exception
    {
	java.net.URL resourceURL = ResourceManager.findResource( fileName );
	if( resourceURL != null )
	{
	    DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
	    DOMImplementation builder = documentBuilder.getDOMImplementation();
	    
	    InputStream is = resourceURL.openStream();
	    
	    System.out.println( "Parsing objects" );
	    Document document = documentBuilder.parse( is );
	    is.close();
	    
	    Element root = document.getDocumentElement();
	    if( root.getTagName().equals( "mapmaker" ) )
	    {
		Element objectCellsElement = XMLUtils.getNamedChild( root, "mapobjectcells" );
		String imageFileName = objectCellsElement.getAttribute( "image" );
		int cellWidth = Integer.parseInt( objectCellsElement.getAttribute( "cellwidth" ) );
		System.out.println( "Image file: " + imageFileName );
		map.addObjectSet( imageFileName );
		Element [] cells = XMLUtils.getNamedChildren( objectCellsElement, "item" );
		int startIndex = mapObjectCellData.size();
		for( int i=0; i < cells.length; i++ )
		{
		    int type = i+startIndex;		    
		    String name = cells[i].getAttribute( "name" );
		    int hitPoints = Integer.parseInt( cells[i].getAttribute( "hitpoints" ) );
		    boolean blocksMovement = new Boolean( cells[i].getAttribute( "blocksmovement" ) ).booleanValue();
		    int costToMove = Integer.parseInt( cells[i].getAttribute( "costtomove" ) );
		    boolean blocksSight = new Boolean( cells[i].getAttribute( "blockssight" ) ).booleanValue();
		    int brokenIndex = startIndex+Integer.parseInt( cells[i].getAttribute( "brokenindex" ) );
		    MapObjectCell newCell = new MapObjectCell( type, name, hitPoints, blocksMovement, costToMove, blocksSight, brokenIndex );
		    mapObjectCellData.add( newCell );
		    System.out.println( "Found object: " + name );
		}
		System.out.println( "Objects: " + mapObjectCellData.size() );
	    }
	    
	}
    }
    
    private boolean loadPreferences()
    {
        String directory = System.getProperty( "user.home" );
        Properties properties = new Properties();
        File prefsFile = new File( directory, ".theGame.properties" );
        if( prefsFile.exists() )
        {
            if( prefsFile.canRead() )
            {
                try
		{
		    FileInputStream fis = new FileInputStream( prefsFile );
		    properties.load( fis );
		    fis.close();
		}
		    catch( Exception e )
		{
			properties = null;
		}
            }
        }
        else
        {
            System.out.println( "Preference file not found" );
            preferences = new Properties();
            properties = null;
            return false;
        }
        if( properties != null )
        {
            preferences = properties;
            missionDirectoryPath = preferences.getProperty( MISSION_FILE_PATH_KEY );
	    baseResourcePath = preferences.getProperty( BASE_RESOURCE_PATH_KEY );
	    if( baseResourcePath == null )
	    {
		return false;
	    }
	    ResourceManager.addBaseURL( baseResourcePath.substring( 0, baseResourcePath.length() ) );
	    System.out.println( "Mission directory: " + missionDirectoryPath );
	    ResourceManager.addBaseURL( "file:"+missionDirectoryPath+"/Resources/" );
        }
        return true;
    }
        
    public void mouseClicked( MouseEvent e )
    {
	Point cellPlace;
	
	//System.out.println( "Mouse click! " + e.toString() );
	if( e.getSource() == floorCellSelector )
	{
	    setMode( PAINT_MAP );
	    if( ( e.getModifiers() & MouseEvent.BUTTON3_MASK ) == MouseEvent.BUTTON3_MASK )
	    {
                rightSelectedType = FLOOR;
		rightSelectedValue = floorCellSelector.getCellType( e.getPoint() );
		//System.out.println( "Right selected type: " + rightSelectedValue );
		rightIcon.setImage( floorCellSelector.getCellImage( rightSelectedValue ) );
            }
	    else if( ( e.getModifiers() & MouseEvent.BUTTON1_MASK ) == MouseEvent.BUTTON1_MASK )
	    {
                leftSelectedType = FLOOR;
                leftSelectedValue = floorCellSelector.getCellType( e.getPoint() );
		leftIcon.setImage( floorCellSelector.getCellImage( leftSelectedValue ) );
	    }
	}
	if( e.getSource() == wallCellSelector )
	{
	    setMode( PAINT_MAP );
	    if( ( e.getModifiers() & MouseEvent.BUTTON3_MASK ) == MouseEvent.BUTTON3_MASK )
	    {
                rightSelectedType = WALL;
                rightSelectedValue = wallCellSelector.getCellType( e.getPoint() );
		//System.out.println( "Right selected type: " + rightSelectedType );
		rightIcon.setImage( wallCellSelector.getCellImage( rightSelectedValue ) );
	    }
	    else if( ( e.getModifiers() & MouseEvent.BUTTON1_MASK ) == MouseEvent.BUTTON1_MASK )
	    {
                leftSelectedType = WALL;
                leftSelectedValue = wallCellSelector.getCellType( e.getPoint() );
		leftIcon.setImage( wallCellSelector.getCellImage( leftSelectedValue ) );
	    }
	}
	if( e.getSource() == doorCellSelector )
	{
	    setMode( PAINT_MAP );
	    if( ( e.getModifiers() & MouseEvent.BUTTON3_MASK ) == MouseEvent.BUTTON3_MASK )
	    {
                rightSelectedType = DOOR;
		rightSelectedValue = doorCellSelector.getCellType( e.getPoint() );
		//System.out.println( "Right selected type: " + rightSelectedType );
		rightIcon.setImage( doorCellSelector.getCellImage( rightSelectedValue ) );
            }
	    else if( ( e.getModifiers() & MouseEvent.BUTTON1_MASK ) == MouseEvent.BUTTON1_MASK )
	    {
                leftSelectedType = DOOR;
		leftSelectedValue = doorCellSelector.getCellType( e.getPoint() );
		leftIcon.setImage( doorCellSelector.getCellImage( leftSelectedValue ) );
	    }
	}
	if( e.getSource() == objectCellSelector )
	{
	    setMode( PAINT_MAP );
	    if( ( e.getModifiers() & MouseEvent.BUTTON3_MASK ) == MouseEvent.BUTTON3_MASK )
	    {
                rightSelectedType = OBJECT;
                rightSelectedValue = objectCellSelector.getCellType( e.getPoint() );
		//System.out.println( "Right selected type: " + rightSelectedType );
		rightIcon.setImage( objectCellSelector.getCellImage( rightSelectedValue ) );
	    }
	    else if( ( e.getModifiers() & MouseEvent.BUTTON1_MASK ) == MouseEvent.BUTTON1_MASK )
	    {
                leftSelectedType = OBJECT;
                leftSelectedValue = objectCellSelector.getCellType( e.getPoint() );
		leftIcon.setImage( objectCellSelector.getCellImage( leftSelectedValue ) );
                System.out.println( "Left selected value: " + leftSelectedValue );
	    }
            
	}
	if( e.getSource() == mapPanel )
	{
	    cellPlace = mapPanel.getCell( e.getPoint() );
	    
	    if( SwingUtilities.isRightMouseButton( e ) )
	    {
		if( mode == PAINT_MAP )
		{
		    edited=true;
		    if( applyLightBox.isSelected() )
		    {
			map.setLight( cellPlace.x, cellPlace.y, (byte)Integer.parseInt( (String)lightLevels.getSelectedItem() ) );
		    }
		    if( applyCellBox.isSelected() )
		    {
                        if( rightSelectedType == FLOOR )
                            map.setCellFloor( cellPlace.x, cellPlace.y, rightSelectedValue );
                        if( rightSelectedType == WALL )
                            map.setCellWall( cellPlace.x, cellPlace.y, rightSelectedValue );
		    }
		}
	    }
	    else if( SwingUtilities.isLeftMouseButton( e ) )
	    {
                
		if( e.isAltDown() )
		{
                    System.out.println( "Picking from map" );
                    MapCell cell = map.getCell( cellPlace.x, cellPlace.y );
                    if( cell != null )
                    {
                        if( leftSelectedType == FLOOR )
			{
                            leftSelectedValue = cell.getFloor();
			    if( leftSelectedValue >= 0 && leftSelectedValue < floorCellSelector.getCellImages().size() )
			    {
				leftIcon.setImage( floorCellSelector.getCellImage( leftSelectedValue ) );
			    }
			    else
			    {
				leftIcon.setImage( null );
			    }
			}
                        if( leftSelectedType == WALL )
			{
                            leftSelectedValue = cell.getWall();
			    if( leftSelectedValue >= 0 && leftSelectedValue < wallCellSelector.getCellImages().size() )
			    {
				leftIcon.setImage( wallCellSelector.getCellImage( leftSelectedValue ) );
			    }
			    else
			    {
				leftIcon.setImage( null );
			    }
			}
                    }
                    else
                    {
                        System.out.println( "Erase" );
                        leftSelectedValue = -1;
			leftIcon.setImage( null );
                    }
		    //leftIcon.setImage( mapPanel.selectImageForCellType( leftSelectedValue ) );
		    //lightLevels.setSelected( cell.getLight() );
		}
		else
		{
		    if( mode == PAINT_MAP )
		    {
			edited = true;
			if( applyLightBox.isSelected() )
			{
			    map.setLight( cellPlace.x, cellPlace.y, (byte)Integer.parseInt( (String)lightLevels.getSelectedItem() ) );
			}
			if( applyCellBox.isSelected() )
			{
			    if( leftSelectedType == FLOOR )
			    {
				if( e.isShiftDown() )
				{
				    map.addCellFloor( cellPlace.x, cellPlace.y, leftSelectedValue );
				}
				else
				    map.setCellFloor( cellPlace.x, cellPlace.y, leftSelectedValue );
			    }
			    if( leftSelectedType == WALL )
				map.setCellWall( cellPlace.x, cellPlace.y, leftSelectedValue );
			    if( leftSelectedType == DOOR )
			    {
				if( leftSelectedValue >= 0 )
				{
				    map.addDoor( new Door( cellPlace.x, cellPlace.y, convertDoorToDirection( leftSelectedValue ) ) );
				    if( e.isShiftDown() )
				    {
					map.getDoor( cellPlace.x, cellPlace.y ).setOpen( true );
				    }
				}
				else
				{
				    map.removeDoor( cellPlace.x, cellPlace.y );
				}
			    }
			    if( leftSelectedType == OBJECT )
			    {
				if( leftSelectedValue >= 0 )
				{
				    int objIndex = leftSelectedValue;
				    //int hitPoints = MapObjectCell.objectHitPoints[ objIndex ];
				    //boolean objBlocksMovement = MapObjectCell.objectBlockingMovement[ objIndex ];
				    //boolean objBlocksSight = MapObjectCell.objectBlockingSight[ objIndex ];
				    MapObjectCell protoType = (MapObjectCell)mapObjectCellData.elementAt( objIndex );
				    map.addObjectCell( new MapObjectCell( cellPlace.x, cellPlace.y, protoType ) );

				    //map.addObjectCell( new MapObjectCell( cellPlace.x, cellPlace.y, objIndex, hitPoints, objBlocksMovement, objBlocksSight ) );
				}
				else
				{
				    map.removeUnboundObjectCells( cellPlace.x, cellPlace.y );
				}
			    }
			}
		    }
		}
	    }
            else if( SwingUtilities.isMiddleMouseButton( e ) )
            {
                System.out.println( "Picking from map" );
                MapCell cell = map.getCell( cellPlace.x, cellPlace.y );
                if( cell != null )
                {
                    if( leftSelectedType == FLOOR )
                        leftSelectedValue = cell.getFloor();
                    if( leftSelectedType == WALL )
                        leftSelectedValue = cell.getWall();
                }
                else
                {
                    System.out.println( "Erase" );
                    leftSelectedValue = -1;
                }
            }
	    if( mode != SELECT )
	    {
		map.clearSelection();
	    }
	    if( mode == MANAGE_ROOMS )
	    {
		currentRoom = map.getRoomForPoint( cellPlace );
		if( currentRoom != null && e.getClickCount() >= 2 )
		{
		    currentRoom.setHotSpot( cellPlace );
		}
		displayRoomInfo( currentRoom );
	    }
	    if( mode == MANAGE_OBJECTS )
	    {
		currentObject = map.getObjectForPoint( cellPlace );
		displayObjectInfo( currentObject );
	    }
	    cellSize = mapPanel.getCellSize();
	    mapPanel.repaint( 10, cellPlace.x*cellSize, cellPlace.y*cellSize, cellSize, cellSize );
	    /*RepaintManager.currentManager( mapPanel ).addDirtyRegion( mapPanel, cell.x*cellSize, cell.y, cellSize*2, cellSize*2 );
	    RepaintManager.currentManager( mapPanel ).paintDirtyRegions();*/
	    overview.repaint();
	}
	if( e.getSource() == overview )
	{
	    cellPlace = overview.getCell( e.getPoint() );
	    mapPanel.scrollRectToVisible( new Rectangle( (cellPlace.x-6)*cellSize, (cellPlace.y-6)*cellSize, 13*cellSize, 13*cellSize ) );	    
	}
    }

    public void mouseDragged( MouseEvent e )
    {
	if( e.getSource() == mapPanel )
	{
	    System.out.println( "Dragging..." );
	    Point scrollPoint = e.getPoint();
	    scrollPoint.x = (scrollPoint.x/cellSize)*cellSize;
	    scrollPoint.y = (scrollPoint.y/cellSize)*cellSize;

	    Rectangle scrollRect = new Rectangle( scrollPoint.x, scrollPoint.y, cellSize*2, cellSize*2 );
	    mapPanel.scrollRectToVisible( scrollRect );
	    
	    painting = true;
	    if( paintFrom == null )
	    {
		paintFrom = new Point( mapPanel.getCell( e.getPoint() ) );
	    }
	    if( paintTo == null )
	    {
		paintTo = new Point( mapPanel.getCell( e.getPoint() ) );
	    }
	    paintFrom.x = Math.min( paintFrom.x, mapPanel.getCell( e.getPoint() ).x );
	    paintFrom.y = Math.min( paintFrom.y, mapPanel.getCell( e.getPoint() ).y );
	    paintTo.x = Math.max( paintTo.x, mapPanel.getCell( e.getPoint() ).x );
	    paintTo.y = Math.max( paintTo.y, mapPanel.getCell( e.getPoint() ).y );
	    
	    if( paintTo.x >= map.getHorizontalSize() )
	    {
		paintTo.x = map.getHorizontalSize()-1;
	    }
	    if( paintTo.y >= map.getVerticalSize() )
	    {
		paintTo.y = map.getVerticalSize()-1;
	    }
	    for( int i = paintFrom.x; i<=paintTo.x; i++ )
	    {
		for( int j = paintFrom.y; j<=paintTo.y; j++ )
		{
		    map.setUtility( i, j, SELECTED );
		}
	    }
	    if( mode == MANAGE_ROOMS )
	    {
		roomStartXField.setText( String.valueOf( paintFrom.x ) );
		roomStartYField.setText( String.valueOf( paintFrom.y ) );
		roomEndXField.setText( String.valueOf( paintTo.x ) );
		roomEndYField.setText( String.valueOf( paintTo.y ) );
	    }
	    if( mode == MANAGE_OBJECTS )
	    {
		objectStrengthField.setText( String.valueOf( -1 ) );
		objectCostToMoveField.setText( String.valueOf( -1 ) );
		currentObject = null;
	    }
	    mapPanel.repaint();
	}
    }
    
    public void actionPerformed( ActionEvent e )
    {
	if( e.getSource() == addCellsButton )
	{
	    addCells();
	}
	if( e.getSource() == addRoomButton )
	{
	    addRoom();
	}
	if( e.getSource() == removeRoomButton )
	{
	    if( currentRoom != null )
	    {
		removeCurrentRoom();
	    }
	}
	if( e.getSource() == addObjectButton )
	{
	    addObject();
	}
	if( e.getSource() == removeObjectButton )
	{
	    removeCurrentObject();   
	}
	if( e.getSource() == showLightsBox )
	{
	    mapPanel.setDrawLights( showLightsBox.isSelected() );
	    mapPanel.repaint();
	}
        if( e.getSource() == newMap )
        {
            if( edited )
            {
                int selection = JOptionPane.showConfirmDialog( this, "Map is not saved. Do you want to proceed?", "", JOptionPane.YES_NO_CANCEL_OPTION );
                if( selection == JOptionPane.CANCEL_OPTION || selection == JOptionPane.NO_OPTION || selection == JOptionPane.CLOSED_OPTION )
                {
                    return;
                }
            }
            String sSize = JOptionPane.showInputDialog( this, "How many cells per map edge?", "Map size", JOptionPane.QUESTION_MESSAGE );
            int size;
            try
            {
                size = Integer.parseInt( sSize );
            }
            catch( NumberFormatException nfe )
            {
                size = 50;
            }
            mapFile = null;
            map = new Map( size, size );
            map.addFloorSet( "ftiles.jpg" );
            map.addFloorSet( "doorbase.png" );
            map.addWallSet( "walltiles.png" );
            map.addDoorSet( "doors.png" );
            map.addObjectSet( "objects.png" );
            mapPanel.changeMap( map );
            overview.setMap( map );
            edited = false;
            validate();
        }
        if( e.getSource() == save )
        {
            if( mapFile == null )
            {
                saveMapWithFileChooser();
            }
            else
            {
                saveMap();
            }
        }	
	if( e.getSource() == saveAs )
	{
	    saveMapWithFileChooser();
	}
	if( e.getSource() == openOld )
	{
            System.out.println( "Open old map" );
	    openOldMapWithFileChooser();
	}
        if( e.getSource() == open )
        {
            System.out.println( "Open map" );
            openMapWithFileChooser();
        }
	if( e.getSource() == mapButton )
	{
	    setMode( PAINT_MAP );   
	}
	if( e.getSource() == selectionButton )
	{
	    setMode( SELECT );   
	}
	if( e.getSource() == roomButton )
	{
	    setMode( MANAGE_ROOMS );
	}
	if( e.getSource() == objectsButton )
	{
	    setMode( MANAGE_OBJECTS );   
	}
	requestFocus();
    }
    
    public void addCells()
    {
	JFileChooser fileChooser = new JFileChooser( missionDirectoryPath );
	int index = selectors.getSelectedIndex();

	SimpleFileFilter filter = filter = new SimpleFileFilter( ".png" );
	if( index == 3 )
	{
	    filter = new SimpleFileFilter( ".xml" );	    
	}
	fileChooser.setFileFilter( filter );
	fileChooser.setDialogTitle( "Choose file" );
	int returnVal = fileChooser.showOpenDialog( this );
	if( returnVal == JFileChooser.APPROVE_OPTION )
	{
	    String fileName = fileChooser.getSelectedFile().getName();
	    
	    if( index == 0 )
	    {
		System.out.println( "Floors" );
		Vector images = new Vector();
		ResourceManager.cropImage( fileName, images, cellSize );
		floorCellSelector.appendImages( images );
		map.addFloorSet( fileName );
	    }
	    
	    if( index == 1 )
	    {
		System.out.println( "Walls" );
		Vector images = new Vector();
		ResourceManager.cropImage( fileName, images, cellSize );
		wallCellSelector.appendImages( images );
		map.addWallSet( fileName );
	    }
	    if( index == 3 )
	    {
		System.out.println( "Objects" );
		try
		{
		    loadMapObjectCells( fileName );
		    //Vector images = new Vector();
		    //ResourceManager.cropImage( fileName, images, cellSize );
		    mapPanel.loadObjectsFromMap();
		    objectCellSelector.setImages( mapPanel.getObjectCells() );
		}
		catch( Exception e )
		{
		    System.err.println( "Error loading objects!" );
		    e.printStackTrace();
		}
	    }
	}
	    
    }
    
    public void addObject()
    {
	String objectName;
	int strength = -1;
	int costToMove = -1;
	objectName = objectNameField.getText().trim();
	try
	{
	    strength = Integer.parseInt( objectStrengthField.getText() );
	    costToMove = Integer.parseInt( objectCostToMoveField.getText() );
	}
	catch( NumberFormatException nfe )
	{
	    Toolkit.getDefaultToolkit().beep();
	    return;
	}
	if( selectedObjectCells != null && currentObject == null )
	{
	    MapObject mapObject = new MapObject();
	    mapObject.setName( objectName );
	    for( int i=0; i < selectedObjectCells.length; i++ )
	    {
		mapObject.addObjectCell( selectedObjectCells[ i ] );
	    }
	    strength = Math.max( strength, mapObject.getStrength() );
	    mapObject.setStrength( strength );
	    System.out.println( "Adding object: " + mapObject );
	    for( int x=mapObject.getTopLeft().x; x <= mapObject.getBottomRight().x; x++ )
	    {
		for( int y=mapObject.getTopLeft().y; y <= mapObject.getBottomRight().y; y++ )
		{
		    map.removeUnboundObjectCells( x, y );
		}
	    }
	    map.addObject( mapObject );
	    map.clearSelection();
	    currentObject = mapObject;
	    displayObjectInfo( currentObject );
	}
	if( currentObject != null )
	{
	    currentObject.setName( objectName );
	    currentObject.setStrength( strength );
	    currentObject.setCostToMove( costToMove );
	    displayObjectInfo( currentObject );
	}
	mapPanel.repaint();
    }
    
    public void addRoom()
    {
	int startX = -1;
	int startY = -1;
	int endX = -1;
	int endY = -1;
	int hotSpotX = -1;
	int hotSpotY = -1;
	try
	{
	    startX = Integer.parseInt( roomStartXField.getText() );
	    startY = Integer.parseInt( roomStartYField.getText() );
	    endX = Integer.parseInt( roomEndXField.getText() );
	    endY = Integer.parseInt( roomEndYField.getText() );
	    hotSpotX = Integer.parseInt( roomHotSpotXField.getText() );
	    hotSpotY = Integer.parseInt( roomHotSpotYField.getText() );

	}
	catch( NumberFormatException nfe )
	{
	    displayRoomInfo( null );
	    Toolkit.getDefaultToolkit().beep();
	    return;
	}
	if( startX < 0 || startY < 0 || endX < 0 || endY < 0 || 
	    startX >= map.getHorizontalSize() || startY >= map.getVerticalSize() ||
	    endX >= map.getHorizontalSize() || endY >= map.getVerticalSize() 
	    )
	{
	    Toolkit.getDefaultToolkit().beep();
	    return;	    
	}
	Point start = new Point( startX, startY );
	Point end = new Point( endX, endY );
	Room room;
	if( currentRoom != null )
	{
	    room = currentRoom;
	}
	else
	{
	    room = new Room();
	}
	room.setTopLeft( start );
	room.setBottomRight( end );
	room.setName( roomNameField.getText().trim() );
	if( hotSpotX < 0 || hotSpotX >= map.getHorizontalSize() || hotSpotY < 0 || hotSpotY >= map.getVerticalSize() )
	{
	    hotSpotX = (startX+endX)/2;
	    hotSpotY = (startY+endY)/2;
	}
	Point hotSpot = new Point( hotSpotX, hotSpotY );
	room.setHotSpot( hotSpot );
	if( currentRoom == null )
	{
	    map.addRoom( room );
	    edited = true;
	}
	currentRoom = room;
	displayRoomInfo( currentRoom );
	map.clearSelection();
	mapPanel.repaint();
    }

    public void saveMap()
    {
        if( edited )
        {
            try
            {
                map.save( mapFile );
                edited = false;
            }
            catch( FileNotFoundException e )
            {
                System.out.println( "File not found!" );
            }
            catch( IOException e )
            {
                System.out.println( e.toString() );
            }
	    catch( Exception e )
	    {
		System.out.println( e.toString() );
	    }
        }
    }
    
    public void saveMapWithFileChooser()
    {
        File file;
	JFileChooser fileChooser = new JFileChooser( missionDirectoryPath );
	SimpleFileFilter filter = new SimpleFileFilter( ".map" );
	fileChooser.setFileFilter( filter );
	fileChooser.setDialogTitle( "Save map" );
	int returnVal = fileChooser.showSaveDialog( this );
	if( returnVal == JFileChooser.APPROVE_OPTION )
	{
            file = fileChooser.getSelectedFile();
            if( file.exists() )
            {
                int selection = JOptionPane.showConfirmDialog( this, "Are you sure you want to overwrite the file?", "", JOptionPane.YES_NO_CANCEL_OPTION );
                if( selection == JOptionPane.CANCEL_OPTION || selection == JOptionPane.NO_OPTION || selection == JOptionPane.CLOSED_OPTION )
                {
                    return;
                }
            }
	    try
	    {
		map.setName( nameField.getText() );
		mapFile = file;
		map.save( mapFile );
		setTitle( "Mapmaker: " + fileChooser.getSelectedFile().getName() );
                edited = false;
	    }
	    catch( FileNotFoundException e )
	    {
		System.out.println( "File not found!" );
	    }
	    catch( IOException e )
	    {
		System.out.println( e.toString() );
	    }
	    catch( Exception e )
	    {
		System.out.println( e.toString() );
	    }
	    
	}
    }
    
    public void openMapWithFileChooser()
    {
	JFileChooser fileChooser = new JFileChooser( missionDirectoryPath );
	SimpleFileFilter filter = new SimpleFileFilter( ".map" );
	fileChooser.setFileFilter( filter );
	fileChooser.setDialogTitle( "Open map" );
	int returnVal = fileChooser.showOpenDialog( this );
	if( returnVal == JFileChooser.APPROVE_OPTION )
	{
	    try
	    {
		map = Map.load( fileChooser.getSelectedFile() );
                mapFile = fileChooser.getSelectedFile();
		setTitle( "Mapmaker: " + fileChooser.getSelectedFile().getName() );
		nameField.setText( map.getName() );
		mapPanel.changeMap( map );
		overview.setMap( map );
                edited = false;
		floorCellSelector.setImages( mapPanel.getFloorCells() );
		wallCellSelector.setImages( mapPanel.getWallCells() );
		objectCellSelector.setImages( mapPanel.getObjectCells() );
	    }
	    catch( FileNotFoundException e )
	    {
		System.out.println( "File not found!" );
	    }
	    catch( IOException e )
	    {
		System.out.println( e.toString() );
	    }
	    catch( Exception e )
	    {
		e.printStackTrace();
	    }
	}	
    }

    public void openOldMapWithFileChooser()
    {
        JFileChooser fileChooser = new JFileChooser();
        SimpleFileFilter filter = new SimpleFileFilter( ".oldmap" );
        fileChooser.setFileFilter( filter );
        fileChooser.setDialogTitle( "Open map" );
        int returnVal = fileChooser.showOpenDialog( this );
        if( returnVal == JFileChooser.APPROVE_OPTION )
        {
            try
        {
            map = Map.loadOld( fileChooser.getSelectedFile() );
            setTitle( "Mapmaker: " + fileChooser.getSelectedFile().getName() );
            nameField.setText( map.getName() );
            mapPanel.changeMap( map );
            overview.setMap( map );
        }
            catch( FileNotFoundException e )
        {
                System.out.println( "File not found!" );
        }
            catch( IOException e )
        {
                System.out.println( e.toString() );
        }
        }
    }
        
    public void mouseEntered( MouseEvent e )
    {}
    
    public void mouseExited( MouseEvent e )
    {}
    
    public void mousePressed( MouseEvent e )
    {
	Point cellPlace;
	if( e.getSource() == mapPanel )
	{
	    cellPlace = mapPanel.getCell( e.getPoint() );


	    if( mode == MANAGE_ROOMS )
	    {
		currentRoom = map.getRoomForPoint( cellPlace );
		displayRoomInfo( currentRoom );
		map.clearSelection();   
	    }
	}
    }
    
    public void mouseReleased( MouseEvent e )
    {
	if( e.getSource() == mapPanel )
	{
	    if( painting )
	    {
		painting = false;
		if( mode == PAINT_MAP )
		{
		    for( int i = paintFrom.x; i <= paintTo.x; i++ )
		    {
			for( int j = paintFrom.y; j <= paintTo.y; j++ )
			{
			    map.setUtility( i, j, 0 );
			    if( applyLightBox.isSelected() )
			    {
				map.setLight( i, j, (byte)Integer.parseInt( (String)lightLevels.getSelectedItem() ) );
			    }
			    if( applyCellBox.isSelected() && leftSelectedType == FLOOR )
			    {
				edited = true;
				if( SwingUtilities.isRightMouseButton( e ) )
				{
				    if( (i%2==0 && j%2==1) || (i%2==1 && j%2==0) )
				    {
					map.setCellFloor( i, j, leftSelectedValue );			
				    }
				    else
				    {
					map.setCellFloor( i, j, rightSelectedValue );
				    }
				}
				else
				{
				    if( e.isShiftDown() )
				    {
					map.addCellFloor( i, j, leftSelectedValue );
				    }
				    else
				    {
					map.setCellFloor( i, j, leftSelectedValue );
				    }
				}
			    }
			    if( applyCellBox.isSelected() && leftSelectedType == WALL )
			    {
				edited = true;
				map.setCellWall( i, j, leftSelectedValue );
			    }
			}
		    }
		}
		paintFrom = null;
		paintTo = null;
		if( mode == PAINT_MAP )
		{
		    map.clearSelection();
		}
	    }
	    if( mode == MANAGE_ROOMS && currentRoom != null )
	    {
		Point cellPlace = mapPanel.getCell( e.getPoint() );
		Room room = map.getRoomForPoint( cellPlace );
		if( room != null && room != currentRoom )
		{
		    room.addLinkedRoom( currentRoom );
		    map.clearSelection();
		    edited = true;
		}
		currentRoom = room;
		displayRoomInfo( currentRoom );
	    }
	    if( mode == MANAGE_OBJECTS )
	    {
		selectedObjectCells = map.getSelectedUnboundObjectCells();
		addObjectButton.setEnabled( selectedObjectCells != null );
		addObjectButton.setToolTipText( "Add object" );
		if( selectedObjectCells != null )
		{
		    int hitPoints = -1;
		    int costToMove = 0;
		    for( int i=0; i < selectedObjectCells.length; i++ )
		    {
			if( selectedObjectCells[ i ].getHitPoints() > hitPoints )
			{
			    hitPoints = selectedObjectCells[ i ].getHitPoints();
			}
			if( selectedObjectCells[ i ].getCostToMove() > costToMove )
			{
			    costToMove = selectedObjectCells[ i ].getCostToMove();
			}
			if( selectedObjectCells[ i ].getCostToMove() < 0 )
			{
			    costToMove = -1;
			}
		    }
		    objectNameField.setText( selectedObjectCells[0].getName() );
		    objectStrengthField.setText( String.valueOf( hitPoints ) );
		    objectCostToMoveField.setText( String.valueOf( costToMove ) );
		}
	    }
	}
	mapPanel.repaint();
	overview.repaint();
	requestFocus();
    }

    public void openSerializedWithFileChooser()
    {
	JFileChooser fileChooser = new JFileChooser();
	SimpleFileFilter filter = new SimpleFileFilter( ".smap" );
	fileChooser.setFileFilter( filter );
	fileChooser.setDialogTitle( "Open map" );
	int returnVal = fileChooser.showOpenDialog( this );
	if( returnVal == JFileChooser.APPROVE_OPTION )
	{
	    try
	    {
                FileInputStream fis = new FileInputStream( fileChooser.getSelectedFile() );
                ObjectInputStream ois = new ObjectInputStream( fis );
		map = (Map)ois.readObject();
		setTitle( "Mapmaker: " + fileChooser.getSelectedFile().getName() );
		nameField.setText( map.getName() );
		mapPanel.changeMap( map );
		overview.setMap( map );
	    }
	    catch( FileNotFoundException e )
	    {
		System.out.println( "File not found!" );
	    }
	    catch( IOException e )
	    {
		System.out.println( e.toString() );
	    }
            catch( ClassNotFoundException e )
            {
                System.out.println( e.toString() );
            }
	}	
    }

    public void removeCurrentRoom()
    {
	if( currentRoom != null )
	{
	    map.removeRoom( currentRoom );
	    currentRoom = null;
	    displayRoomInfo( currentRoom );
	    mapPanel.repaint();
	}
    }
    
    public void removeCurrentObject()
    {
	if( currentObject != null )
	{
	    map.removeObject( currentObject );
	    currentObject = null;
	    displayObjectInfo( currentObject );
	    mapPanel.repaint();
	}
    }
    
    public void serializeWithFileChooser()
    {
	JFileChooser fileChooser = new JFileChooser();
	SimpleFileFilter filter = new SimpleFileFilter( ".smap" );
	fileChooser.setFileFilter( filter );
	fileChooser.setDialogTitle( "Save map" );
	int returnVal = fileChooser.showSaveDialog( this );
	if( returnVal == JFileChooser.APPROVE_OPTION )
	{
	    try
	    {
		map.setName( nameField.getText() );
		saveMapBySerial( fileChooser.getSelectedFile() );
		setTitle( "Mapmaker: " + fileChooser.getSelectedFile().getName() );
	    }
	    catch( FileNotFoundException e )
	    {
		System.out.println( "File not found!" );
	    }
	    catch( IOException e )
	    {
		System.out.println( e.toString() );
	    }
	}
    }

    public void saveMapBySerial( File file ) throws IOException, FileNotFoundException
    {
        FileOutputStream fos = new FileOutputStream( file );
        ObjectOutputStream oos = new ObjectOutputStream( fos );
        oos.writeObject( map );
        oos.flush();
        oos.close();
        fos.close();
        
    }

    public void setMode( int mode )
    {
	if( this.mode != SELECT )
	{
	    map.clearSelection();
	    mapPanel.repaint();
	    selectionButton.setSelected( true );
	}
	this.mode = mode;
	if( mode == MANAGE_ROOMS )
	{
	    detailViewSwitcher.show( detailViewsPanel, ROOM_DETAILS_KEY );
	    roomButton.setSelected( true );
	}
	if( mode == PAINT_MAP )
	{
	    detailViewSwitcher.show( detailViewsPanel, PAINT_DETAILS_KEY );
	    mapButton.setSelected( true );
	}
	if( mode == MANAGE_OBJECTS )
	{
	    detailViewSwitcher.show( detailViewsPanel, OBJECT_DETAILS_KEY );
	    objectsButton.setSelected( true );
	}
	mapPanel.setDrawRooms( mode == MANAGE_ROOMS );
	mapPanel.setDrawObjectBounds( mode == MANAGE_OBJECTS );
	mapPanel.repaint();
    }
    
    
    public void mouseMoved( MouseEvent e )
    {}
    
    public static void main( String [] args ) 
    {
	new MapMaker();
    }
}
