// Map.java
// Started on June 16th, 2001
// Copyright (c) Virtu Halttunen

import TheGame.*;
import java.awt.Point;
import java.util.*;
import java.io.*;
import java.awt.geom.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;
import java.util.zip.*;

public class EditorMap
{
    public static final int NORTH = 0;
    public static final int NORTH_EAST = 1;
    public static final int EAST = 2;
    public static final int SOUTH_EAST = 3;
    public static final int SOUTH = 4;
    public static final int SOUTH_WEST = 5;
    public static final int WEST = 6;
    public static final int NORTH_WEST = 7;
    
    private MapCell[][] cellMap;
    private Vector doors;
    private Vector objects;
    private HashMap rooms;
    private short[][] utilityMap;
    private int horizontalSize;
    private int verticalSize;
    private Vector floorSetNames;
    private Vector wallSetNames;
    private Vector objectSetNames;
    private Vector doorSetNames;
    private String name;
    private MissionEditor editor;
    private boolean doorsLocked = false;
    
    public EditorMap( int horizontalSize, int verticalSize )
    {
	this.horizontalSize = horizontalSize;
	this.verticalSize = verticalSize;
        cellMap = new MapCell[ horizontalSize ][ verticalSize ];
        doors = new Vector();
        objects = new Vector();
	rooms = new HashMap();
	utilityMap = new short[ horizontalSize ][ verticalSize ];
	floorSetNames = new Vector();
	wallSetNames = new Vector();
	objectSetNames = new Vector();
	doorSetNames = new Vector();
    }
    
    public void setEditor( MissionEditor editor )
    {
	this.editor = editor;
    }
    
    public void setCell( int x, int y, MapCell cell )
    {
        cellMap[x][y] = cell;
    }

    public MapCell getCell( int x, int y )
    {
	//System.out.println( "Map: getCellType("+x+","+y+")" );
	return cellMap[x][y];
    }
    
    public MapCell getCell( Point point )
    {
	//System.out.println( "Map: getCellType() " + point.toString() );
	return cellMap[ point.x ][ point.y ];
    }
                
    public boolean isPassable( int x, int y )
    {
        MapObject mapObject;
        if( x < 0 ) return false;
        if( x >= horizontalSize ) return false;
        if( y < 0 ) return false;
        if( y >= verticalSize ) return false;
        if( cellMap[x][y] == null )
            return false;
	if( cellMap[x][y].getWall() > -1 )
	{
	    return false;
	}
        else
        {
            if( cellMap[x][y].hasDoor() )
            {
                if( getDoor( x, y ).isOpen() )
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            if( cellMap[x][y].hasObjects() )
            {
                for( int i = 0; i < getNumberOfObjects(); i++ )
                {
                    mapObject = getObject( i );
                    if( mapObject.isLocatedIn( x, y ) && mapObject.isBlockingMovement( new Point( x, y ) ) )
                    {
                        return false;
                    }
                }
            }
        }
        return true;
    }
 
    public boolean isPassable( Point cell )
    {
        return isPassable( cell.x, cell.y );
    }
    
    public boolean isPassable( Point cell, int direction )
    {
	Point target = getNeighbor( cell, direction );
	if( target.x < 0 ) return false;
	if( target.x >= horizontalSize ) return false;
	if( target.y < 0 ) return false;
	if( target.y >= verticalSize ) return false;
	return isPassable( target );
    }
    
    public boolean isPassable( Point point, boolean canOpenDoors )
    {
        if( !canOpenDoors )
            return isPassable( point );
        boolean passable = isPassable( point );
        passable = ( passable || isClosedDoor( point.x, point.y ) );
        return passable;
    }
    
    public boolean isPassable( int x, int y, boolean canOpenDoors )
    {
        if( !canOpenDoors )
            return isPassable( x, y );
        boolean passable = isPassable( x, y );
        passable = ( passable || isClosedDoor( x, y ) );
        return passable;
    }
    
    public boolean isPassable( Point cell, int direction, boolean canOpenDoors )
    {
        if( !canOpenDoors )
            return isPassable( cell, direction );
        boolean passable = isPassable( cell, direction );
        passable = ( passable || isClosedDoor( cell.x, cell.y ) );
        return passable;
    }

    public boolean isClosedDoor( Point point )
    {
        if( cellMap[ point.x ][ point.y ] == null )
        {
            return false;
        }
        else
        {
            if( cellMap[ point.x ][ point.y ].hasDoor() )
            {
                Door door = getDoor( point.x, point.y );
                return door.isClosed();
            }
        }
        return false;
    }


    public boolean isClosedDoor( int x, int y )
    {
        if( cellMap[ x ][ y ] == null )
        {
            return false;
        }
        else
        {
            if( cellMap[ x ][ y ].hasDoor() )
            {
                Door door = getDoor( x, y );
                return door.isClosed();
            }
        }
        return false;
    }

    public boolean isClosedDoor( Point cell, int direction )
    {
	Point target = getNeighbor( cell, direction );
	if( target.x < 0 ) return false;
	if( target.x >= horizontalSize ) return false;
	if( target.y < 0 ) return false;
	if( target.y >= verticalSize ) return false;
	return isClosedDoor( target.x, target.y );
    }
    
    public boolean isOpenDoor( Point point )
    {
        if( cellMap[ point.x ][ point.y ] == null )
        {
            return false;
        }
        else
        {
            if( cellMap[ point.x ][ point.y ].hasDoor() )
            {
                Door door = getDoor( point.x, point.y );
                return !door.isClosed();
            }
        }
        return false;
    }

    
    public boolean isOpenDoor( int x, int y )
    {
        if( cellMap[ x ][ y ] == null )
        {
            return false;
        }
        else
        {
            if( cellMap[ x ][ y ].hasDoor() )
            {
                Door door = getDoor( x, y );
                return !door.isClosed();
            }
        }
        return false;
    }

    public boolean isOpenDoor( Point cell, int direction )
    {
	Point target = getNeighbor( cell, direction );
	if( target.x < 0 ) return false;
	if( target.x >= horizontalSize ) return false;
	if( target.y < 0 ) return false;
	if( target.y >= verticalSize ) return false;
	return isOpenDoor( target.x, target.y );
    }
   
    public static Point getNeighbor( Point cell, int direction )
    {
	Point target = new Point( cell );
	switch( direction )
	{
	    case NORTH: target.y--; break;
	    case NORTH_EAST: target.x++; target.y--; break;
	    case EAST: target.x++; break;
	    case SOUTH_EAST: target.x++; target.y++; break;
	    case SOUTH: target.y++; break;
	    case SOUTH_WEST: target.x--; target.y++; break;
	    case WEST: target.x--; break;
	    case NORTH_WEST: target.x--; target.y--; break;
	}
	return target;
    }
    public static boolean isNeighbor( Point one, Point two )
    {
        if( Math.abs( one.x - two.x ) <= 1 && Math.abs( one.y - two.y ) <= 1 )
            return true;
        return false;
    }
    public void addDoor( Door door )
    {
        if( cellMap[ door.getX() ][ door.getY() ] != null )
        {
            if( cellMap[ door.getX() ][ door.getY() ].hasDoor() )
            {
                removeDoor( door.getX(), door.getY() );
            }
            cellMap[ door.getX() ][ door.getY() ].setDoor( true );
            doors.add( door );
        }
    }

    public void removeDoor( int x, int y )
    {
        if( cellMap[x][y].hasDoor() )
        {
            Door theDoor = getDoor( x, y );
            doors.remove( theDoor );
            cellMap[x][y].setWall( -1 );
        }
    }
    
    public Door getDoor( int x, int y )
    {
        Door theDoor = null;
        for( int i = 0; i < doors.size(); i++ )
        {
            if( ((Door)doors.elementAt( i )).isLocatedIn( x, y ) )
                theDoor = (Door)doors.elementAt( i );
        }
        return theDoor;
    }

    public void openDoor( Point point )
    {
        if( cellMap[point.x][point.y].hasDoor() )
        {
            Door door = getDoor( point.x, point.y );
            if( door.isClosed() && !door.isLocked() )
            {
                door.setOpen( true );
                openDoor( getNeighbor( point, door.getDirection() ) );
            }
        }
    }

    public void closeDoor( Point point )
    {
        if( cellMap[point.x][point.y].hasDoor() )
        {
            Door door = getDoor( point.x, point.y );
            if( !door.isClosed() && !door.isLocked() )
            {
                door.setOpen( false );
                closeDoor( getNeighbor( point, door.getDirection() ) );
            }
        }
    }


    public void lockAllDoors()
    {
        for( int i = 0; i < doors.size(); i++ )
        {
            ((Door)doors.elementAt(i)).setLocked( true );
        }
    }
    
    public static int getDirection( Point from, Point to, int originalDirection )
    {
	int direction = originalDirection;
	if( !from.equals( to ) )
	{
	    // Uses Arc2D to save the trouble of calculating the angle between the points
	    Arc2D.Double helper = new Arc2D.Double( from.x-1, from.y-1, 2, 2, 0.0, 0.0, Arc2D.OPEN );
	    helper.setAngles( new Point( from.x, from.y-1 ), to );
	    int angle = (int)helper.getAngleExtent();
	    if( angle < 23 ) direction = NORTH;
	    if( angle >= 23 && angle < 68 ) direction = NORTH_WEST;
	    if( angle >= 68 && angle < 113 ) direction = WEST;
	    if( angle >= 113 && angle < 158 ) direction = SOUTH_WEST;
	    if( angle >= 158 && angle < 203 ) direction = SOUTH;
	    if( angle >= 203 && angle < 248 ) direction = SOUTH_EAST;
	    if( angle >= 248 && angle < 293 ) direction = EAST;
	    if( angle >= 293 && angle < 338 ) direction = NORTH_EAST;
	    if( angle >= 338 ) direction = NORTH;
	}
	
	return direction;
    }
   
    public boolean hasLineOfSight( Point start, Point end )
    {
        if( start.x < 0 || start.x >= getHorizontalSize() ) return false;
        if( start.y < 0 || start.y >= getVerticalSize() ) return false;
        if( end.x < 0 || end.x >= getHorizontalSize() ) return false;
        if( end.y < 0 || end.y >= getVerticalSize() ) return false;
	// Uses Bresenham's line drawing algorithm with a small change
	int dX = Math.abs( end.x-start.x );	// Difference in coordinates
	int dY = Math.abs( end.y-start.y );
	int checkX = start.x;
	int checkY = start.y;	
	
	int xIncr, yIncr;
	if( start.x > end.x ) 
	    xIncr=-1;
	else
	    xIncr=1;
	    
	if( start.y > end.y )
	    yIncr=-1;
	else
	    yIncr=1;
			    
	if( dX >= dY )
	{
	    int dPr = dY*2;		// Amount to increment decision to right
	    int dPru = dPr - (dX*2); 	// Amount to increment up
	    int p = dPr - dX;		// decision start value
	    
	    while( dX > 0 )
	    {
		if( p > 0 )	// Is the next one up and right 
		{
		    checkX+=xIncr;
		    checkY+=yIncr;
		    p+=dPru;
		}
		else		// Or just right
		{
		    checkX+=xIncr;
		    p+=dPr;
		}
		dX--;
		// Check the cell
		if( !isPassable( checkX, checkY ) )
		{
		    return false;
		}
	    }
	}
	else
	{
	    int dPr = dX*2;
	    int dPru = dPr - (dY*2);
	    int p = dPr-dY;
	    
	    while( dY>0 )
	    {

		if( p > 0 )
		{
		    checkX+=xIncr;
		    checkY+=yIncr;
		    p+=dPru;
		}
		else
		{
		    checkY+=yIncr;
		    p+=dPr;
		}
		dY--;
		//Check the cell
		if( !isPassable( checkX, checkY ) )
		{
		    return false;
		}
	    }
	}
	return true;
    }
   
    public void resetUtilityMap()
    {
	for( int i=0; i < horizontalSize; i++ )
	    for( int j=0; j < verticalSize; j++ )
		utilityMap[i][j]=0;
    }
    
    private Point findPrevious( Point point )
    {
	int level = utilityMap[point.x][point.y];
        int smallest;
	int direction;
	Point testPoint;
	Point previous = null;
        smallest = level;
	for( int i = 7; i>=0; i-- )
	{
	    testPoint = getNeighbor( point, i );
            if( (utilityMap[testPoint.x][testPoint.y] <= smallest) && (utilityMap[testPoint.x][testPoint.y] > 0) )
            {
                smallest = utilityMap[testPoint.x][testPoint.y];
                previous = testPoint;
            }
	}
        // Extra check for CommSentry and doors
        if( previous == null )
        {
            System.out.println( "Previous node on path not found! Point: " + point.toString() + " at level " + level);
            for( int i=0; i<verticalSize; i++ )
            {
                for( int j=0; j<horizontalSize; j++ )
                {
                    System.out.print( utilityMap[j][i] +" " );
                    if( utilityMap[j][i] < 10 )
                        System.out.print( " " );
                }
                System.out.print("\n");
            }
        }
	return previous;
    }

    public void setCellFloor( int x, int y, int type )
    {
        if( cellMap[x][y] == null )
        {
            cellMap[x][y] = new MapCell();
        }
        cellMap[x][y].setFloor( type );
    }

    public void addCellFloor( int x, int y, int type )
    {
        if( cellMap[x][y] == null )
        {
            cellMap[x][y] = new MapCell();
            cellMap[x][y].setFloor( type );
        }
        else
        {
            cellMap[x][y].addFloor( type );
        }
    }

    public void setCellWall( int x, int y, int type )
    {
        if( cellMap[x][y] == null )
        {
            cellMap[x][y] = new MapCell();
        }
        cellMap[x][y].setWall( type );
    }
    
    public byte getLight( int x, int y )
    {
        if( cellMap[x][y] != null )
            return cellMap[x][y].getLight();
        return 0;
    }

    public byte getLight( Point point )
    {
        if( cellMap[ point.x ][ point.y ] != null )
            return cellMap[ point.x ][ point.y ].getLight();
        return 0;
    }

    public void setLight( int x, int y, byte light )
    {
        if( cellMap[x][y] != null )
            cellMap[x][y].setLight( light );
    }
    
    public int getUtility( int x, int y )
    {
	return utilityMap[x][y];
    }
    
    public void setUtility( int x, int y, int value )
    {
	utilityMap[x][y] = (short)value;
    }
    
    public void addObject( MapObject object )
    {
	if( !objects.contains( object ) )
	{
	    objects.add( object );
	    for( int x=object.getTopLeft().x; x <=object.getBottomRight().x; x++ )
	    {
		for( int y=object.getTopLeft().y; y <= object.getBottomRight().y; y++ )
		{
		    if( object.isLocatedIn( x, y ) )
		    {
			cellMap[x][y].setObjects( true );
		    }
		}
	    }
	}
    }
    
    public void removeObject( MapObject object )
    {
	if( objects.contains( object ) )
	{
	    objects.remove( object );
	    MapObject otherObject;
	    for( int x=object.getTopLeft().x; x <=object.getBottomRight().x; x++ )
	    {
		for( int y=object.getTopLeft().y; y <= object.getBottomRight().y; y++ )
		{
		    cellMap[x][y].setObjects( false );
		    for( int c=0; c < objects.size(); c++ )
		    {
			otherObject = getObject( c );
			if( otherObject.isLocatedIn( x, y ) )
			{
			    cellMap[x][y].setObjects( true );
			}
		    }
		}
	    }
	    
	}
    }
    
    public void addRoom( Room room )
    {
	if( !rooms.containsKey( room.getId() ) )
	{
	    rooms.put( room.getId(), room );
	    System.out.println( "Added room: " + room.getName() + " (" + room.getId() +")" );
	}
    }
    
    public Room getRoom( String id )
    {
	return (Room)rooms.get( id );
    }
    
    public Room getRoomForPoint( Point point )
    {
	Room [] theRooms = getRooms();
	for( int i=0; i < theRooms.length; i++ )
	{
	    if( point.x >= theRooms[i].getTopLeft().x && point.x <= theRooms[i].getBottomRight().x &&
		point.y >= theRooms[i].getTopLeft().y && point.y <= theRooms[i].getBottomRight().y )
	    {
		// Found the room
		return theRooms[i];
	    }
	}
	return null;
    }
    
    public Room [] getRooms()
    {
	Collection collection = rooms.values();
	Room [] theRooms = new Room[ collection.size() ];
	
	Iterator iterator = collection.iterator();
	int i=0;
	while( iterator.hasNext() )	
	{
	    theRooms[ i++ ] = (Room)(iterator.next());
	}
	return theRooms;
    }
    
    public void removeRoom( Room room )
    {
	if( rooms.containsKey( room.getId() ) )
	{
	    String [] linkedRoomIds = room.getLinkedRoomIds();
	    Room otherRoom;
	    for( int i=0; i < linkedRoomIds.length; i++ )
	    {
		String id = linkedRoomIds[ i ];
		otherRoom = (Room)rooms.get( id );
		if( otherRoom != null )
		{
		    otherRoom.removeLinkedRoom( room );
		}
	    }
	    rooms.remove( room.getId() );
	}
    }
    
    public int getNumberOfObjects()
    {
        return objects.size();
    }

    public MapObject getObject( int index )
    {
        return (MapObject)objects.elementAt( index );
    }

    public MapObject getObjectForPoint( Point point )
    {
	if( cellMap[ point.x ][ point.y ] == null )
	{
	    return null;   
	}
	if( cellMap[ point.x ][ point.y ].hasObjects() )
	{
	    for( int i=0; i < objects.size(); i++ )
	    {
		MapObject object = getObject( i );
		if( object.isLocatedIn( point.x, point.y ) )
		{
		    return object;   
		}
	    }
	}
	return null;
    }    
    
    public void addWallSet( String setName )
    {
	wallSetNames.add( new String( setName ) );
    }
    
    public Vector getWallSets()
    {
	return wallSetNames;
    }
    
    public void addFloorSet( String setName )
    {
	floorSetNames.add( new String( setName ) );
    }
    
    public Vector getFloorSets()
    {
	return floorSetNames;
    }
    
    public void addDoorSet( String setName )
    {
	doorSetNames.add( new String( setName ) );
    }
    
    public Vector getDoorSets()
    {
	return doorSetNames;
    }
    
    public void addObjectSet( String setName )
    {
	objectSetNames.add( new String( setName ) );
    }
    
    public Vector getObjectSets()
    {
	return objectSetNames;
    }
    
    public void setName( String name )
    {
	this.name = new String( name );
    }
    
    public String getName()
    {
	return new String( name );
    }
    
    public int getHorizontalSize()
    {
	return horizontalSize;
    }
    
    public int getVerticalSize()
    {
	return verticalSize;
    }

    public static EditorMap load( File file ) throws Exception
    {
	EditorMap newMap = null;
	
	if( file.exists() && file.canRead() )
	{
	    DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
	    DOMImplementation builder = documentBuilder.getDOMImplementation();
	    
	    FileInputStream fis = new FileInputStream( file );
	    GZIPInputStream gis = new GZIPInputStream( fis );
	    
	    System.out.println( "Parsing map file: " + file.getName() );
	    Document document = documentBuilder.parse( gis );
	    gis.close();
	    fis.close();
	    
	    Element root = document.getDocumentElement();
	    String mapName = root.getAttribute( "name" );
	    System.out.println( "Map name: " + mapName );
	    int mapWidth = Integer.parseInt( root.getAttribute( "width" ) );
	    int mapHeight = Integer.parseInt( root.getAttribute( "height" ) );
	    newMap = new EditorMap( mapWidth, mapHeight );
	    newMap.setName( mapName );
	    Element setsElement = XMLUtils.getNamedChild( root, "graphicssets" );
	    Element [] floorSets = XMLUtils.getNamedChildren( setsElement, "floorset" );
	    System.out.println( "Reading floor sets" );
	    for( int i=0; i < floorSets.length; i++ )
	    {
		if( i != Integer.parseInt( floorSets[i].getAttribute( "index" ) ) )
		{
		    System.out.println( "Index mismatch!" );
		}
		String setName = floorSets[i].getAttribute( "file" );
		if( setName != null )
		{
		    newMap.addFloorSet( setName );
		}
	    }
	    Element [] wallSets = XMLUtils.getNamedChildren( setsElement, "wallset" );
	    System.out.println( "Reading wall sets" );
	    for( int i=0; i < wallSets.length; i++ )
	    {
		if( i != Integer.parseInt( wallSets[i].getAttribute( "index" ) ) )
		{
		    System.out.println( "Index mismatch!" );
		}
		String setName = wallSets[i].getAttribute( "file" );
		if( setName != null )
		{
		    newMap.addWallSet( setName );
		}
	    }
	    Element [] objectSets = XMLUtils.getNamedChildren( setsElement, "objectset" );
	    System.out.println( "Reading object sets" );
	    for( int i=0; i < objectSets.length; i++ )
	    {
		if( i != Integer.parseInt( objectSets[i].getAttribute( "index" ) ) )
		{
		    System.out.println( "Index mismatch!" );
		}
		String setName = objectSets[i].getAttribute( "file" );
		if( setName != null )
		{
		    newMap.addObjectSet( setName );
		}
	    }
	    Element [] doorSets = XMLUtils.getNamedChildren( setsElement, "doorset" );
	    System.out.println( "Reading door sets" );
	    for( int i=0; i < doorSets.length; i++ )
	    {
		if( i != Integer.parseInt( doorSets[i].getAttribute( "index" ) ) )
		{
		    System.out.println( "Index mismatch!" );
		}
		String setName = doorSets[i].getAttribute( "file" );
		if( setName != null )
		{
		    newMap.addDoorSet( setName );
		}
	    }
	    // Load cells
	    Element cellsElement = XMLUtils.getNamedChild( root, "cells" );
	    Element [] cellElements = XMLUtils.getNamedChildren( cellsElement, "cell" );
	    System.out.println( cellElements.length + " cells found" );
	    for( int c=0; c < cellElements.length; c++ )
	    {
		MapCell cell = new MapCell();
		int x = Integer.parseInt( cellElements[c].getAttribute( "x" ) );
		int y = Integer.parseInt( cellElements[c].getAttribute( "y" ) );
		byte light = Byte.parseByte( cellElements[c].getAttribute( "light" ) );
		cell.setLight( light );
		Element [] floorElements = XMLUtils.getNamedChildren( cellElements[c], "floor" );
		if( floorElements != null )
		{
		    //System.out.println( floorElements.length + " floors found for cell: " + x + ", " + y );
		    for( int f=0; f < floorElements.length; f++ )
		    {
			int index = Integer.parseInt( floorElements[f].getAttribute( "index" ) );
			if( index != f )
			{
			    System.out.println( "Index mismatch!" );
			}
			cell.addFloor( Integer.parseInt( floorElements[f].getAttribute( "tile" ) ) );
		    }
		}
		Element wallElement = XMLUtils.getNamedChild( cellElements[c], "wall" );
		if( wallElement != null )
		{
		    cell.setWall( Integer.parseInt( wallElement.getAttribute( "type" ) ) );
		}
		newMap.setCell( x, y, cell );
	    }
	    // Load doors
	    Element doorsElement = XMLUtils.getNamedChild( root, "doors" );
	    if( doorsElement != null )
	    {
		Element [] doorElements = XMLUtils.getNamedChildren( doorsElement, "door" );
		System.out.println( doorElements.length + " doors found" );
		for( int d=0; d < doorElements.length; d++ )
		{
		    int x = Integer.parseInt( doorElements[d].getAttribute( "x" ) );
		    int y = Integer.parseInt( doorElements[d].getAttribute( "y" ) );
		    int direction = Integer.parseInt( doorElements[d].getAttribute( "direction" ) );
		    boolean locked = doorElements[d].getAttribute( "locked" ).trim().equalsIgnoreCase( "true" );
		    boolean open = doorElements[d].getAttribute( "open" ).trim().equalsIgnoreCase( "true" );
		    Door door = new Door( x, y, direction );
		    door.setLocked( locked );
		    door.setOpen( open );
		    newMap.addDoor( door );
		}
	    }
	    // Load objects
	    Element objectsElement = XMLUtils.getNamedChild( root, "objects" );
	    if( objectsElement != null )
	    {
		Element [] objectElements = XMLUtils.getNamedChildren( objectsElement, "object" );
		System.out.println( objectElements.length + " objects found" );
		for( int o=0; o < objectElements.length; o++ )
		{
		    String objectId = objectElements[o].getAttribute( "id" );
		    String name = objectElements[o].getAttribute( "name" );
		    int strength = Integer.parseInt( objectElements[o].getAttribute( "strength" ) );
		    int moveCost = Integer.parseInt( objectElements[o].getAttribute( "movecost" ) );
		    MapObject mapObject = new MapObject( objectId, name, strength, moveCost );
		    /// Load object cells
		    Element [] objectCellElements = XMLUtils.getNamedChildren( objectElements[o], "objectcell" );
		    if( objectCellElements != null )
		    {
			for( int c=0; c < objectCellElements.length; c++ )
			{
			    int x = Integer.parseInt( objectCellElements[c].getAttribute( "x" ) );
			    int y = Integer.parseInt( objectCellElements[c].getAttribute( "y" ) );
			    int type = Integer.parseInt( objectCellElements[c].getAttribute( "type" ) );
			    int destroyedType = Integer.parseInt( objectCellElements[c].getAttribute( "destroyedtype" ) );
			    int hitPoints = Integer.parseInt( objectCellElements[c].getAttribute( "hitpoints" ) );
			    boolean blocksMovement = objectCellElements[c].getAttribute( "blocksmovement" ).trim().equalsIgnoreCase( "true" );
			    boolean blocksSight = objectCellElements[c].getAttribute( "blocksvision" ).trim().equalsIgnoreCase( "true" );
			    MapObjectCell objectCell = new MapObjectCell( x, y, type, hitPoints, blocksMovement, blocksSight );
			    objectCell.setDestroyedType( destroyedType );
			    mapObject.addObjectCell( objectCell );
			}
		    }
		    //System.out.println( "Loaded object " + name + " (" + objectId + "), " + objectCellElements.length + " components" );
		    newMap.addObject( mapObject );
		}
	    }
	    // Load rooms
	    Element roomsElement = XMLUtils.getNamedChild( root, "rooms" );
	    if( roomsElement != null )
	    {
		Element [] roomElements = XMLUtils.getNamedChildren( roomsElement, "room" );
		for( int r=0; r < roomElements.length; r++ )
		{
		    String roomId = roomElements[r].getAttribute( "id" );
		    String roomName = roomElements[r].getAttribute( "name" );
		    int minX = Integer.parseInt( roomElements[r].getAttribute( "minx" ) );
		    int maxX = Integer.parseInt( roomElements[r].getAttribute( "maxx" ) );
		    int minY = Integer.parseInt( roomElements[r].getAttribute( "miny" ) );
		    int maxY = Integer.parseInt( roomElements[r].getAttribute( "maxy" ) );
		    int hotSpotX = Integer.parseInt( roomElements[r].getAttribute( "hotspotx" ) );
		    int hotSpotY = Integer.parseInt( roomElements[r].getAttribute( "hotspoty" ) );
		    Room room = new Room( roomId, roomName, minX, minY, maxX, maxY );
		    room.setHotSpot( new Point( hotSpotX, hotSpotY ) );
		    Element [] linkElements = XMLUtils.getNamedChildren( roomElements[r], "link" );
		    if( linkElements != null )
		    {
			for( int l=0; l < linkElements.length; l++ )
			{
			    room.addLink( linkElements[l].getAttribute( "target" ) );
			}
		    }		    
		    newMap.addRoom( room );
		}
	    }
	    System.out.println( "Loaded map" );
	}
	return newMap;
	
    }
}
