// MissionMapPanel.java
// Started on April 12th, 2002 as GamePanel
// Copyright (c) Virtu Halttunen

import TheGame.Sprite;
import TheGame.CharacterSprite;
import TheGame.MapCell;
import TheGame.MapObject;
import TheGame.MapObjectCell;
import TheGame.Room;
import TheGame.Door;
import TheGame.GameAction;
import java.awt.*;
import java.awt.image.*;
import javax.swing.*;
import java.util.*;
import java.applet.*;
import java.net.*;

public class MissionMapPanel extends JPanel implements Scrollable
{
    public final static int LIGHT_LEVELS = 32;
    public static boolean playSounds = true;
    public static boolean drawShootingHints = true;
    private boolean alwaysShowMonsters = false;
    private boolean drawPlayerHint = true;
    private String resourcePath = null;
    private ClassLoader imageLoader = null;

    private MissionEditor editor;
    private EditorMap map;
    private Dimension viewportSize;
    private int cellSize;
    private Vector floorCells;
    private Vector wallCells;
    private Image [][] doorCells;
    private Vector objectCells;
    private Color [] lights;
    private Color highlightColor = Color.green;
    private Color hasViewColor = Color.yellow;
    private Color noViewColor = Color.red;
    private HashMap characterSprites;
    private HashMap weaponSprites;
    private Vector stainSprites;
    private Vector bufferedMoveActions;
    private Vector bufferedAttackActions; // includes player deaths
    private Vector bufferedDoorActions;
    private int frame = 0;
    private int openDoorIndex = 0;
    private int closedDoorIndex = 0;
    private int doorFrame = -1;
    private AudioClip doorClip;
    private AudioClip boom;
    private AudioClip miss;
    private Point targetReticleLocation = new Point( -1, -1 );
    private ProgressMonitor monitor;
    private int weaponFrame = -1;
    
    public MissionMapPanel( MissionEditor editor, int cellSize, int visibleHorizontal, int visibleVertical )
    {
	super();
	map = editor.getMap();
	this.editor = editor;
        URL fileURL = null;
        try
        {
            fileURL = new URL( MissionEditor.RESOURCE_PATH );
        }
        catch( MalformedURLException mfue )
        {}
        URL [] urls = new URL[1];
        urls[0] = fileURL;
        imageLoader = new URLClassLoader( urls );        
	this.cellSize = cellSize;
        monitor = new ProgressMonitor( this, "Loading...", "Loading sprites", 0, 10 ); // Progress
        int progress = 0;
	viewportSize = new Dimension( visibleHorizontal*cellSize, visibleVertical*cellSize );
	Dimension mySize = new Dimension ( map.getHorizontalSize()*cellSize, map.getVerticalSize()*cellSize );
	setMinimumSize( mySize );
	setPreferredSize( mySize );
	setMaximumSize( mySize );
	floorCells = new Vector();
	wallCells = new Vector();
	objectCells = new Vector();
        bufferedMoveActions = new Vector();
        bufferedAttackActions = new Vector();
        bufferedDoorActions = new Vector();
	CharacterReference character;
	String characterClass;
	CharacterSprite newSprite;
        characterSprites = new HashMap( editor.getMonsters().size()*3, 0.75f );
        weaponSprites = new HashMap();
        monitor.setMaximum( monitor.getMaximum() + editor.getPlayerCharacters().size() + editor.getMonsters().size() );
	for( int i = 0; i < editor.getPlayerCharacters().size(); i++ )
	{
            character = editor.getPlayerCharacter( i );
            addSpriteForCharacter( character );
            monitor.setProgress( ++progress );
	}
        for( int i = 0; i < editor.getMonsters().size(); i++ )
	{
            character = editor.getMonster( i );
            addSpriteForCharacter( (CharacterReference)character );
            monitor.setProgress( ++progress );
        }
	stainSprites = new Vector();
        monitor.setNote( "Loading map... " );
	loadFloorsFromMap();
        monitor.setProgress( progress+=2 );
	loadWallsFromMap();
        monitor.setProgress( progress+=2 );
	loadDoorsFromMap();
        monitor.setProgress( progress+=2 );
	loadObjectsFromMap();
        monitor.setProgress( progress+=2 );
	lights = new Color[ LIGHT_LEVELS ];
	for( int i = 0; i < LIGHT_LEVELS; i++ )
	{
	    lights[ i ] = new Color( 1, 1, 1, i*(255/LIGHT_LEVELS));
	}
        monitor.setProgress( progress+=2 );
        monitor.close();
	//setCursor( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
    }
    
    protected void paintComponent( Graphics g )
    {
	//super.paintComponent( g );
        long startTime = System.currentTimeMillis();
        MapCell cell;
	CharacterSprite sprite;
	Sprite stain;
        Sprite objective;
	GameAction action;
        Point door;
	//System.out.println( "MapPanel: repaint" );
	Rectangle newRect = g.getClipBounds();
	Point min = getCell( new Point( newRect.x, newRect.y ) );
	int minX = min.x;
	int minY = min.y;
	Point max = getCell( new Point( newRect.x+newRect.width, newRect.y+newRect.height ) );
	int maxX = max.x;
	int maxY = max.y;
	
	if( maxX == map.getHorizontalSize() ) maxX = maxX-1;
	if( maxY == map.getVerticalSize() ) maxY = maxY-1;

        int numCells = (maxX-minX)*(maxY-minY);
        
	for( int i = minX; i <= maxX; i++ )
	{
	    for( int j = minY; j <= maxY; j++ )
	    {
		g.setColor( Color.black );
		g.fillRect( i*cellSize, j*cellSize, cellSize, cellSize );

                cell = map.getCell( i, j );
                if( cell != null )
                {
		    g.setColor( Color.black );
		    g.fillRect( i*cellSize, j*cellSize, cellSize, cellSize );

                    int floor;
                    for( int f = 0; f < cell.getFloorSize(); f++ )
                    {
                        floor = cell.getFloor( f );
                        if( floor >= 0 )
                            g.drawImage( (Image)floorCells.elementAt( floor ), i*cellSize, j*cellSize, cellSize, cellSize, this );
                    }
                }
	    }
	}
	// Draw stains
	for( int i = 0; i < stainSprites.size(); i++ )
	{
	    stain = (Sprite)stainSprites.elementAt( i );
	    if( newRect.intersects( stain.getBounds() ) )
	    {
                g.drawImage( stain.getSprite(), stain.getLocation().x, stain.getLocation().y, this );
	    }
	}
    
	// Draw objects
	for( int x=minX; x <=maxX; x++ )
	{
	    for( int y=minY; y <= maxY; y++ )
	    {
		for( int i=0; i < map.getNumberOfObjects(); i++ )
		{
		    MapObject mapObject = map.getObject( i );
		    if( mapObject.isLocatedIn( x, y ) )
		    {
			MapObjectCell[] mapObjectCells = mapObject.getCells( x, y );
			if( mapObjectCells != null )
			{
			    for( int j=0; j < mapObjectCells.length; j++ )
			    {
				if( mapObjectCells[j].isLocatedIn( x, y ) )
				{
				    int objIndex = mapObjectCells[j].getType();
				    if( mapObjectCells[j].isDestroyed() )
				    {
					objIndex = mapObjectCells[j].getDestroyedType();
					if( objIndex < 0 )
					{
					    objIndex = mapObjectCells[j].getType();
					}
				    }
				    g.drawImage( (Image)objectCells.elementAt( objIndex ), x*cellSize, y*cellSize, cellSize, cellSize, this );
				}
			    }
			}
		    }
		}
	    }
	}
	// Draw objectives
	for( int i = minX; i <= maxX; i++ )
	{
	    for( int j = minY; j <= maxY; j++ )
	    {
                for( int k=0; k < editor.getNumberOfObjectives(); k++ )
                {
                    if( editor.getObjective(k).getNumberOfSprites() > 0 )
                    {
                        if( editor.getObjective(k).isLocatedIn( i, j ) )
                        {
                            objective = editor.getObjective(k).getSprite( i, j );
                            if( objective != null )
                            {
                                g.drawImage( objective.getSprite(), i*cellSize+objective.getXOffset(), j*cellSize+objective.getYOffset(), this );
                            }
                        }
                    }
                }
		if( editor.getSecondaryObjective() != null )
		{
		    if( editor.getSecondaryObjective().isLocatedIn( i, j ) )
		    {
			objective = editor.getSecondaryObjective().getSprite( i, j );
			if( objective != null )
			{
			    g.drawImage( objective.getSprite(), i*cellSize+objective.getXOffset(), j*cellSize+objective.getYOffset(), this );
			}
		    }
		}
		if( editor.getPrimaryObjective() != null )
		{
		    if( editor.getPrimaryObjective().isLocatedIn( i, j ) )
		    {
			objective = editor.getPrimaryObjective().getSprite( i, j );
			if( objective != null )
			{
			    g.drawImage( objective.getSprite(), i*cellSize+objective.getXOffset(), j*cellSize+objective.getYOffset(), this );
			}
		    }
		}
            }
        }
	// Draw monster sprites
	for( int i=0; i < editor.getMonsters().size(); i++ )
	{
	    sprite = (CharacterSprite)characterSprites.get( editor.getMonster( i ) );
            g.setColor( noViewColor );
            Point point = editor.getMonster( i ).getLocation();
            g.drawRect( point.x*cellSize+1, point.y*cellSize+1, cellSize-3, cellSize-3 );
            g.drawRect( point.x*cellSize+2, point.y*cellSize+2, cellSize-5, cellSize-5 );
            if( sprite != null )
            {
                action = sprite.getCharacter().getLastAction();
                if( newRect.intersects( sprite.getBounds() ) )
                {
                    if( sprite.getCharacter().isAlive() )
                    {
                        Point loc = sprite.getCharacter().getLocation();
                        sprite.setLocation( loc.x*cellSize, loc.y*cellSize );
                        sprite.setDirection( sprite.getCharacter().getDirection() );
                        g.drawImage( sprite.getSprite( action ), sprite.getLocation().x, sprite.getLocation().y, this );
                    }
                }
            }
	}
	// Draw player sprites
	for( int i=0; i < editor.getPlayerCharacters().size(); i++ )
	{
	    sprite = (CharacterSprite)characterSprites.get( editor.getPlayerCharacter( i ) );;
	    action = sprite.getCharacter().getLastAction();
	    if( newRect.intersects( sprite.getBounds() ) )
	    {
                if( sprite.getCharacter().isAlive() )
                {
                    Point loc = sprite.getCharacter().getLocation();
                    sprite.setLocation( loc.x*cellSize, loc.y*cellSize );
                    sprite.setDirection( sprite.getCharacter().getDirection() );
                    g.drawImage( sprite.getSprite( action ), sprite.getLocation().x, sprite.getLocation().y, this );
                }
	    }
	}
        // Draw walls
        for( int i = minX; i <= maxX; i++ )
            for( int j = minY; j <= maxY; j++ )
            {
                cell = map.getCell( i, j );
                if( cell != null )
                {
                    int wall = cell.getWall();
                    if( wall >= 0 )
                    {
                        g.drawImage( (Image)wallCells.elementAt( wall ), i*cellSize, j*cellSize, cellSize, cellSize, this );
                    }
                    else
                    {
                        // If it's a door
                        if( wall < -1 )
                        {
                            drawDoor( g, i, j );
                        }
                    }
                }
            }
	// Draw lights
	for( int i = minX; i <= maxX; i++ )
	    for( int j = minY; j <= maxY; j++ )
	    {
		g.setColor( lights[ map.getLight( i, j ) ] );
		g.fillRect( i*cellSize, j*cellSize, cellSize, cellSize );
	    }
        // Draw player starting points
	g.setColor( highlightColor );
	Point point;
	for( int i=0; i < editor.getNumberOfPlayerStartPoints(); i++ )
	{
	    point = editor.getPlayerCharacter( i ).getLocation();
	    g.drawRect( point.x*cellSize+1, point.y*cellSize+1, cellSize-3, cellSize-3 );
	    g.drawRect( point.x*cellSize+2, point.y*cellSize+2, cellSize-5, cellSize-5 );
	}

        long endTime = System.currentTimeMillis();
        //System.out.println( "Drawing " + numCells + " cells took " + (int)(endTime-startTime) + " ms" );
    }

    private void drawDoor( Graphics g, int x, int y )
    {
        Door theDoor = map.getDoor( x, y );
        Point door1;
        Point door2;
        int doorAnimationFrame = closedDoorIndex;

        if( theDoor != null )
        {
            int dir = theDoor.getDirection();
            int dIndex = 0;
            if( dir == EditorMap.NORTH ) dIndex = 1;
            if( dir == EditorMap.SOUTH ) dIndex = 0;
            if( dir == EditorMap.EAST ) dIndex = 2;
            if( dir == EditorMap.WEST ) dIndex = 3;
            
            // Draw the door
            if( theDoor.isOpen() )
                doorAnimationFrame = openDoorIndex;
            else
                doorAnimationFrame = closedDoorIndex;

            g.drawImage( doorCells[ dIndex ][ doorAnimationFrame ], x*cellSize, y*cellSize, cellSize, cellSize, this );
        }
    }
    
    public Point getCell( Point point )
    {
	int x=0;
	int y=0;
	
	x = point.x/cellSize;
	y = point.y/cellSize;
	if( x < 0 ) x = 0;
	if( x >= map.getHorizontalSize() ) x = map.getHorizontalSize()-1;
	if( y < 0 ) y = 0;
	if( y >= map.getVerticalSize() ) y = map.getVerticalSize()-1;
	
	return new Point( x, y );
    }

    public int getCellSize()
    {
        return cellSize;
    }
            
    public void addSpriteForCharacter( CharacterReference character )
    {
        String characterClass;
        CharacterSprite newSprite;
        
        try
        {
            characterClass = "TheGame."+character.getClassName();
            System.out.println( "Adding a sprite for " + characterClass );
	    if( characterClass.endsWith( MissionEditor.GENERIC_MONSTER_KEY ) )
	    {
		System.out.println( "Generic monster" );
		// Create a dummy monster, so the stats are certainly loaded
		String type = characterClass.substring( 8, characterClass.indexOf( "(" ) );
		TheGame.GenericMonster.create( type );
		newSprite = new TheGame.GenericMonsterSprite();
		((TheGame.GenericMonsterSprite)newSprite).setType( type );
	    }
	    else
	    {
		newSprite = (CharacterSprite)Class.forName( characterClass+"Sprite", true, imageLoader ).newInstance();
	    }
            newSprite.setCharacter( character );
            newSprite.setLocation( character.getLocation().x*cellSize, character.getLocation().y*cellSize );
            newSprite.setDirection( character.getDirection() );
            characterSprites.put( character, newSprite );
            System.out.println( "Added sprite for: " + characterClass );
        }
        catch( Exception e )
        {
            System.out.println( "Couldn't add sprite for " + character.getClassName() );
            System.out.println( e.toString() );
            characterSprites.put( character, null );
        }

    }

    public void removeSpriteForCharacter( CharacterReference character )
    {
        characterSprites.remove( character );
    }

    public void setDrawPlayerHint( boolean aBool )
    {
	drawPlayerHint = aBool;
    }
    
    public void centerOnCharacter( TheGame.Character character )
    {
        Rectangle visRect = getVisibleRect();
        int x = visRect.width/cellSize;
        int y = visRect.height/cellSize;
        scrollRectToVisible( new Rectangle( (character.getLocation().x-x/2)*cellSize, (character.getLocation().y-y/2)*cellSize, x*cellSize, y*cellSize ) );
    }


    private void loadFloorsFromMap()
    {
	Vector floorVector = map.getFloorSets();
	if( floorVector.size() > 0 )
	{
	    for( int i = 0; i < floorVector.size(); i++ )
	    {
		TheGame.ResourceManager.cropImage( (String)floorVector.elementAt( i ), floorCells, cellSize );
	    }
	}	
    }

    private void loadWallsFromMap()
    {
	Vector wallVector = map.getWallSets();
	if( wallVector.size() > 0 )
	{
	    for( int i = 0; i < wallVector.size(); i++ )
	    {
		TheGame.ResourceManager.cropImage( (String)wallVector.elementAt( i ), wallCells, cellSize );
	    }
	}	
    }

    private void loadDoorsFromMap()
    {
	System.out.println( "Loading door sets" );
        System.out.println( "Resource path: " + resourcePath );
	Vector doorVector = map.getDoorSets();
	Vector doorImages = new Vector();
	if( doorVector.size() > 0 )
	{
	    for( int i = 0; i < doorVector.size(); i++ )
	    {
                TheGame.ResourceManager.cropImage( (String)doorVector.elementAt( i ), doorImages, cellSize );
	    }
	}
	doorCells = new Image[4][doorImages.size()/4];
	System.out.println( doorImages.size()/4 + " door frame(s) found" );
	openDoorIndex = (doorImages.size()/4)-1;
	System.out.println( "Open door is at " + openDoorIndex );
	closedDoorIndex = 0;
	for( int i=0; i < 4; i++ )
	{
	    for( int j=0; j < doorImages.size()/4;j++ )
	    {
		doorCells[i][j] = (Image)doorImages.elementAt( i+j*4 );
	    }
	}	
    }

    private void loadObjectsFromMap()
    {
	Vector objectVector = map.getObjectSets();
	if( objectVector.size() > 0 )
	{
	    for( int i = 0; i < objectVector.size(); i++ )
	    {
		TheGame.ResourceManager.cropImage( (String)objectVector.elementAt( i ), objectCells, cellSize );
	    }
	}	
    }
    
    public void setMap( EditorMap map )
    {
	this.map = map;
	Dimension mySize = new Dimension ( map.getHorizontalSize()*cellSize, map.getVerticalSize()*cellSize );
        characterSprites = new HashMap();
	setMinimumSize( mySize );
	setPreferredSize( mySize );
	setMaximumSize( mySize );
	floorCells = new Vector();
	wallCells = new Vector();
	objectCells = new Vector();
	loadFloorsFromMap();
	loadWallsFromMap();
	loadDoorsFromMap();
	loadObjectsFromMap();
	repaint();
    }
    
    public Vector getFloorCells()
    {
	return floorCells;
    }
    
    public Vector getWallCells()
    {
	return wallCells;
    }
    
    public Vector getObjectCells()
    {
	return objectCells;
    }
    
/*    public Vector getDoorCells()
    {
	return doorCells;
    }*/
    
    public Dimension getPreferredScrollableViewportSize()
    {
	return viewportSize;
    }
    
    public int getScrollableUnitIncrement( Rectangle visibleRect, int orientation, int direction ) 
    {
	//get the current position
	int currentPosition = 0;
	if( orientation == SwingConstants.HORIZONTAL )
	    currentPosition = visibleRect.x;
	else
	    currentPosition = visibleRect.y;
    
	//return the number of pixels between currentPosition
	//and the nearest tick mark in the indicated direction
	if ( direction < 0 ) 
	{
	    int newPosition = currentPosition - (currentPosition / cellSize) * cellSize;
	    return (newPosition == 0) ? cellSize : newPosition;
	} 
	else 
	{
	    return ((currentPosition / cellSize) + 1) * cellSize - currentPosition;
	}
    }
    
    public int getScrollableBlockIncrement( Rectangle visibleRect, int orientation, int direction ) 
    {
	int amount = 0;
	if( orientation == SwingConstants.HORIZONTAL )
	{
	    amount = getScrollableUnitIncrement( visibleRect, orientation, direction );
	    if( amount < cellSize )
		amount = amount+cellSize;
	    return visibleRect.width - amount;
	}
	else
	{
	    amount = getScrollableUnitIncrement( visibleRect, orientation, direction );
	    if( amount < cellSize )
		amount = amount+cellSize;
	    return visibleRect.height - amount;
	}
    }
    
    public boolean getScrollableTracksViewportWidth()
    {
	return false;
    }
    
    public boolean getScrollableTracksViewportHeight()
    {
	return false;
    }

    
}