// BattleRobot.java
// Started on September 7th, 2003
// Copyright (c) Virtu Halttunen

package TheGame;

import java.awt.Point;
import java.util.Vector;

public class BattleRobot extends Monster implements Armed, Robot
{
    public static final int ROAMING = 0;
    public static final int SEEKING = 1;
    public static final int ATTACKING = 2;
    public static final int SHOOTING = 3;
    public static final int WAITING = 4;

    public static final double REACTIVATION_PROBABILITY = 0.075;

    private Character target = null;
    private boolean hasMoreActions = false;
    private int state = ROAMING;
    private Weapon weapon;
    private int skill = 4;
    private Point start;
    private Point end;
    private boolean hasTriedReactivation = false;
    private boolean hasReactivated = false;
    private boolean disabled = false;
    private int hits = 0;

    public BattleRobot()
    {
        location = new Point( 0, 0 );
        direction = Map.NORTH;
        alive = true;
        lastAction = new MoveAction( this, location, location );
        maximumActionPoints = 5;
        currentActionPoints = 5;
        maximumHitPoints = 5;
        currentHitPoints = 5;
        weapon = new SubmachineGun();
    }

    public void setSpottedPlayer( Character character, boolean canSee )
    {
	setDormant( false );
        if( canSee )
        {
            target = character;
            state = SEEKING;
            //System.out.println( "BattleRobot spotted player: " + index );
        }
    }

    public void setState( int state )
    {
        this.state = state;
    }

    public Weapon getWeapon()
    {
        return weapon;
    }

    public void setWeapon( Weapon weapon )
    {}    
    
    public GameAction playAction()
    {
        int x;
        int y;

        Point cell;
        Vector enemies = new Vector();
        Character check;

        hasMoreActions = true;

	if( alive && !disabled )
        {
            for( int i=0; i < activeGame.getPlayerCharacters().size(); i++ )
            {
                check = activeGame.getPlayerCharacter( i );
                if( activeGame.getMap().hasLineOfSight( check.getLocation(), location ) )
                    enemies.add( check );
            }
            //System.out.println( "BattleRobot: playAction" );
            if( target != null )
                if( !target.isAlive() )
                {
                    target = null;
                    state = ROAMING;
                }
                else
                {
                    state = SEEKING;
                }
            if( currentActionPoints <= 0 )
            {
                lastAction = new WaitAction( this, new Point( location ) );
                return lastAction;
            }
            // If next to an enemy, attack
            //System.out.println( "BattleRobot checking neighbours" );
            for( int i=0; i < enemies.size(); i++ )
            {
                check = (Character)enemies.elementAt(i);
                if( check.isAlive() )
                {
                    cell = check.getLocation();
                    if( activeGame.getMap().isNeighbor( location, cell ) )
                    {
                        if( currentActionPoints >= 2 )
                        {
                            state = ATTACKING;
                            direction = Map.getDirection( location, cell, direction );
                            lastAction = new MeleeAction( this, check );
                            return lastAction;
                        }
                        else
                        {
                            state = WAITING;
                            lastAction = new WaitAction( this, new Point( location ) );
                            hasMoreActions = false;
                            return lastAction;
                        }
                    }
                }
            }
            // If not next to an enemy, but can see one, shoot it
            if( enemies.size() > 0 )
            {
                target = activeGame.getClosestCharacter( this, enemies );
                if( activeGame.getMap().getLineOfFire( location, target.getLocation(), weapon.getBlockingSize() ).equals( target.getLocation() ) && ( location.distance( target.getLocation() ) <= 16 ) && currentActionPoints > 2 )
                {
                    state = SHOOTING;
                    currentActionPoints--; // Let's decrease this here, so the robot will shoot less
                    direction = Map.getDirection( location, target.getLocation(), direction );
                    lastAction = new ShootAction( this, weapon, target.getLocation(), true );
                    return lastAction;
                }
                else
                {
                    state = SEEKING;
                }
            }
            if( state == SEEKING )
            {
                //System.out.println( "BattleRobot seeking targets" );
                if( target != null )
                {
                    GameAction action = SeekerAI.seek( this, target.getLocation() );
                    if( action != null )
                    {
                        lastAction = action;
                        return lastAction;
                    }
                    else
                    {
                        state = ROAMING;
                        target = null;
                    }
                }
            }
            if( state == ROAMING )
            {
                // Walks between start and end
                if( location.equals( end ) )
                {
                    //System.out.println( "Battle robot reached end at " + Objective.pointToString( end ) + ", going to " + Objective.pointToString( start ) );
                    end = start;
                    start = location;
                }
                GameAction action = SeekerAI.seek( this, end );
                if( action != null )
                {
                    lastAction = action;
                    return lastAction;
                }
                else
                {
                    //System.out.println( "BattleRobot doesn't know what to do!" );
                    lastAction = new MoveAction( this, location, location );
                }
            }
        }
        return lastAction;
    }
	
    public void preAction()
    {
	if( !alive && !disabled && !hasReactivated )
        {
            // May reactivate
            if( !hasTriedReactivation && ( Math.random() < REACTIVATION_PROBABILITY ) )
            {
                // Each character instance can only reactivate once. We will however
                // spawn a new monster instance for the reactivated version.
                hasReactivated = true;
                
                BattleRobot battleRobot = new BattleRobot();
                battleRobot.setGame( activeGame );
                battleRobot.setLocation( location.x, location.y );
                int healthLevel = maximumHitPoints;
                if( Math.random() < 0.75 )
                {
                    healthLevel = healthLevel-1;
                }
                battleRobot.setReactivated( healthLevel );
                activeGame.spawnMonster( battleRobot );
                
                System.out.println( "Battle robot reactivated!" );
            }
            else
            {
                hasTriedReactivation = true;
                hasMoreActions = false;
            }
        }
	
    }

        public GameAction getLastAction()
        {
            //System.out.println( "Get action from BattleRobot " +index );
            return lastAction;
        }

        public void resetActionPoints()
        {
            //System.out.println( "BattleRobot reset action points" );
            currentActionPoints = maximumActionPoints;
            hasTriedReactivation = false;
        }

        public boolean hasMoreActions()
        {
            if( currentActionPoints > 0 && hasMoreActions )
            {
                //System.out.println( "BattleRobot has " + currentActionPoints + " action points left, attacking: " + attacking +", seeking: " + seeking );
                return true;
            }
            return false;
        }

	public int getMeleeAbility()
	{
	    return 3;   
	}
	
	public int getMeleeCost()
	{
	    return 2;
	}

        public int getWeaponSkill()
        {
            return skill;
        }
	
        public boolean isAlive()
        {
            return alive;
        }

        public boolean hasReactivated()
        {
            return hasReactivated;
        }

        public void setReactivated( int hitPoints )
        {
            if( hitPoints > 0 )
            {
                currentHitPoints = hitPoints;
                maximumHitPoints = hitPoints;
                currentActionPoints = hitPoints-1;
                maximumActionPoints = hitPoints-1;
            }
            if( maximumActionPoints < 3 )
            {
                currentActionPoints = 3;
                maximumActionPoints = 3;
            }
        }

        public void setGame( Game game )
        {
            activeGame = game;
            start = activeGame.getRandomPointForMonster();
            end = activeGame.getRandomPointForMonster();
        }

        public void setDisabled( boolean disabled )
        {
            this.disabled = disabled;
            if( disabled )
            {
                alive = false;
                lastAction = new DeadAction( this, new Point( location ) );
            }
        }

        public boolean getDisabled()
        {
            return disabled;
        }

        public boolean applyDamage( int damage )
        {
            damage = damage-2;
            if( damage >= 100 )
            {
                maximumActionPoints = 0;
                currentActionPoints = 0;
                setDisabled( true );
                alive = false;
                return false;
            }
            if( damage < 0 )
            {
                damage = 0;
            }
            currentHitPoints = currentHitPoints-damage;
            lastAction = new WoundAction( this, location );
            if( currentHitPoints <= 0 )
            {
                alive = false;
                lastAction = new DeadAction( this, new Point( location ) );
                maximumActionPoints = 0;
                currentActionPoints = 0;
            }
            return alive;
        }
	
	public void addKill()
	{}
	
	public void registerHit()
	{
	    hits++;
	    if( hits >= 10 )
	    {
		hits-=10;
		skill++;
	    }
	}

        public int getSize()
        {
            return 75;
        }
}