// Character.java
// Started on June 20th, 2001
// Copyright (c) Virtu Halttunen

package TheGame;

/**
 * This class is the abstract base for all character classes in the game.
 *
 * @author	Virtu Halttunen
 * @version	0.1
 */
import java.awt.Point;

public abstract class Character
{
    protected Game activeGame;
    protected Point location;
    protected int direction;
    protected GameAction lastAction;
    protected boolean alive = true;
    protected int maximumActionPoints;
    protected int currentActionPoints;
    protected int maximumHitPoints;
    protected int currentHitPoints;
    protected String name = null;
    protected String id = null;

    public Point getLocation()
    {
        return location;
    }

    public void setLocation( Point location )
    {
        this.location = new Point( location );
        lastAction = new WaitAction( this, this.location );
    }

    public void setLocation( int x, int y )
    {
	location.x = x;
	location.y = y;
        lastAction = new WaitAction( this, location );
    }
    
    public int getDirection()
    {
	return direction;
    }
    
    public void setDirection( int direction )
    {
	this.direction = direction;
    }
    
    public void setGame( Game game )
    {
	activeGame = game;
    }

    /**
     *
     *  This method is used by the Game to save and load the monsters.
     *
     *  @return the classname of the monster, without package names
     */
    public String getClassName()
    {
        String className = getClass().getName();
        if( className.startsWith( "TheGame." ) )
        {
            className = className.substring( 8 );
        }
        return className;
    }
    
    /**
     *
     *  This method is the recommended way of recognizing friends and foes.
     *  It can also be used to display the human-readable monster type name.
     *
     *  @return the type of the monster
     */
    public String getType()
    {
        return getClassName();
    }
    
    public Game getGame()
    {
        return activeGame;
    }
    
    public String getName()
    {
        return name;
    }
    
    public void setName( String name )
    {
        this.name = name;
    }
    
    public GameAction doAction( GameAction action )
    {
        return action;
    }
        
    public GameAction getLastAction()
    {
	return lastAction;
    }
    
    /**
     * Damages the character
     *
     * @param damage	the amount of damage inflicted on character
     * @return 		<code>true</code> if the character is still alive
     */     
    public boolean applyDamage( int damage )
    {
	return true;
    }

    public void healDamage( int hitPoints )
    {
        currentHitPoints+=hitPoints;
        if( currentHitPoints > maximumHitPoints )
        {
            currentHitPoints = maximumHitPoints;
        }
    }
    
    public int getMaximumHitPoints()
    {
	return maximumHitPoints;
    }
    
    public int getCurrentHitPoints()
    {
	return currentHitPoints;
    }
    
    /**
     * Returns the melee ability of the character. This represents the character's
     * skill in close combat. Higher is deadlier. Negative number means the character
     * is unable to defend itself at all
     *
     * @return 		melee ability of the character, usually in range of 0 to 10
     */
    public int getMeleeAbility()
    {
        return 0;
    }
    
    /**
     * How many action points the character's melee attack costs
     *
     * @return 		cost of melee attack
     */
    public int getMeleeCost()
    {
        return 1;
    }

    /**
     * Checks whether this character hates the other character.
     * The feeling doesn't have to ne mutual, so this
     * character may hate the other even if the opposite isn't true.
     *
     * @param character	the character to be checked
     * @return 		<code>true</code> if this character hates the other
     */
    public boolean isFoe( Character character )
    {
        return false;
    }

    /**
     * Checks whether this and the other character hate each other.
     * The feeling is mutual, if one character hates the other, both
     * are enemies to each other.
     *
     * @param character	the character to be checked
     * @return 		<code>true</code> if the characters hate each other
     */
    public boolean isHostile( Character character )
    {      
        if( character == this )
        {
            return false;
        }
        return ( isFoe( character ) || character.isFoe( this ) );
    }

    public boolean isShownOnMap()
    {
        return true;
    }
    
    public boolean isAlive()
    {
	return alive;
    }
    
    public boolean leavesCorpse()
    {
//        System.out.println( this.getClass().getName() + " leaves a corpse" );
        return true;
    }
    
    public int getMaximumActionPoints()
    {
	return maximumActionPoints;
    }

    public void setMaximumActionPoints( int maximumActionPoints )
    {
        this.maximumActionPoints = maximumActionPoints;
    }
    
    public int getCurrentActionPoints()
    {
	return currentActionPoints;
    }

    public void resetActionPoints()
    {
        currentActionPoints = maximumActionPoints;
    }

    public String getId()
    {
        return id;
    }
    
    public void setId( String id )
    {
        this.id = id;
    }
    
    public void resetHitPoints()
    {
        currentHitPoints = maximumHitPoints;
    }

    /**
     * Returns the size of the character. The size varies from 0 to 100
     * Two or more characters can inhabit the same square only if the sum
     * of their respective sizes is less than 100. Size also affects ranged
     * weapon attacks. If the character's size is greater than the weapon's
     * minBlockingSize, the character blocks the line of fire for that weapon.
     *
     * @return 		the size of the character
     */
    public int getSize()
    {
	return 50;
    }
}