// Game.java
// Started on June 20th, 2001
// Copyright (c) Virtu Halttunen

package TheGame;

import java.util.Vector;
import java.util.Observable;
import java.awt.Point;
import java.io.*;

/**
 * This class handles the game logic. A new instance is created for
 * every individual game.
 *
 * @author	Virtu Halttunen
 * @version	0.1
 */
public class Game extends Observable
{    
    public static int MOVE = 0;
    public static int ATTACK_RANGED = 1;
    public static int ATTACK_MELEE = 2;
    public static int DIE = 3;
    public static int DEAD = 4;
    public static int USE = 5;
    
    public static int PLAYER_MOVE = 0;
    public static int MONSTER_MOVE = 1;
    public static final int PLAYER = 0;
    public static final int MONSTER = 1;
    public static final int OBJECTIVE = 2;
    public static final int OBJECT = 3;

    //private Mission mission;
    private Map map;
    private Vector playerCharacters;
    private Vector monsters;
    private Vector playerStartingPoints;
    private Objective primaryObjective = null;
    private Objective secondaryObjective = null;
    private boolean endsAtBase = false;
    private Vector objectives;

    private int state = PLAYER_MOVE;
    private int activeCharacter = 0;
    private int maximumMonsterActions = 0;
    private boolean hasMoreMonsterActions = false;
    private Vector attackActions = new Vector();
    private int turns = 0;
    private boolean failed = false;
    private String missionName;
    private String longMissionDescription;
    private String [] debriefMessages = { "Complete victory", "Victory", "Failure", "Complete failure" };
    private int [] missionBonusPoints = { 9, 6, 4, 2 };
    private String [] nextMissionFileNames = { null, null, null, null };
    private boolean [] canChangeMissionTeamAfter = { true, true, true, true };
        
    public Game( Mission mission )
    {
	//this.mission = mission;
	this.map = mission.getMap();
	map.setGame( this );
	int x, y;
	playerCharacters = new Vector();
	setPlayerStartingPoints( mission.getPlayerStartPoints() );
	setMonsters( mission.getMonsters() );
	primaryObjective = mission.getPrimaryObjective();
	secondaryObjective = mission.getSecondaryObjective();
        setObjectives( mission.getObjectives() );
	endsAtBase = mission.getEndsAtBase();
	canChangeMissionTeamAfter = mission.getCanChangeMissionTeamAfter();
	missionName = mission.getName();
	longMissionDescription = mission.getDescription();
	debriefMessages = mission.getDebriefMessages();
	missionBonusPoints = mission.getMissionBonusPoints();
	nextMissionFileNames = mission.getNextMissionFileNames();
    }

    public boolean canCharacterMove( Character theCharacter, Point point )
    {
        return canCharacterMove( theCharacter, point, false, false );
    }

    public boolean canCharacterMove( Character theCharacter, Point point, boolean canOpenDoors )
    {
        return canCharacterMove( theCharacter, point, false, canOpenDoors );
    }

    public boolean canCharacterMove( Character theCharacter, int direction, boolean canOpenDoors )
    {
        return canCharacterMove( theCharacter, direction, false, canOpenDoors );
    }

    public boolean canCharacterMove( Character theCharacter, int direction, boolean ignoreObjects, boolean canOpenDoors )
    {
        Point point = map.getNeighbor( theCharacter.getLocation(), direction );
        return canCharacterMove( theCharacter, point, ignoreObjects, canOpenDoors );
    }

    public boolean canCharacterMove( Character theCharacter, Point point, boolean ignoreObjects, boolean canOpenDoors )
    {
        if( !theCharacter.isAlive() )
        {
            return false;
        }
        if( map.isPassable( point, canOpenDoors ) && ( theCharacter.getCurrentActionPoints() > 0 ) )
        {
            boolean blocked = false;
            Character character;
            MapObject mapObject;
            int minBlockingSize = 100-theCharacter.getSize();
	    Character [] characters = getVisibleCharacters( theCharacter );
            
            for( int i=0; i< characters.length; i++ )
            {
                character = characters[i];
                if( character != theCharacter && character.getLocation().equals( point ) && character.getSize() > minBlockingSize && character.isAlive() )
                {
                    blocked = true;
                    break;
                }
            }
            for( int i=0; i< getNumberOfObjectives(); i++ )
            {
                if( ( getObjective(i).getSize() > minBlockingSize ) && getObjective(i).isLocatedIn( point ) )
                {
                    blocked = true;
                    break;
                }
            }
            if( map.getCell( point ).hasObjects() && !ignoreObjects )
            {
                for( int i = 0; i < map.getNumberOfObjects(); i++ )
                {
                    mapObject = map.getObject( i );
                    if( mapObject.isLocatedIn( point ) && mapObject.isBlockingMovement( point ) )
                    {
                        blocked = true;
                        break;
                    }
                }
            }
            return !blocked;
        }
        return false;
    }
    
    public boolean canCharacterMove( Character theCharacter, Point point, Character [] blockingCharacters, boolean ignoreObjects, boolean canOpenDoors )
    {
        if( !theCharacter.isAlive() )
        {
            return false;
        }
        if( map.isPassable( point, canOpenDoors ) && ( theCharacter.getCurrentActionPoints() > 0 ) )
        {
            boolean blocked = false;
            Character character;
            MapObject mapObject;
            int minBlockingSize = 100-theCharacter.getSize();
            
            for( int i=0; i< blockingCharacters.length; i++ )
            {
                character = blockingCharacters[i];
                if( character != theCharacter && character.getLocation().equals( point ) && character.getSize() > minBlockingSize && character.isAlive() )
                {
                    blocked = true;
                    break;
                }
            }
            for( int i=0; i< getNumberOfObjectives(); i++ )
            {
                if( ( getObjective(i).getSize() > minBlockingSize ) && getObjective(i).isLocatedIn( point ) )
                {
                    blocked = true;
                    break;
                }
            }
            if( map.getCell( point ).hasObjects() && !ignoreObjects )
            {
                for( int i = 0; i < map.getNumberOfObjects(); i++ )
                {
                    mapObject = map.getObject( i );
                    if( mapObject.isLocatedIn( point ) && mapObject.isBlockingMovement( point ) )
                    {
                        blocked = true;
                        break;
                    }
                }
            }
            return !blocked;
        }
        return false;
    }
    
    public boolean canChangeMissionTeamAfter()
    {
	int status = getMissionStatus();
	return canChangeMissionTeamAfter[ status ];
    }
    
    public boolean moveCharacter( Character character, int direction )
    {
        GameAction action;

        if( ! character.isAlive() )
        {
            return false;
        }
        if( ( map.isPassable( character.getLocation(), direction ) ) && ( character.getCurrentActionPoints() > 0 ) )
        {
            Point target = map.getNeighbor( character.getLocation(), direction );	    
            if( !map.isBlockedByInhabitant( target, 100-character.getSize(), false ) )
            {
                action = character.doAction( new MoveAction( character, character.getLocation(), target ) );
                setChanged();
                notifyObservers( action );
                //System.out.println( "Move character " + character.toString() + ": Direction " + direction + " Target " + target.toString() );
                checkObjectives( action );
                return true;
            }
        }
        return false;        
    }
    
    public boolean movePlayer( int index, int direction )
    {
	boolean success = false;
	Character character = (Character)playerCharacters.elementAt( index );
	Point target = map.getNeighbor( character.getLocation(), direction );
	if( target.x < 0 || target.x >= map.getHorizontalSize() || target.y < 0 || target.y >= map.getVerticalSize() )
	{
	    handleOutOfAP( character );
	    return false;
	}
	// Check for enemies
	for( int i=monsters.size()-1; i >= 0; i-- )
	{
	    Monster monster = (Monster)(monsters.elementAt( i ));
	    if( monster.isAlive() && monster.getLocation().equals( target ) && monster.isFoe( character ) )
	    {
		return false;
	    }
	}
        success = moveCharacter( character, direction );
	if( !success )
	{
	    // Check whether the move was blocked by an object
	    MapObject mapObject = map.getObjectForPoint( target );
	    if( mapObject != null )
	    {
		success = handlePush( character, mapObject, direction );
	    }
	}
	if( success )
	{
	    // Notify the enemies
	    checkMonsterVisibility( character );
	}
	return success;
    }
    
    public void checkMonsterVisibility( Character character )
    {
	for( int i=0; i < monsters.size(); i++ )
	{
	    Monster monster = getMonster( i );
	    if( monster.isAlive() )
	    {
		if( character == null )
		{
		    for( int p=0; p < playerCharacters.size(); p++ )
		    {
			Character check = getPlayerCharacter( p );
			if( isCellVisibleToCharacter( monster, check.getLocation() ) )
			{
			    monster.setSpottedPlayer( check, true );
			}
		    }
		}
		else
		{
		    if( isCellVisibleToCharacter( monster, character.getLocation() ) )
		    {
			monster.setSpottedPlayer( character, true );
		    }
		}
	    }
	}
    }
 
    public boolean openDoor( Character character, Point door )
    {
        Point otherHalf = new Point( door );
        Door theDoor = map.getDoor( door.x, door.y );
        if( theDoor != null )
        {
            if( theDoor.isClosed() && !theDoor.isLocked() )
            {
                otherHalf = map.getNeighbor( door, theDoor.getDirection() );
                if( !map.isClosedDoor( otherHalf.x, otherHalf.y ) )
                {
                    otherHalf = door;
                }
                map.openDoor( door );
		checkMonsterVisibility( null );
                setChanged();
                notifyObservers( new DoorAction( character, door, otherHalf, true ) );
                return true;
            }
	}
	return false;
    }
        
    public boolean closeDoor( Character doorUser, Point door )
    {
        Character character;
        Point otherHalf = new Point( door );
        Door theDoor = map.getDoor( door.x, door.y );
	if( theDoor != null )
	{
            if( theDoor.isOpen() && !theDoor.isLocked() )
            {
                otherHalf = map.getNeighbor( door, theDoor.getDirection() );
                if( !map.isOpenDoor( otherHalf.x, otherHalf.y ) )
                {
                    otherHalf = door;
                }
                // Check the characters
                for( int i = 0; i < playerCharacters.size(); i++ )
                {
                    character = getPlayerCharacter( i );
                    if( character.getLocation().equals( door ) || character.getLocation().equals( otherHalf ))
                    {
                        if( character.isAlive() )
                        {
                            character.applyDamage( Integer.MAX_VALUE );
                        }
                    }
                }
		WoundAction woundAction;
                for( int i = 0; i < monsters.size(); i++ )
                {
                    character = getMonster( i );
                    if( character.getLocation().equals( door ) || character.getLocation().equals( otherHalf ) )
                    {
                        if( character.isAlive() )
                        {
			    woundAction = new WoundAction( character, character.getLocation() );
			    setChanged();
			    notifyObservers( woundAction );
                            character.applyDamage( Integer.MAX_VALUE );
			    /*
			    if( !character.isAlive() && character.leavesCorpse() )
			    {
				addCorpse( character );
			    }
			     */
                        }
                    }
                }
		for( int i=0; i < map.getNumberOfObjects(); i++ )
		{
		    MapObject object = map.getObject( i );
		    if( object.isLocatedIn( door ) || object.isLocatedIn( otherHalf ) )
		    {
			if( object.getCostToMove() > 0 )
			{
			    // Static objects cannot be moved to doors, so let's not destroy them if they
			    // happen to partially sit on them on the map
			    object.applyDamage( Integer.MAX_VALUE );
			    ObjectDestroyedEvent objectDestroyedEvent = new ObjectDestroyedEvent( object );
			    setChanged();
			    notifyObservers( objectDestroyedEvent );			    
			}
		    }
		}
		SyncEvent sync = new SyncEvent(); // Sync the possible deaths
		sync.setSyncDoors( false );
		setChanged();
		notifyObservers( sync );
                map.closeDoor( door );
                DoorAction action = new DoorAction( doorUser, door, otherHalf, false );
                checkObjectives( action );
                setChanged();
                notifyObservers( action );
                return true;
            }
	}
	return false;
    }

    private void handleAreaDamage( Character attacker, Point [] area, Weapon weapon )
    {
        Point point;
        Character character;
        Objective objective;
        GameAction action;
        int damage = weapon.getDamage();
	Vector damagedObjects = new Vector();
	boolean objectivesKilled = false;
        for( int i=0; i < area.length; i++ )
        {
            // Check player characters in area
            for( int j = 0; j < getPlayerCharacters().size(); j++ )
            {
                character = getPlayerCharacter( j );
                if( character.getLocation().equals( area[ i ] ) )
                {
                    if( character.isAlive() )
                    {
			action = new WoundAction( character, character.getLocation() );
			setChanged();
			notifyObservers( action );

                        character.applyDamage( damage );
                        if( !character.isAlive() )
                        {
                            if( attacker instanceof Armed )
                            {
                                ((Armed)attacker).addKill();
                            }
                        }
                    }
                }
            }
            // And monsters
            for( int j=0; j < getMonsters().size(); j++ )
            {
                character = getMonster( j );
                if( character.getLocation().equals( area[ i ] ) )
                {
                    if( character.isAlive() )
                    {
                        action = new WoundAction( character, character.getLocation() );
			setChanged();
			notifyObservers( action );
			character.applyDamage( damage );
			if( !character.isAlive() )
                        {
                            if( attacker instanceof Armed )
                            {
                                ((Armed)attacker).addKill();
                            }
			    if( character.leavesCorpse() )
			    {
				addCorpse( character );
			    }
			}
		    }
                }
            }
            MapObject object;
            // And objects
	    boolean objectsDestroyed = false;
            for( int j=0; j < map.getNumberOfObjects(); j++ )
            {
                object = map.getObject( j );
                if( !damagedObjects.contains( object ) && object.isLocatedIn( area[ i ] ) && !object.isDestroyed() )
                {
		    damagedObjects.add( object );
                    if( object.applyDamage( damage ) )
		    {
			ObjectDestroyedEvent objectDestroyedEvent = new ObjectDestroyedEvent( object );
			setChanged();
			notifyObservers( objectDestroyedEvent );
			objectsDestroyed = true;
		    }
                }
            }
            // And finally, Objectives
            for( int j=0; j < getNumberOfObjectives(); j++ )
            {
                objective = getObjective( j );
                if( !damagedObjects.contains( objective ) && objective.isDestructible() && objective.isLocatedIn( area[ i ] ) )
                {
		    damagedObjects.add( objective );
                    if( objective.handleDamage( area[ i ], damage ) )
		    {
                        objectivesKilled = true;
		    }
                }
            }
            if( objectivesKilled || objectsDestroyed )
	    {
                setMapChanged();
	    }
        }

    }

    public boolean handleFirstAid( Character healer, Character target )
    {
        int status = -1;
        if( !map.isNeighbor( healer.getLocation(), target.getLocation() ) )
        {
            return false;
        }
        int firstAidAbility = 0;
        if( healer instanceof Marine )
        {
            firstAidAbility = ((Marine)healer).getAttribute( Marine.FIRST_AID );
        }
        if( healer.getCurrentActionPoints() >= 2 )
        {
            int damage = target.getMaximumHitPoints()-target.getCurrentHitPoints();
            int skillRoll = rollDie();
            int difficulty = 15-target.getCurrentHitPoints();
            if( difficulty < 5 )
            {
                difficulty = 5;
            }
            if( skillRoll == 1 && firstAidAbility < 5 )
            {
                // Fumble
                //System.out.println( "Fumbled first aid!" );
                target.applyDamage( (int)(Math.random()*3) );
                status = HealAction.FUMBLE;
            }
            else if( ( skillRoll + firstAidAbility ) >= difficulty )
            {
                int oldHealth = target.getCurrentHitPoints();
                target.healDamage( 1+(int)(Math.random()*3) );
                int newHealth = target.getCurrentHitPoints();
                if( healer instanceof Marine && difficulty > 5 )
                {
                    ((Marine)healer).grantExperienceTo( Marine.FIRST_AID, (int)(Math.random()*2) );
                }
                if( newHealth > oldHealth )
                {
                    status = HealAction.HEAL;
                }
                else
                {
                    status = HealAction.FULL_HEALTH;
                }
            }
            else
            {
                //System.out.println( "Failed first aid" );
                status = HealAction.ERROR;
            }
            
            UseAction healerAction = new UseAction( healer, target.getLocation() );
            healerAction.setCost( 2 );
            healer.doAction( healerAction );
            setChanged();
            notifyObservers( healerAction );
            HealAction healAction = new HealAction( target, status );
            setChanged();
            notifyObservers( healAction );
            return true;
        }
	handleOutOfAP( healer );
        return false;
    }
    
    // Returns true if attack is handled
    public boolean handleRangedAttack( Character attacker, Point target ) throws IllegalArgumentException
    {
        if( !(attacker instanceof Armed) )
        {
            throw new IllegalArgumentException( "Only armed characters can try to shoot" );
        }
        boolean handled = false;
        GameAction action;
        Weapon weapon;
        weapon = ((Armed)attacker).getWeapon();
        int actionCost = weapon.getActionsToFire();
	if( ( attacker.getCurrentActionPoints() < actionCost ) || ( actionCost <= 0 ) )
	{
	    handleOutOfAP( attacker );
	    return false;
	}	
	ShootAction shootAction = new ShootAction( attacker, weapon, target, false );
	attackActions.add( shootAction );
	alertCharactersForNoise( attacker );
	
	setChanged();
	notifyObservers( new SyncEvent() );
	
	//System.out.println( "Added ShootAction from " + attacker.getClass().getName() + ", attacks: " + attackActions.size() );
	if( attackActions.size() == 1 )
	{
	    processAttackActions();
	}
	checkObjectives( shootAction );
	return true;
    }
     	
    public void processAttackActions()
    {
	ShootAction shootAction;
	Character attacker;
	Weapon weapon;
	Point target;
	boolean hit = false;
	int actionCost = 0;
	GameAction action;
	
	while( attackActions.size() > 0 )
	{
	    action = (GameAction)attackActions.remove( 0 ); //elementAt( 0 );
	    if( action instanceof NewCorpseAction )
	    {
		setChanged();
		notifyObservers( action );
	    }
	    if( action instanceof ShootAction )
	    {
		shootAction = (ShootAction)action;//shootActions.elementAt( 0 );
		attacker = shootAction.getActor();
		weapon = shootAction.getWeapon();
		actionCost = weapon.getActionsToFire();
		target = shootAction.getEndPoint();
		
		if( weapon.hasAreaEffect() )
		{
		    int skillLevel = ((Armed)attacker).getWeaponSkill();
		    int difficulty = (int)attacker.getLocation().distance( target );
		    //System.out.print( "Shooting at range " + difficulty );
		    difficulty = (difficulty-5)*2;
		    //System.out.println( ", difficulty " + difficulty );
		    int roll = rollDie();
		    if( skillLevel+roll >= difficulty )
		    {
			// It's a hit
			hit = true;
			target = getMap().getLineOfFire( attacker.getLocation(), target, weapon.getBlockingSize() );
			((Armed)attacker).registerHit();
		    }
		    else
		    {
			hit = false;
			// Let's offset the target
			System.out.print( "Missed shooting at (" + target.x + "," + target.y +"), " );
			int offset = difficulty-roll;
			offset=offset/2; // Let's make it somewhat smaller until we get the difficulties properly in place
			int newX = target.x+(int)Math.random()*2*offset-offset;
			if( newX < 0 ) newX = 0;
			if( newX > map.getHorizontalSize() ) newX = map.getHorizontalSize();
			int newY = target.y+(int)Math.random()*2*offset-offset;
			if( newY < 0 ) newY = 0;
			if( newY > map.getVerticalSize() ) newY = map.getVerticalSize();
			target = new Point( newX, newY );
			target = getMap().getLineOfFire( attacker.getLocation(), target, weapon.getBlockingSize() );
			//System.out.println( "hitting (" + target.x + "," + target.y +") instead" );
		    }
		    //System.out.println( "Hitting " + target.x + ", " + target.y );
		    Point [] area = weapon.getArea( this, attacker.getLocation(), target );
		    shootAction.setTarget( target );
		    shootAction.setCost( actionCost );
		    shootAction.setHit( hit );
		    action = attacker.doAction( shootAction );
		    setChanged();
		    notifyObservers( action );
		    
		    handleAreaDamage( attacker, area, weapon );
		}
		else
		{
		    Point targetPoint = getMap().getLineOfFire( attacker.getLocation(), target );
		    //System.out.println( "Lands in " + targetPoint.toString() );
		    int targetIndex = -1;
		    int targetType = -1;
		    
		    // Target is selected in reverse order because sprites are drawn in direct order
		    // That way, the topmost monster becomes the target if there are several of them in the same cell
		    // Objectives and players are checked first, so if there's a monster in the cell, too, it will become target
		    for( int i = getNumberOfObjectives()-1; i >= 0; i-- )
		    {
			if( getObjective( i ).isDestructible() && getObjective( i ).isLocatedIn( targetPoint ) )
			{
			    targetType = OBJECTIVE;
			    targetIndex = i;
			    break;
			}
		    }
		    MapObject mapObject = map.getObjectForPoint( targetPoint );
		    if( mapObject != null && !mapObject.isDestroyed() )
		    {
			targetType = OBJECT;
			targetIndex = 1; // Does not actually matter, except to mark a hit
		    }
		    for( int i = getPlayerCharacters().size()-1; i >= 0; i-- )
		    {
			if( getPlayerCharacter( i ).getLocation().equals( targetPoint ) )
			{
			    if( getPlayerCharacter( i ).isAlive() )
			    {
				targetType = PLAYER;
				targetIndex = i;
				break;
			    }
			}
		    }
		    for( int i = getMonsters().size()-1; i >= 0; i-- )
		    {
			if( getMonster( i ).getLocation().equals( targetPoint ) )
			{
			    if( getMonster( i ).isAlive() )
			    {
				targetType = MONSTER;
				targetIndex = i;
				break;
			    }
			}
		    }
		    if( targetIndex >= 0 && attacker.getCurrentActionPoints() >= actionCost && actionCost > 0 )
		    {
			hit = false;
			if( targetIndex > -1 )
			{
			    int skillLevel = ((Armed)attacker).getWeaponSkill();
			    int difficulty = (int)attacker.getLocation().distance( target );
			    //System.out.print( "Shooting at range " + difficulty );
			    difficulty = (difficulty-5)*2;
			    //System.out.println( ", difficulty " + difficulty );
			    if( skillLevel+rollDie() >= difficulty )
			    {
				// It's a hit
				hit = true;
				((Armed)attacker).registerHit();
			    }
			    else
			    {
				hit = false;
			    }
			}
			shootAction.setCost( actionCost );
			shootAction.setHit( hit );
			action = attacker.doAction( shootAction );
			setChanged();
			notifyObservers( action );
			if( hit )
			{
			    int damage = weapon.getDamage();
			    if( targetType == PLAYER )
			    {
				action = new WoundAction( getPlayerCharacter( targetIndex ), targetPoint );
				setChanged();
				getPlayerCharacter( targetIndex ).applyDamage( damage );
				notifyObservers( action );
			    }
			    if( targetType == MONSTER )
			    {
				action = new WoundAction( getMonster( targetIndex ), targetPoint );
				setChanged();
				getMonster( targetIndex ).applyDamage( damage );
				notifyObservers( action );
				if( !getMonster( targetIndex ).isAlive() )
				{
				    ((Armed)attacker).addKill();
				    if( getMonster( targetIndex ).leavesCorpse() )
				    {
					addCorpse( getMonster( targetIndex ) );
				    }
				}
			    }
			    boolean objectivesKilled = false;
			    if( targetType == OBJECT )
			    {
				mapObject = map.getObjectForPoint( targetPoint );
				if( mapObject != null )
				{
				    if( mapObject.applyDamage( damage ) )
				    {
					ObjectDestroyedEvent objectDestroyedEvent = new ObjectDestroyedEvent( mapObject );
					setChanged();
					notifyObservers( objectDestroyedEvent );					
					objectivesKilled = true;
				    }
				}
			    }
			    if( targetType == OBJECTIVE )
			    {
				if( getObjective( targetIndex ).handleDamage( targetPoint, damage ) )
				{
				    objectivesKilled = true;
				}
			    }
			    if( objectivesKilled )
			    {
				setMapChanged();
			    }
			}
		    }
		    else
		    {
			handleOutOfAP( attacker );
		    }
		}
	    }
	}
    }
    
    public boolean handleUse( Character character, Point point )
    {
	// Dead characters don't do anything
	if( !character.isAlive() )
	{
	    return false;   
	}
	if( isPlayerCharacter( character ) )
	{
	    // Search for suitable target
	    Character target = null;
	    
	    for( int i=0; i < playerCharacters.size(); i++ )
	    {
		target = getPlayerCharacter( i );
		if( target.getLocation().equals( point ) )
		{
		    handleFirstAid( character, target );
		    return true;
		}
	    }
	}
	// Check objectives
	UseAction useAction = new UseAction( character, point );
	Objective objective;
	for( int i = 0; i < getNumberOfObjectives() && !useAction.isConsumed(); i++ )
	{
	    objective = getObjective( i );
	    if( objective.isUsable() )
	    {
		if( Map.isNeighbor( character.getLocation(), point ) || objective.isRemoteControlled() )
		{
		    if( objective.handleAction( useAction ) )
		    {
			character.doAction( useAction );
		    }
		}
	    }
	}
	if( useAction.isConsumed() )
	{
	    // Some objective claimed the action. Stop here
	    setChanged();
	    notifyObservers( useAction );   // Animation will occur in reverse direction now, but...
	    return true;
	}
	// If no Objective claimed the action, try the objects
	MapObject mapObject = map.getObjectForPoint( point );
	int direction = map.getDirection( character.getLocation(), point, character.getDirection() );
	if( mapObject != null )
	{
	    if( handlePush( character, mapObject, direction ) )
	    {
		return true;
	    }
	}
	//System.out.println( "Use action not consumed, trying doors" );
	if( character.getCurrentActionPoints() >= 1 )
	{
	    if( Map.isNeighbor( character.getLocation(), point ) )
	    {
		if( map.isClosedDoor( point ) )
		{
		    useAction.setCost( 1 );
		    character.doAction( useAction );
		    setChanged();
		    notifyObservers( useAction );
		    if( openDoor( character, point ) )
		    {
			checkObjectives( useAction );
			return true;
		    }
		}
		if( map.isOpenDoor( point ) )
		{
		    useAction.setCost( 1 );
		    character.doAction( useAction );
		    setChanged();
		    notifyObservers( useAction );
		    if( closeDoor( character, point ) )
		    {
			checkObjectives( useAction );
			return true;
		    }
		}
	    }
	}
	else
	{
	    handleOutOfAP( character );   
	}
	return false;
    }
    
    public void setMapChanged()
    {
        setChanged();
	notifyObservers( new SyncEvent() );
	setChanged();
        notifyObservers( new MapEvent() );
    }

    public boolean handleMelee( Character attacker, Character target )
    {
	//System.out.println( "Attacking " + target.getClassName() );
	GameAction action;
	boolean success = false;
	if( attacker.getCurrentActionPoints() > 0 && map.isNeighbor( attacker.getLocation(), target.getLocation() ) )
	{
	    success = true;
	    int direction = map.getDirection( attacker.getLocation(), target.getLocation(), attacker.getDirection() );
	    attacker.setDirection( direction );
	    boolean targetDefends = false;
	    int attackerMeleeAbility = attacker.getMeleeAbility();
	    int defenderMeleeAbility = target.getMeleeAbility();
	    // if the target has action points and is not attacked directly from behind, can defend itself
	    if( (target.getCurrentActionPoints() > 0) && (target.getDirection() != direction) && (defenderMeleeAbility >= 0) )
	    {
		targetDefends = true;
		//System.out.println( "Target defends" );
	    }
	    action = new MeleeAction( attacker, target, targetDefends );
	    ((MeleeAction)action).setCost( attacker.getMeleeCost() );
	    attacker.doAction( action );
	    setChanged();
	    notifyObservers( action );
	    int damage = 0;
	    if( targetDefends )
	    {
		target.setDirection( (direction+4)%8 );
		int attackerScore = attackerMeleeAbility + rollDie();
		int defenderScore = defenderMeleeAbility + rollDie();
		//System.out.println( "Difference: " + (attackerScore-defenderScore) );
		if( attackerScore > defenderScore )
		{
		    damage = 1+(int)( Math.random()*attacker.getMeleeAbility() );
		}
		if( attackerScore < defenderScore )
		{
		    // Ouch, attacker takes damage!
		    //System.out.println( "Defender overwhelms attacker!" );
		    damage = 1+(int)( Math.random()*target.getMeleeAbility() );
		    Character temp = attacker;
		    attacker = target;
		    target = temp;
		}
	    }
	    else
	    {
		damage = 1+(int)(Math.random()*(attackerMeleeAbility));
		if( target.getDirection() == direction )
		{
		    // Double damage from behind
		    damage *= 2;
		}
	    }
	    if( damage > 0 )
	    {
		action = new WoundAction( target, target.getLocation() );
		setChanged();
		target.applyDamage( damage );
		//System.out.println( target.getType() + " takes " + damage + " points of damage" );
		notifyObservers( action );
		if( !target.isAlive() )
		{
		    if( attacker instanceof Marine )
		    {
			((Marine)attacker).addKill();
		    }
		    if( target.leavesCorpse() )
		    {
			addCorpse( target );
		    }
		}
	    }
	}
	return success;
	
    }
    
    public void handleMessage( Message message )
    {
        setChanged();
        notifyObservers( message );
    }
    
    public void handleMessage( String message )
    {
	String messageString = message.replaceAll( "\\\\n", "\n" );
	Message theMessage = new Message( null, messageString );
	handleMessage( theMessage );
    }
    
    public void handleOutOfAP( Character character )
    {
	GameAction action = new NoAPAction( character, character.getLocation() );
	setChanged();
	notifyObservers( action );
    }
    
    public boolean handlePush( Character character, MapObject mapObject, int direction )
    {
	boolean success = true;
	Character check;
	int costToMove = mapObject.getCostToMove();
	if( costToMove < 0 )
	{
	  return false;
	}
	if( map.canMoveObject( mapObject, direction ) && character.getCurrentActionPoints() >= costToMove )
	{
	    boolean doesMove = true;
	    // Get the area of the object
	    MapObjectCell [] cells = mapObject.getCells();
	    // Check whether there are any characters in the area
	    for( int i=0; i < cells.length; i++ )
	    {
		Point point = Map.getNeighbor( new Point( cells[i].getX(), cells[i].getY() ), direction );
		for( int p=0; p < playerCharacters.size(); p++ )
		{
		    check = getPlayerCharacter( p );
		    if( check.getLocation().equals( point ) )
		    {
			WoundAction action = new WoundAction( check, check.getLocation() );
			setChanged();
			notifyObservers( action );
			check.applyDamage( costToMove );
			doesMove = false;
		    }
		}
		for( int m=0; m < monsters.size(); m++ )
		{
		    check = getMonster( m );
		    if( check.getLocation().equals( point ) && check.isAlive() )
		    {
			WoundAction action = new WoundAction( check, check.getLocation() );
			setChanged();
			notifyObservers( action );
			check.applyDamage( costToMove );
			if( !check.isAlive() )
			{
			    if( character instanceof Armed )
			    {
				((Armed)character).addKill();
			    }
			    if( check.leavesCorpse() )
			    {
				addCorpse( check );
			    }
			}
			doesMove = false;
		    }
		}
		for( int o=0; o < objectives.size(); o++ )
		{
		    Objective objective = getObjective( o );
		    if( objective.isLocatedIn( point ) && objective.isDestructible() )
		    {
			objective.handleDamage( point, costToMove );
			doesMove = false;
		    }
		}
	    }
	    if( doesMove )
	    {		
		Point target = map.getNeighbor( character.getLocation(), direction );
		MoveAction mAction = new MoveAction( character, character.getLocation(), target );
		mAction.setCost( costToMove );
		character.doAction( mAction );   
		setChanged();
                notifyObservers( mAction );
		PushObjectAction pushAction = new PushObjectAction( character, mapObject, target, map.getNeighbor( target, direction ) );
		setChanged();
		notifyObservers( pushAction );
		map.moveObject( mapObject, direction );		
                //System.out.println( "Move character " + character.toString() + ": Direction " + direction + " Target " + target.toString() );
                checkObjectives( mAction );
	    }
	    else
	    {
		Point target = map.getNeighbor( character.getLocation(), direction );
		PushObjectAction pushAction = new PushObjectAction( character, mapObject, target, map.getNeighbor( target, direction ) );
		pushAction.setBounce( true );
		setChanged();
		notifyObservers( pushAction );

		UseAction uAction = new UseAction( character, character.getLocation() );
		uAction.setCost( costToMove );
		character.doAction( uAction );
		setChanged();
		notifyObservers( uAction );
		checkObjectives( uAction );
	    }
	}
	else
	{
	    success = false;   
	}
	return success;
    }
    
    public void doMonsterActions()
    {
	state = MONSTER_MOVE;
	int maximumActionPoints = 0;
        int check = 0;

	// Do pre-actions
	for( int i=0; i < monsters.size(); i++ )
	{
	    //if( getMonster( i ).isAlive() && !getMonster( i ).isDormant() )
	    {
		getMonster( i ).preAction();
	    }
	}
        // Count the maximum number of action points;
        for( int i = 0; i < monsters.size(); i++ )
        {
            check = getMonster( i ).getMaximumActionPoints();
            if( check > maximumActionPoints )
	    {
                maximumActionPoints = check;
	    }
        }
	// Do the actions
        for( int i = 0; i < maximumActionPoints; i++ )
	{
            System.out.print( "." );
	    processMonsterActions();
	    /// Sync
	    SyncEvent syncEvent = new SyncEvent();
	    syncEvent.setSyncDoors( true );
	    syncEvent.setSyncMove( true );
	    syncEvent.setSyncAttacks( true );
	    setChanged();
	    notifyObservers( syncEvent );	    
	}
        System.out.print( "\n" );
	// Do post-actions
	for( int i=0; i < monsters.size(); i++ )
	{
	  //  if( getMonster( i ).isAlive() && !getMonster( i ).isDormant() )
	   // {
		getMonster( i ).postAction();
	   // }
	}
	state = PLAYER_MOVE;
    }
    
    private boolean processMonsterActions()
    {
	hasMoreMonsterActions = false;
        GameAction action = null;
        MoveAction mAction;
        Vector actions = new Vector();
	for( int i = 0; i < monsters.size(); i++ )
	{
	    if( getMonster( i ).isAlive() && !getMonster( i ).isDormant() )
	    {
		try
		{
		    action = getMonster( i ).playAction();
		}
		catch( Exception e )
		{
		    System.out.println( "Monster borked!" );
		    e.printStackTrace();
		    action = new WaitAction( getMonster( i ), getMonster( i ).getLocation() );
		}
		actions.add( action );
		if( action instanceof ShootAction )
		{
		    handleRangedAttack( ((ShootAction)action).getActor(), ((ShootAction)action).getEndPoint() );
		    checkObjectives( action );
		}
		if( action instanceof MoveAction )
		{
		    MoveAction move = (MoveAction)action;
		    // If can't make the move, ignore it. The character will still lose the action points
		    if( !map.isPassable( move.getEndPoint() ) && !((Monster)move.getActor()).ignoresWalls() )
		    {
			move.getActor().setLocation( move.getStartPoint() );
			move.setEndPoint( move.getStartPoint() );
		    }
		    getMonster( i ).doAction( action );
		    setChanged();
		    notifyObservers( move );
		    checkObjectives( action );
		}
		if( action instanceof MeleeAction )
		{
		    MeleeAction melAction = (MeleeAction)action;
		    handleMelee( melAction.getAttacker(), melAction.getDefender() );
		    checkObjectives( action );
		}
		if( action instanceof UseAction )
		{
		    // HandleUse also checks the objectives, no need to re-do
		    handleUse( ((UseAction)action).getActor(), ((UseAction)action).getPoint() );
		}
		if( getMonster( i ).hasMoreActions() )
		{
		    hasMoreMonsterActions = true;
		    //System.out.println( getMonster( i ).getClassName() + " has more actions" );
		}
	    }
	}
	/*
        for( int i=0; i < actions.size(); i++ )
        {
            action = (GameAction)actions.elementAt( i );
            checkObjectives( action );
        }*/
	return hasMoreMonsterActions;
    }
    
    public void checkObjectives( GameAction action )
    {
        for( int i = 0; i < getNumberOfObjectives(); i++ )
        {
	    getObjective(i).handleAction( action );
        }
    }
    
    public void setupNewTurn()
    {
	for( int i = 0; i < getPlayerCharacters().size(); i++ )
	{
	    getPlayerCharacter( i ).resetActionPoints();
            ((Marine)getPlayerCharacter( i )).checkActionPoints();
	}
	for( int i = 0; i < getMonsters().size(); i++ )
	{
	    getMonster( i ).resetActionPoints();
	}
	// Advance the objectives
	for( int i = 0; i < getNumberOfObjectives(); i++ )
        {
            if( getObjective(i) != primaryObjective )
            {
                getObjective( i ).isCompleted();
            }	
        }
	hasMoreMonsterActions = false;
	state = PLAYER_MOVE;
        turns++;
    }
    
    public int getState()
    {
        return state;
    }

    public int getMaximumMonsterActions()
    {
	return maximumMonsterActions;
    }

    public String getDebriefMessage()
    {
	int status = getMissionStatus();
	String debriefMessage = debriefMessages[ status ];
	if( debriefMessage.indexOf( "$CLIP-" ) > -1 )
	{
	    //System.out.println( "Clipping: " + missionDesc );
	    int clipStart = debriefMessage.indexOf( "$CLIP-" );
	    int clipEnd = debriefMessage.indexOf( "$", clipStart+1 );
	    //System.out.println( "Range: " + clipStart + " to " + clipEnd );
	    String id = debriefMessage.substring( clipStart+6, clipEnd );
	    //System.out.println( "Trying to clip id: " + id );
	    String clipContent = "";
	    Objective objective = getObjectiveById( id );
	    if( objective != null )
	    {
		clipContent = objective.getDescriptionClip();
		//System.out.println( "Clip content: " + clipContent );
	    }
	    String clipString = "$CLIP-"+id+"$";
	    //debriefMessage = debriefMessage.substring( 0, clipStart ) + clipContent + debriefMessage.substring( clipEnd+1 );
	    debriefMessage = debriefMessage.replaceAll( "\\$CLIP-"+id+"\\$", clipContent );
	}
	return debriefMessage;
    }
    
    public int getMissionBonus()
    {
	int status = getMissionStatus();
	return missionBonusPoints[ status ];
    }
    
    public String getMissionDescription()
    {
        String missionDesc = longMissionDescription;
        if( missionDesc.indexOf( "$CLIP-" ) > -1 )
        {
	    //System.out.println( "Clipping: " + missionDesc );
	    int clipStart = missionDesc.indexOf( "$CLIP-" );
	    int clipEnd = missionDesc.indexOf( "$", clipStart+1 );
	    //System.out.println( "Range: " + clipStart + " to " + clipEnd );
	    String id = missionDesc.substring( clipStart+6, clipEnd );
	    //System.out.println( "Trying to clip id: " + id );
	    String clipContent = "";
	    Objective objective = getObjectiveById( id );
	    if( objective != null )
	    {
		clipContent = objective.getDescriptionClip();
		//System.out.println( "Clip content: " + clipContent );
	    }
	    String clipString = "$CLIP-"+id+"$";
            //missionDesc = missionDesc.substring( 0, clipStart ) + clipContent + missionDesc.substring( clipEnd+1 );
	    missionDesc = missionDesc.replaceAll( "\\$CLIP-"+id+"\\$", clipContent );

        }
        return missionDesc;
    }
    
    public String getMissionName()
    {
        return missionName;
    }
    
    public String getNextMissionFileName()
    {
	int status = getMissionStatus();
	return nextMissionFileNames[ status ];
    }
    
    public int getMissionStatus()
    {
	int status = Mission.VICTORY;
	if( !isCompleted() )
	{
	    status = Mission.FAILURE;
	}
	if( primaryObjective != null )
	{
	    if( !primaryObjective.isSuccessful() )
	    {
		status = Mission.FAILURE;
	    }
	    if( secondaryObjective != null )
	    {
		if( primaryObjective.isSuccessful() && secondaryObjective.isSuccessful() )
		{
		    status = Mission.COMPLETE_VICTORY;
		}
		if( !primaryObjective.isSuccessful() && !secondaryObjective.isSuccessful() )
		{
		    status = Mission.COMPLETE_FAILURE;
		}
	    }
	}
	return status;
    }

    public Character [] getVisibleCharacters( Character character )
    {
        Vector temp = new Vector();
        Character check;
        // First, check player characters
        for( int i = 0; i < playerCharacters.size(); i++ )
        {
            check = getPlayerCharacter( i );
            if( check.isAlive() && isCellVisibleToCharacter( character, check.getLocation() ) && ( character != check ) )
            {
                temp.add( check );
            }
        }
        // Then do the same for monsters
        for( int i = 0; i < monsters.size(); i++ )
        {
            check = getMonster( i );
            if( check.isAlive() && isCellVisibleToCharacter( character, check.getLocation() ) && ( character != check ) )
            {
                temp.add( check );
            }
        }
        Character [] characters = new Character[ temp.size() ];
        for( int i = 0; i < temp.size(); i++ )
        {
            characters[ i ] = (Character)temp.elementAt( i );
        }
        return characters;
    }
    
    
    public Character [] getVisibleEnemies( Character character )
    {
        Vector temp = new Vector();
        Character check;
        // First, check player characters
        for( int i = 0; i < playerCharacters.size(); i++ )
        {
            check = getPlayerCharacter( i );
            if( check.isAlive() && check.isHostile( character ) && isCellVisibleToCharacter( character, check.getLocation() ) && check != character )
            {
                temp.add( check );
            }
        }
        // Then do the same for monsters
        for( int i = 0; i < monsters.size(); i++ )
        {
            check = getMonster( i );
            if( check.isAlive() && check.isHostile( character ) && isCellVisibleToCharacter( character, check.getLocation() ) && check != character )
            {
                temp.add( check );
            }
        }
        Character [] characters = new Character[ temp.size() ];
        for( int i = 0; i < temp.size(); i++ )
        {
            characters[ i ] = (Character)temp.elementAt( i );
        }
        return characters;
    }

    public Character getClosestCharacter( Character character, Vector targets )
    {
        Character check = null;
        Character target = null;
        double range = Double.MAX_VALUE;
        double checkRange;
        for( int i = 0; i < targets.size(); i++ )
        {
            check = (Character)targets.elementAt(i);
            checkRange = character.getLocation().distance( check.getLocation() );
            if( checkRange < range )
            {
                target = check;
                range = checkRange;
            }
        }
        return target;
    }

    public Character getClosestCharacter( Character character, Character[] targets )
    {
        Character check = null;
        Character target = null;
        double range = Double.MAX_VALUE;
        double checkRange = 0;
        for( int i = 0; i < targets.length; i++ )
        {
            check = targets[ i ];
            checkRange = character.getLocation().distance( check.getLocation() );
            if( ( checkRange < range ) && check.isAlive() )
            {
                target = check;
                range = checkRange;
            }
        }
        return target;
    }
    
    
    public boolean isMonsterVisibleToPlayer( int index )
    {
        Character character;
	Point monsterLocation = getMonster( index ).getLocation();
	for( int i=0; i < playerCharacters.size(); i++ )
	{
            character = getPlayerCharacter( i );
	    if( map.hasLineOfSight( character.getLocation(), monsterLocation ) && character.isAlive() )
	    {
		return true;
	    }
	}
	return false;
    }
    
    public boolean isCellVisibleToCharacter( Character character, Point cell )
    {
        return map.hasLineOfSight( character.getLocation(), cell );
    }
 
    public boolean isCellVisibleToPlayer( Point cell )
    {
	for( int i=0; i < playerCharacters.size(); i++ )
	{
	    if( map.hasLineOfSight( getPlayerCharacter( i ).getLocation(), cell ) && getPlayerCharacter( i ).isAlive() )
	    {
		return true;
	    }
	}
	return false;
    }

       
    public void addPlayerCharacter( Character character )
    {
	playerCharacters.add( character );
    }
    
    public Character getPlayerCharacter( int index )
    {
	return (Character)playerCharacters.elementAt( index );
    }
    
    public Vector getPlayerCharacters()
    {
	return playerCharacters;
    }
    
    public void alertCharactersForNoise( Character character )
    {
	Room room = map.getClosestRoomForPoint( character.getLocation() );
	if( room != null )
	{
	    if( isPlayerCharacter( character ) )
	    {
		for( int i=0; i < monsters.size(); i++ )
		{
		    Monster monster = getMonster( i );
		    Room otherRoom = map.getClosestRoomForPoint( monster.getLocation() );
		    if( room.hasLinked( otherRoom ) && monster.isAlive() )
		    {
			//System.out.println( "Alerting " + monster.getClass().getName() );
			monster.setSpottedPlayer( character, false );
		    }
		}
	    }
	    else	    
	    {
		Character alerted = null;
		for( int i=0; i < playerCharacters.size(); i++ )
		{
		    Character player = getPlayerCharacter( i );
		    Room otherRoom = map.getClosestRoomForPoint( character.getLocation() );
		    if( room.hasLinked( otherRoom ) && character.isAlive() )
		    {
			if( alerted == null )
			{
			    alerted = player;
			}
			else
			{
			    if( alerted.getLocation().distance( character.getLocation() ) < player.getLocation().distance( character.getLocation() ) )
			    {
				if( !map.hasLineOfSight( player.getLocation(), character.getLocation() ) )
				{
				    alerted = player;
				}
			    }
			}
		    }
		}
		if( alerted != null )
		{
		    handleMessage( new Message( alerted, ((Marine)alerted).getName() + " hears noise from " + room.getName() ) );
		}
	    }
	}
    }
    
    public void setActiveCharacter( int index )
    {
	activeCharacter = index;
    }

    public boolean isAbortable()
    {
        boolean returned = true;
        Character character;
        Point location;
        boolean characterReturned;

        for( int i = 0; i < getPlayerCharacters().size(); i++ )
        {
            character = getPlayerCharacter(i);
            characterReturned = false;
            if( character.isAlive() )
            {
                for( int j = 0; j < getPlayerStartingPoints().size(); j++ )
                {
                    location = ((MapNode)getPlayerStartingPoints().elementAt(j)).getPoint();
                    if( (Math.abs( character.getLocation().x-location.x ) <= 1) && (Math.abs( character.getLocation().y-location.y ) <= 1) )
                        characterReturned = true;
                }
            }
            else
            {
                // Dead count as returned
                characterReturned=true;
            }
            returned = (returned && characterReturned );
        }
        return returned;
    }
    
    public boolean isActiveCharacter( int index )
    {
	if( activeCharacter == index )
	{
	    return true;
	}
	return false;
    }
    
    public int getActiveCharacterIndex()
    {
	return activeCharacter;
    }
    
    public Character getActiveCharacter()
    {
	return getPlayerCharacter( activeCharacter );
    }

    public void addMonster( Character character )
    {
	int low = 0;
	int high = monsters.size();
	int check = 0;
	Character test;
	int iterations = 0;
	if( high > 0 )
	{
	    if( character.getSize() >= ((Character)monsters.elementAt( high-1 ) ).getSize() )
	    {
		character.setGame( this );
		monsters.add( character );
		return;
	    }
	}
	while( (low <= high) && (iterations < monsters.size() ) )
	{
	    iterations++;
	    //System.out.println( "Testing range: " + low + " ... " + high );
	    check = (low+high)/2;
	    test = (Character)monsters.elementAt( check );
	    //System.out.println( "Value at check (" + check + ") is " + test.getSize() );
	    if( test.getSize() <= character.getSize() )
	    {
		low = check+1;
	    }
	    else
	    {
		high = check-1;
	    }
	}
        character.setGame( this );
	//System.out.println( "Slot: " + (low) + " after " + iterations + " iterations" );
	monsters.insertElementAt( character, low );
    }

    public void addCorpse( Character character )
    {
        GameAction action = new NewCorpseAction( character, character.getLocation() );
	attackActions.add( action );
	//System.out.println( "Added corpse for " + character.getClass().getName() );
	if( attackActions.size() == 1 )
	{
	    processAttackActions();
	}
    }
        
    public void spawnMonster( Character character )
    {
        addMonster( character );
        setChanged();
        notifyObservers( new MonsterBirth( character ) );
    }
        
    public Monster getMonster( int index )
    {
	return (Monster)monsters.elementAt( index );
    }
    
    public Monster getMonsterById( String id )
    {
	Monster monster = null;
	boolean found = false;
	for( int i=0; (i < monsters.size()) && !found; i++ )
	{
	    String theId = getMonster( i ).getId();
	    if( theId != null && theId.equals( id ) )
	    {
		found = true;
		monster = getMonster( i );
	    }
	}
	return monster;
    }
    
    public Vector getMonsters()
    {
	return monsters;
    }

    public void setPlayerStartingPoints( Vector vector )
    {
	playerStartingPoints = new Vector();
        Point point;
        int dir;
        for( int i=0; i < vector.size(); i++ )
        {
            point = ((MapNode)vector.elementAt( i )).getPoint();
            dir = ((MapNode)vector.elementAt( i )).getDirection();
            if( point.x < 0 || point.y < 0 )
            {
                point = getRandomPointForMonster();
                dir = (int)(Math.random()*8);
            }
            playerStartingPoints.add( new MapNode( point, dir ) );
        }
    }
    
    public Vector getPlayerStartingPoints()
    {
        return playerStartingPoints;
    }
    
    public Point getPlayerStartingPoint( int index )
    {
	if( index < 0 || index >= playerStartingPoints.size() )
	{
	    return null;
	}
	MapNode mapNode = (MapNode)playerStartingPoints.elementAt( index );
	return mapNode.getPoint();
    }
    
    public boolean isPlayerCharacter( Character character )
    {
	return playerCharacters.contains( character );
    }

    public boolean isPlayerStartingPoint( Point point )
    {
        for( int i=0; i < playerStartingPoints.size(); i++ )
        {
            if( point.equals( ((MapNode)playerStartingPoints.elementAt( i )).getPoint() ) )
                return true;
        }
        return false;
    }
        
    public void setPlayers( Vector playerCharacters )
    {
        this.playerCharacters = playerCharacters;
    }

    public void addObjective( Objective objective )
    {
        objective.setGame( this );
	if( objective.getId() == null )
	{
	    objective.setId( String.valueOf( objectives.size() ) );
	}
        objectives.add( objective );
    }
    
    public int rollDie()
    {
	int result = 0;
	do
	{
	    result = 1+(int)(Math.random()*10);
	} while( 0 == (int)(Math.random()*20) );
	return result;
    }
    
    public void initialize()
    {
        // Initialize player locations
        Marine player;
        MapNode node;
        for( int i = 0; i < playerCharacters.size(); i++ )
        {
            node = (MapNode)playerStartingPoints.elementAt( i );
            player = (Marine)getPlayerCharacter(i);
            player.setLocation( node.getPoint() );
            player.setDirection( node.getDirection() );
            player.setGame( this );
        }
        // Initialize actions
        for( int i = 0; i < playerCharacters.size(); i++ )
        {
            player = (Marine)getPlayerCharacter(i);
            player.checkActionPoints();
        }
        System.out.println( "Initialized player starting location(s)" );
        // Intialize monsters
        Character monster;
        Point point;
        int dir;
        if( monsters.size() > 0 )
        {
            for( int i = 0; i < monsters.size(); i++ )
            {
                monster = (Character)monsters.elementAt( i );
                point = monster.getLocation();
                if( point.x < 0 || point.y < 0 )
                {
                    point = getRandomPointForMonster();
                    dir = (int)(Math.random()*8);
                    monster.setLocation( point );
                    monster.setDirection( dir );
                }
                monster.setGame( this );
            }
        }
        else
        {
            randomizeMonsters();
        }
        System.out.println( "Initialized monsters" );
        // Intialize objectives
        Objective objective;
	if( primaryObjective != null )
	{
	    objectives.add( primaryObjective );
	}
	if( secondaryObjective != null )
	{
	    objectives.add( secondaryObjective );
	}
        for( int i = 0; i < objectives.size(); i++ )
        {
            objective = getObjective( i );
            //System.out.println( "Initializing objective " + i );
	    if( objective.getGame() != this )
	    {
		objective.setGame( this );
	    }
        }
        System.out.println( "Initialized objectives" );
    }
    
    public void setMonsters( Vector theMonsters )
    {
        monsters = new Vector();
	for( int i = 0; i < theMonsters.size(); i++ )
	{
	    addMonster( (Character)theMonsters.elementAt( i ) );
	}
    }

    public void setObjectives( Vector theObjectives )
    {
        objectives = theObjectives;
	
	// Default Objectives. Should get these from an XML configuration file
	Objective objective = new CommandBobsObjective();
	addObjective( objective );
	objective = new MineBehaviorObjective();
        addObjective( objective );
	
    }

    public void randomizeMonsters()
    {	
	// Calculate the team's skills
	int totalLevels = 0;
	for( int i=0; i < playerCharacters.size(); i++ )
	{
	    Marine marine = (Marine)getPlayerCharacter( i );
	    for( int j=0; j < marine.ATTRIBUTES; j++ )
	    {
		totalLevels += marine.getAttribute( j );
	    }
	}
	int averageLevels = (int)(totalLevels/(double)getPlayerCharacters().size());
	System.out.println( "Average levels in team: " + averageLevels );
	int monsterAverage = averageLevels+10;
	int monsterCount = (int)(monsterAverage/2 + (Math.random()*monsterAverage) );
	System.out.println( "Monsters: " + monsterCount );
	
        int theFaction = (int)(Math.random()*8);
        MonsterFaction faction = new DefaultFaction();
	
        switch( theFaction )
        {
            case 0:
            case 1:
                faction = new RipperFaction(); break;
            case 2:
            case 3:
                faction = new BruteFaction(); break;
            case 4:
                faction = new DroidFaction(); break;
        }
        faction.addMonsters( this, monsterCount );
    }

    public Point getRandomPointForMonster()
    {
        int x, y;
        // Isn't failsafe. If the player manages to see the whole map, gets caught in an endless loop
        do
        {
            x = (int)(Math.random()*map.getHorizontalSize());
            y = (int)(Math.random()*map.getVerticalSize());
	} while( !map.isPassable( x, y ) || isCellVisibleToPlayer( new Point( x, y ) ) );
	//System.out.println( "Random point for monster: " + x + ", " + y );
        return new Point( x, y );
    }


    public Map getMap()
    {
	return map;
    }
    
    public int getNumberOfObjectives()
    {
        return objectives.size();
    }
    
    public Objective getObjective( int index )
    {
        return (Objective)objectives.elementAt( index );
    }

    public Objective getPrimaryObjective()
    {
	return primaryObjective;
    }
    
    public Objective getSecondaryObjective()
    {
	return secondaryObjective;
    }
    
    public Objective getObjectiveById( String id )
    {
	if( id.equalsIgnoreCase( "Primary" ) )
	{
	    return primaryObjective;
	}
	if( id.equalsIgnoreCase( "Secondary" ) )
	{
	    return secondaryObjective;
	}
	
	for( int i=0; i < objectives.size(); i++ )
	{
	    String theId = ((Objective)objectives.elementAt( i )).getId();
	    if( theId != null && theId.equals( id ) )
	    {
		return (Objective)objectives.elementAt( i );
	    }
	}
	
	System.err.println( "No match for objective id: " + id );
	return null;
    }
    
    public boolean isCompleted()
    {
        boolean completed = true;

	Character character;
	if( !primaryObjective.isCompleted() )
	{
	    completed = false;
	}
	if( endsAtBase )
	{
	    boolean hasReturned = false;
	    for( int i = 0; i < playerCharacters.size(); i++ )
	    {
		hasReturned = false;
		character = getPlayerCharacter(i);
		if( character.isAlive() )
		{
		    for( int j = 0; j < getPlayerStartingPoints().size(); j++ )
		    {
			Point location = ((MapNode)getPlayerStartingPoints().elementAt(j)).getPoint();
			if( (Math.abs( character.getLocation().x-location.x ) <= 1) && (Math.abs( character.getLocation().y-location.y ) <= 1) )  
			{
			    hasReturned = true;
			}
		    }
		}
		else
		{
		    // Dead count as returned
		    hasReturned=true;
		}
		completed = (completed && hasReturned );
	    }
	}
        return completed;
    }

    public int getElapsedTurns()
    {
        return turns;
    }
}