// GameMapPanel.java
// Started on June 16th, 2001 as GamePanel
// Copyright (c) Virtu Halttunen

package TheGame;

import java.awt.*;
import java.awt.font.GlyphVector;
import java.awt.font.FontRenderContext;
import java.awt.image.*;
import javax.swing.*;
import java.util.*;
import java.applet.*;

/**
 * This is the main view. It shows the game map, manages the sprites
 * and handles the sounds and animations. The Game events are passed
 * from the Game to the GameMapPanel via the Observer-Observable
 * pattern.
 *
 * @author	Virtu Halttunen
 * @version	0.1
 */
public class GameMapPanel extends JPanel implements Scrollable, Observer
{
    public final static int LIGHT_LEVELS = 32;
    public static boolean playSounds = true;
    private boolean drawShootingHints = false;
    private boolean alwaysShowMonsters = false;
    private boolean drawPlayerHint = true;
    private Rectangle highlightArea = null;
    private int slowifyRate = 1;

    private Game game;
    private Map map;
    private Dimension viewportSize;
    private int cellSize;
    private Vector floorCells;
    private Vector wallCells;
    private Image [][] doorCells;
    private Image [] healthIcons;
    private Image noAPIcon;
    private Image screenIcon = null;
    private Point screenIconLocation = new Point( 0, 0 );
    private Vector objectCells;
    private Color [] lights;
    private Color highlightColor = Color.green;
    private Color hasViewColor = Color.yellow;
    private Color noViewColor = Color.red;
    private HashMap characterSprites;
    private HashMap weaponSprites;
    private HashMap movingObjects;
    private Vector stainSprites;
    private Vector bufferedMoveActions;
    private Vector bufferedAttackActions; // includes player deaths
    private Vector bufferedDoorActions;
    private Vector bufferedPushActions;
    private int frame = 0;
    private int openDoorIndex = 0;
    private int closedDoorIndex = 0;
    private int doorFrame = -1;
    private AudioClip doorClip;
    private AudioClip boom;
    private AudioClip miss;
    private Point targetReticleLocation = new Point( -1, -1 );
    private Image targetReticle;
    private int reticleOffset = 0;
    private ProgressMonitor monitor;
    private WeaponSprite weaponEffect = null;
    private int weaponFrame = -1;
    private BufferedImage messageImage = null;
    private long lastMessageTime = 0;
    private Vector previousMessages;
    private int currentMessageIndex = 0;
    private Rectangle messageImageBounds = new Rectangle( 0, 0, 0, 0 );
    private boolean animating = false;
        
    public GameMapPanel( Game game, int cellSize, int visibleHorizontal, int visibleVertical )
    {
	super();
	map = game.getMap();
	this.game = game;
        this.game.addObserver( this );
	this.cellSize = cellSize;
        monitor = new ProgressMonitor( this, "Loading...", "Loading sprites", 0, 10 ); // Progress
        int progress = 0;
	viewportSize = new Dimension( visibleHorizontal*cellSize, visibleVertical*cellSize );
	Dimension mySize = new Dimension ( map.getHorizontalSize()*cellSize, map.getVerticalSize()*cellSize );
	setMinimumSize( mySize );
	setPreferredSize( mySize );
	setMaximumSize( mySize );
	floorCells = new Vector();
	wallCells = new Vector();
	objectCells = new Vector();
        bufferedMoveActions = new Vector();
        bufferedAttackActions = new Vector();
        bufferedDoorActions = new Vector();
	bufferedPushActions = new Vector();
	Character character;
	String characterClass;
	CharacterSprite newSprite;
        characterSprites = new HashMap( game.getMonsters().size()*3, 0.75f );
        weaponSprites = new HashMap();
	movingObjects = new HashMap();
	previousMessages = new Vector();
        ResourceManager.setObserver( this );
        monitor.setMaximum( monitor.getMaximum() + game.getPlayerCharacters().size() + game.getMonsters().size() );
	for( int i = 0; i < game.getPlayerCharacters().size(); i++ )
	{
            character = game.getPlayerCharacter( i );
            addSpriteForCharacter( character );
            addSpriteForWeapon( ((Marine)character).getWeapon() );
            monitor.setProgress( ++progress );
	}
        for( int i = 0; i < game.getMonsters().size(); i++ )
	{
            character = game.getMonster( i );
            addSpriteForCharacter( character );
            monitor.setProgress( ++progress );
        }
        targetReticle = ResourceManager.loadImage( "TargetReticle.png" );
	stainSprites = new Vector();
        monitor.setNote( "Loading map... " );
	loadFloorsFromMap();
        monitor.setProgress( progress+=2 );
	loadWallsFromMap();
        monitor.setProgress( progress+=2 );
	loadDoorsFromMap();
        monitor.setProgress( progress+=2 );
	loadObjectsFromMap();
        monitor.setProgress( progress+=2 );
        loadIcons();
	lights = new Color[ LIGHT_LEVELS ];
	for( int i = 0; i < LIGHT_LEVELS; i++ )
	{
	    lights[ i ] = new Color( 1, 1, 1, i*(255/LIGHT_LEVELS));
	}
        monitor.setNote( "Loading sounds... " );
        if( playSounds )
        {
            doorClip = ResourceManager.loadAudioClip( "door.wav" );
            boom = ResourceManager.loadAudioClip( "Tsup.wav" );
            miss = ResourceManager.loadAudioClip( "miss.wav" );
            System.out.println( "Intialized sound effects" );
        }
        monitor.setProgress( progress+=2 );
        monitor.close();
	//setCursor( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
        //setDoubleBuffered( false );
    }
    
    protected void paintComponent( Graphics g )
    {
	//super.paintComponent( g );
        long startTime = System.currentTimeMillis();
        MapCell cell;
	CharacterSprite sprite;
	Sprite stain;
        Sprite objective;
	GameAction action;
        Point door;
	boolean doDrawShootingHints = ( drawShootingHints && game.getState() == Game.PLAYER_MOVE );
	
	//System.out.println( "MapPanel: repaint" );
	Rectangle newRect = g.getClipBounds();//getVisibleRect();
	Point min = getCell( new Point( newRect.x, newRect.y ) );
	int minX = min.x;
	int minY = min.y;
	Point max = getCell( new Point( newRect.x+newRect.width, newRect.y+newRect.height ) );
	int maxX = max.x;
	int maxY = max.y;
	
	if( maxX == map.getHorizontalSize() ) maxX = maxX-1;
	if( maxY == map.getVerticalSize() ) maxY = maxY-1;

        int numCells = (maxX-minX)*(maxY-minY);
        
	for( int i = minX; i <= maxX; i++ )
	{
	    for( int j = minY; j <= maxY; j++ )
	    {
                g.setColor( Color.black );
                g.fillRect( i*cellSize, j*cellSize, cellSize, cellSize );
                cell = map.getCell( i, j );
                if( cell != null )
                {
                    int floor;
                    for( int f = 0; f < cell.getFloorSize(); f++ )
                    {
                        floor = cell.getFloor( f );
                        if( floor >= 0 )
			{
                            g.drawImage( (Image)floorCells.elementAt( floor ), i*cellSize, j*cellSize, cellSize, cellSize, this );
			}
		    }
                }
            }
	}
	// Draw stains
	for( int i = 0; i < stainSprites.size(); i++ )
	{
	    stain = (Sprite)stainSprites.elementAt( i );
	    if( newRect.intersects( stain.getBounds() ) && stain.isVisible() || alwaysShowMonsters )
	    {
		Image stainImage = stain.getSprite();
		if( stainImage != null )
		{
		    g.drawImage( stainImage, stain.getBounds().x, stain.getBounds().y, this );
		}
	    }
	}	
	if( highlightArea != null )
	{
	    g.setColor( Color.red );
	    g.fillRect( highlightArea.x, highlightArea.y, highlightArea.width, highlightArea.height );
	}
        // Draw objects
	for( int i=0; i < map.getNumberOfObjects(); i++ )
	{
	    MapObject mapObject = map.getObject( i );
	    MapObjectCell [] cells = mapObject.getCells();
	    if( cells != null )
	    {
		for( int j=0; j < cells.length; j++ )
		{
		    int x = cells[j].getX();
		    int y = cells[j].getY();
		    if( x >= minX && x <= maxX && y >= minY && y <= maxY ) 
		    {
			int xLocation = x*cellSize;
			int yLocation = y*cellSize;
			if( movingObjects.containsKey( mapObject ) )
			{
			    Point offset = (Point)movingObjects.get( mapObject );
			    xLocation += offset.x;
			    yLocation += offset.y;
			}
			int objIndex = cells[j].getType();
			/*
			if( cells[j].isDestroyed() )
			{
			    objIndex = cells[j].getDestroyedType();
			    if( objIndex < 0 )
			    {
				objIndex = cells[j].getType();
			    }
			}*/
			if( !cells[j].isDestroyed() )
			{
			    g.drawImage( (Image)objectCells.elementAt( objIndex ), xLocation, yLocation, cellSize, cellSize, this );
		    
			}
		    }
		}
	    }
	}
	
	// Draw objectives
	for( int i = minX; i <= maxX; i++ )
	    for( int j = minY; j <= maxY; j++ )
	    {
                for( int k=game.getNumberOfObjectives()-1; k >= 0; k-- )
                {
                    if( game.getObjective(k).getNumberOfSprites() > 0 )
                    {
                        if( game.getObjective(k).isLocatedIn( i, j ) )
                        {
                            objective = game.getObjective(k).getSprite( i, j );
                            
                            // draw marker under monsters when in shooting mode
                            if( doDrawShootingHints && game.getObjective(k).isDestructible() )
                            {
                                if( game.isCellVisibleToCharacter( game.getActiveCharacter(), new Point( i, j ) ) )
                                {
                                    g.setColor( hasViewColor );
                                }
                                else
                                {
                                    g.setColor( noViewColor );
                                }
                                g.drawRect( i*cellSize, j*cellSize, cellSize-1, cellSize-1 );
                                g.drawRect( i*cellSize+1, j*cellSize+1, cellSize-3, cellSize-3 );
    
                            }
                            if( objective != null )
                            {
				Image objectiveSprite = objective.getSprite();
				if( objectiveSprite != null )
				{
				    g.drawImage( objectiveSprite, i*cellSize+objective.getXOffset(), j*cellSize+objective.getYOffset(), this );
				}
				else
				{
				    System.err.println( "Null sprite drawing " + game.getObjective( k ).getClassName() );
				}
                            }
                        }
                    }		    
                }
            }
        
	// Draw monster sprites
	for( int i=0; i < game.getMonsters().size(); i++ )
	{
	    sprite = (CharacterSprite)characterSprites.get( game.getMonster( i ) );;
	    action = sprite.getAction();
	    if( newRect.intersects( sprite.getBounds() ) )
	    {
		if( sprite.isVisible() || alwaysShowMonsters )
		{
		    // draw marker under monsters when in shooting mode
		    if( doDrawShootingHints )
		    {
			if( game.isCellVisibleToCharacter( game.getActiveCharacter(), game.getMonster( i ).getLocation() ) )
			{
			    g.setColor( hasViewColor );
			}
			else
			{
			    g.setColor( noViewColor );
			}
			g.drawRect( sprite.getVisualBounds().x, sprite.getVisualBounds().y, sprite.getVisualBounds().width-1, sprite.getVisualBounds().height-1 );
			g.drawRect( sprite.getVisualBounds().x+1, sprite.getVisualBounds().y+1, sprite.getVisualBounds().width-3, sprite.getVisualBounds().height-3 );
		    }
		    Image image = sprite.getSprite( action );
		    if( image != null )
		    {
			g.drawImage( image, sprite.getLocation().x, sprite.getLocation().y, this );
		    }
		}
	    }
	}
	// Draw player sprites
	for( int i=0; i < game.getPlayerCharacters().size(); i++ )
	{
	    sprite = (CharacterSprite)characterSprites.get( game.getPlayerCharacter( i ) );;
	    action = sprite.getCharacter().getLastAction();
	    if( newRect.intersects( sprite.getBounds() ) )
	    {
		if( game.isActiveCharacter( i ) && drawPlayerHint )
		{
		    g.setColor( highlightColor );
		    g.drawRect( sprite.getVisualBounds().x, sprite.getVisualBounds().y, sprite.getVisualBounds().width-1, sprite.getVisualBounds().height-1 );
		    g.drawRect( sprite.getVisualBounds().x+1, sprite.getVisualBounds().y+1, sprite.getVisualBounds().width-3, sprite.getVisualBounds().height-3 );
		}
		if( sprite.getCharacter().isAlive() )
		    g.drawImage( sprite.getSprite( action ), sprite.getLocation().x, sprite.getLocation().y, this );
	    }
	}
        // Draw walls
        for( int i = minX; i <= maxX; i++ )
            for( int j = minY; j <= maxY; j++ )
            {
                cell = map.getCell( i, j );
                if( cell != null )
                {
                    int wall = cell.getWall();
                    if( wall >= 0 )
                    {
                        g.drawImage( (Image)wallCells.elementAt( wall ), i*cellSize, j*cellSize, cellSize, cellSize, this );
                    }
                    else
                    {
                        // If it's a door
                        if( wall < -1 )
                        {
                            drawDoor( g, i, j );
                        }
                    }
                }
            }
        // Draw target reticle
        if( targetReticleLocation.x > -1 && targetReticleLocation.y > -1 )
        {
            g.drawImage( targetReticle, targetReticleLocation.x*cellSize+reticleOffset, targetReticleLocation.y*cellSize+reticleOffset, this );
        }
        // Draw weapon effects
        if( weaponEffect != null && weaponFrame > -1 )
        {
            weaponEffect.drawFrame( g, this, weaponFrame );
        }
	// Draw lights
	for( int i = minX; i <= maxX; i++ )
	    for( int j = minY; j <= maxY; j++ )
	    {
		g.setColor( lights[ map.getLight( i, j ) ] );
		g.fillRect( i*cellSize, j*cellSize, cellSize, cellSize );
	    }
        // Draw interface icons
        if( screenIcon != null )
        {
	    //System.out.println( "Drawing screenIcon" );
            g.drawImage( screenIcon, screenIconLocation.x, screenIconLocation.y, this );
        }
	if( messageImage != null )
	{
	    Rectangle rect = getVisibleRect();
	    g.drawImage( messageImage, rect.x+10, rect.y+rect.height-messageImageBounds.y, this );
	}
	//Toolkit.getDefaultToolkit().sync();
        long endTime = System.currentTimeMillis();
        //System.out.println( "Drawing " + numCells + " cells took " + (int)(endTime-startTime) + " ms" );
    }

    private void drawDoor( Graphics g, int x, int y )
    {
        Door theDoor = map.getDoor( x, y );
        Point door1;
        Point door2;
        int doorAnimationFrame = closedDoorIndex;

        if( theDoor != null )
        {
            int dir = theDoor.getDirection();
            int dIndex = 0;
            if( dir == Map.NORTH ) dIndex = 1;
            if( dir == Map.SOUTH ) dIndex = 0;
            if( dir == Map.EAST ) dIndex = 2;
            if( dir == Map.WEST ) dIndex = 3;
            
            // Draw the door
            if( theDoor.isOpen() )
                doorAnimationFrame = openDoorIndex;
            else
                doorAnimationFrame = closedDoorIndex;

            // If there's a door animation that we have to draw, check it
            if( doorFrame > -1 )
            {
                for( int i = 0; i < bufferedDoorActions.size(); i++ )
                {
                    door1 = ((DoorAction)bufferedDoorActions.elementAt( i )).getDoor1();
                    door2 = ((DoorAction)bufferedDoorActions.elementAt( i )).getDoor2();
                    if( ( door1.x == x && door1.y == y ) || ( door2.x == x && door2.y == y ) )
                    {
                        // This door is animating
                        if( ((DoorAction)bufferedDoorActions.elementAt( i )).isOpening()  )
                        {
                            doorAnimationFrame = doorFrame;
                        }
                        else
                        {
                            doorAnimationFrame = openDoorIndex-doorFrame;
                        }
                    }
                }
            }
            g.drawImage( doorCells[ dIndex ][ doorAnimationFrame ], x*cellSize, y*cellSize, cellSize, cellSize, this );
        }
    }
     
    public boolean animateGameAction( GameAction action )
    {
        if( action instanceof MoveAction )
        {
            //animateMoveAction( (MoveAction)action );
	    animateBufferedMoveActions();
        }
        else if( action instanceof MeleeAction )
        {
            animateMeleeAction( (MeleeAction)action );
        }
        else if( action instanceof ShootAction )
        {
            animateShootAction( (ShootAction)action );
        }
        else if( action instanceof UseAction )
        {
            animateUseAction( (UseAction)action );
        }
        else if( action instanceof DeadAction )
        {
            CharacterSprite sprite = getSpriteFor( action.getActor() );
            if( getVisibleRect().intersects( sprite.getBounds() ) )
            {
                RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
            }
         
        }
        return true;
     }
    
    /*
    public boolean animateMoveAction( MoveAction action )
    {
        long startTime = System.currentTimeMillis();
        long elapsedTime;
        int pauseTime;
        Character character = action.getActor();
        CharacterSprite sprite = (CharacterSprite)characterSprites.get( character );
	sprite.setAction( action );
	    
        //System.out.println( "Animating walk for " + sprite.getClass().getName() + ", frames: " + sprite.getNumberOfWalkFrames() );
        int x, y;
        int oldX = action.getStartPoint().x;
        int oldY = action.getStartPoint().y;
        int newX = action.getEndPoint().x;
        int newY = action.getEndPoint().y;
    
        if( game.isCellVisibleToPlayer( action.getStartPoint() ) || game.isCellVisibleToPlayer( action.getEndPoint() ) )
        {
	    if( character.isAlive() )
	    {
		sprite.setVisible( true );
	    }
            //System.out.println( "Double buffered: " + isDoubleBuffered() );
            for( int i=0; i < sprite.getNumberOfWalkFrames(); i++ )
            {
                elapsedTime = System.currentTimeMillis();
                x = oldX*cellSize+(int)( Math.sin(Math.PI/2.0*(double)i/(double)sprite.getNumberOfWalkFrames())*(newX*cellSize-oldX*cellSize) );
                y = oldY*cellSize+(int)( Math.sin(Math.PI/2.0*(double)i/(double)sprite.getNumberOfWalkFrames())*(newY*cellSize-oldY*cellSize) );
                //System.out.println( "Sprite location " + x +"," + y );
                sprite.setLocation( x, y );
                sprite.setDirection( character.getDirection() );
                
                RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
                RepaintManager.currentManager( this ).paintDirtyRegions();
                try
                {
                    elapsedTime = System.currentTimeMillis()-elapsedTime;
                    //System.out.println( "Elapsed time " + (int)elapsedTime + " ms" );
                    pauseTime = (120/sprite.getNumberOfWalkFrames())-(int)elapsedTime;
                    if( pauseTime >= 0 )
                    {
                        //System.out.println( "Pausing... " );
                        Thread.sleep( pauseTime );
                    }
                }
                catch( InterruptedException e )
                {
                }
            }
            long endTime = System.currentTimeMillis();
            System.out.println( "Animation took " + (int)(endTime-startTime) + " ms" );
            return true;
        }
        else
        {
            sprite.setVisible( false );
            sprite.setLocation( newX*cellSize, newY*cellSize );
            sprite.setDirection( character.getDirection() );
        }
        long endTime = System.currentTimeMillis();
        System.out.println( "Animation took " + (int)(endTime-startTime) + " ms" );
        return false;
    }
     */
    
    public boolean animateBufferedMoveActions()
    {
	animating = true;
        long startTime = System.currentTimeMillis();
        long elapsedTime = 0;
        int pauseTime;
        MoveAction action;
        Character character;
        CharacterSprite sprite;
        int x, y, oldX, oldY, newX, newY;
        boolean needsPainting = false;
	int maxFrame = 10;
	int theFrame = 0;
        
        // Check visibility and set locations for non-visible sprites
        for( int i = 0; i < bufferedMoveActions.size(); i++ )
        {
            action = (MoveAction)bufferedMoveActions.elementAt( i );
            sprite = (CharacterSprite)characterSprites.get( action.getActor() );
	    sprite.setAction( action );
	    maxFrame = Math.max( maxFrame, sprite.getNumberOfWalkFrames() );
            sprite.setVisible( game.isCellVisibleToPlayer( action.getStartPoint() ) || game.isCellVisibleToPlayer( action.getEndPoint() ) || alwaysShowMonsters );
	    if( !action.getActor().isAlive() )
	    {
		sprite.setVisible( false );
	    }
            if( !sprite.isVisible() )
            {
                sprite.setLocation( action.getEndPoint().x*cellSize, action.getEndPoint().y*cellSize );
                sprite.setDirection( action.getActor().getDirection() );
            }
        }
        // Animate visible sprites
        for( int f = 0; f < maxFrame; f++ )
        {
            for( int i = 0; i < bufferedMoveActions.size(); i++ )
            {
                elapsedTime = System.currentTimeMillis();
                action = (MoveAction)bufferedMoveActions.elementAt( i );
                character = action.getActor();
                sprite = (CharacterSprite)characterSprites.get( character );
                if( sprite.isVisible() )
                {
		    theFrame = (int)( sprite.getNumberOfWalkFrames() * (f/(double)maxFrame) );

                    oldX = action.getStartPoint().x;
                    oldY = action.getStartPoint().y;
                    newX = action.getEndPoint().x;
                    newY = action.getEndPoint().y;
                    x = oldX*cellSize+(int)( Math.sin(Math.PI/2.0*(double)f/(double)maxFrame )*(newX*cellSize-oldX*cellSize) );
                    y = oldY*cellSize+(int)( Math.sin(Math.PI/2.0*(double)f/(double)maxFrame )*(newY*cellSize-oldY*cellSize) );
                    //System.out.println( "Sprite location " + x +"," + y );
                    sprite.setLocation( x, y );
		    sprite.setFrame( theFrame );
                    sprite.setDirection( character.getDirection() );
                    
                    if( getVisibleRect().intersects( sprite.getBounds() ) )
                    {
                        needsPainting = true;
                        RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
                    }
                }
            }
            if( needsPainting )
            {
                RepaintManager.currentManager( this ).paintDirtyRegions();
                try
                {
                    elapsedTime = System.currentTimeMillis()-elapsedTime;
                    //System.out.println( "Elapsed time " + (int)elapsedTime + " ms" );
                    pauseTime = (12)-(int)elapsedTime;
                    if( pauseTime >= 0 )
                        Thread.sleep( pauseTime );
                }
                catch( InterruptedException ie )
                {
                }
            }
               
        }
        // Clear the animation buffer
        bufferedMoveActions.clear();
        long endTime = System.currentTimeMillis();
        System.out.println( "Animation took " + (int)(endTime-startTime) + " ms" );
	animating = false;
        return true;
    }
    
    public boolean animateMeleeAction( MeleeAction action )
    {
        Character attacker = action.getAttacker();
        CharacterSprite sprite = (CharacterSprite)characterSprites.get( attacker );
	sprite.setAction( action );
        Character defender = action.getDefender();
	CharacterSprite defenderSprite = (CharacterSprite)characterSprites.get( defender );
	    
        //System.out.println( "Animating walk for " + sprite.getClass().getName() + ", frames: " + sprite.getNumberOfWalkFrames() );
        int x, y;
        int oldX = attacker.getLocation().x;
        int oldY = attacker.getLocation().y;
        int newX = defender.getLocation().x;
        int newY = defender.getLocation().y;
	int tx, ty;
	int maxFrame=0;
	int theFrame=0;
    
        if( game.isCellVisibleToPlayer( attacker.getLocation() ) || alwaysShowMonsters )
        {
            sprite.setVisible( true );
            if( !getVisibleRect().intersects( sprite.getBounds() ) )
            {
                centerOnCharacter( attacker );
                try
                {
                    Thread.sleep( 100 );
                }
                catch( InterruptedException ie )
                {
                }
            }

	    int direction = Map.getDirection( attacker.getLocation(), defender.getLocation(), attacker.getDirection() );
	    if( action.getTargetDefends() )
	    {
		maxFrame = Math.max( sprite.getNumberOfMeleeAttackFrames(), defenderSprite.getNumberOfMeleeAttackFrames() );
	    }
	    else
	    {
		maxFrame = sprite.getNumberOfMeleeAttackFrames();
	    }
            for( int i=0; i < maxFrame; i++ )
            {
		theFrame = (int)( sprite.getNumberOfMeleeAttackFrames() * (i/(double)maxFrame) );
		x = oldX*cellSize+(int)( 0.5*Math.sin(Math.PI*(double)i/(double)(maxFrame-1))*(newX*cellSize-oldX*cellSize) );
		y = oldY*cellSize+(int)( 0.5*Math.sin(Math.PI*(double)i/(double)(maxFrame-1))*(newY*cellSize-oldY*cellSize) );
                //System.out.println( "Sprite location " + x +"," + y );
                sprite.setLocation( x, y );
                sprite.setDirection( direction );
		sprite.setFrame( theFrame );
		
		if( action.getTargetDefends() )
		{
		    theFrame = (int)( defenderSprite.getNumberOfMeleeAttackFrames() * (i/(double)maxFrame) );
		    tx = newX*cellSize+(int)( 0.5*Math.sin(Math.PI*(double)i/(double)(maxFrame-1))*(oldX*cellSize-newX*cellSize) );
                    ty = newY*cellSize+(int)( 0.5*Math.sin(Math.PI*(double)i/(double)(maxFrame-1))*(oldY*cellSize-newY*cellSize) );
		    //System.out.println( "Sprite location " + x +"," + y );
		    defenderSprite.setLocation( tx, ty );
		    defenderSprite.setDirection( (direction+4)%8 );
		    defenderSprite.setFrame( theFrame );
		}
		
                RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
		RepaintManager.currentManager( this ).addDirtyRegion( this, defenderSprite.getLocation().x-cellSize, defenderSprite.getLocation().y-cellSize, defenderSprite.getBounds().width+cellSize*2, defenderSprite.getBounds().height+cellSize*2 );
		RepaintManager.currentManager( this ).paintDirtyRegions();
                try
                {
                    Thread.sleep( 250/sprite.getNumberOfMeleeAttackFrames() );
                }
                catch( InterruptedException ie )
                {
                }
            }
            return true;
        }
        else
        {
            sprite.setVisible( false );
        }
        return false;
    }

    public void animateHealAction( HealAction action )
    {
        //System.out.println( "Animating HealAction" );
        Character character = action.getActor();
        CharacterSprite sprite = (CharacterSprite)characterSprites.get( character );
	sprite.setAction( action );

        int status = action.getStatus();
        if( status == HealAction.HEAL ) { screenIcon = healthIcons[ 0 ]; }
        if( status == HealAction.FULL_HEALTH ) { screenIcon = healthIcons[ 1 ]; }
        if( status == HealAction.ERROR ) { screenIcon = healthIcons[ 2 ]; }
        if( status == HealAction.FUMBLE ) { screenIcon = healthIcons[ 3 ]; }

        int x, y;
        int oldX = character.getLocation().x;
        int oldY = character.getLocation().y;


        for( int i=0; i < 10; i++ )
        {
            x = oldX*cellSize;
            y = oldY*cellSize+(int)( 0.5*Math.sin(Math.PI*(double)i/(double)10)*(-24) );
            //System.out.println( "Sprite location " + x +"," + y );
            screenIconLocation = new Point( x, y );

            RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
            RepaintManager.currentManager( this ).paintDirtyRegions();
            try
            {
                Thread.sleep( 500/10 );
            }
            catch( InterruptedException ie )
            {
            }
        }
        screenIconLocation = new Point( 0, 0 );
        screenIcon = null;
    }
    
    public void animateNoAPAction( NoAPAction action )
    {
        //System.out.println( "Animating NoAPAction" );
        Character character = action.getActor();
        CharacterSprite sprite = (CharacterSprite)characterSprites.get( character );
	sprite.setAction( new MoveAction( character, character.getLocation(), character.getLocation() ) );
	
        screenIcon = noAPIcon;
	
        int x, y;
        int oldX = character.getLocation().x;
        int oldY = character.getLocation().y;
	Rectangle rect = getVisibleRect();
	if( !rect.contains( oldX*cellSize, oldY*cellSize ) )
	{
	    oldX = (rect.x+rect.width/2)/cellSize;
	    oldY = (rect.y+rect.height/2)/cellSize;
	}
	
	
        for( int i=0; i < 10; i++ )
        {
            x = oldX*cellSize;
            y = oldY*cellSize+(int)( 0.5*Math.sin(Math.PI*(double)i/(double)10)*(-cellSize) );
            //System.out.println( "Sprite location " + x +"," + y );
            screenIconLocation = new Point( x, y );
	    
            RepaintManager.currentManager( this ).addDirtyRegion( this, oldX*cellSize-cellSize, oldY*cellSize-cellSize, oldX*cellSize+cellSize*3, oldY*cellSize+cellSize*3 );
            RepaintManager.currentManager( this ).paintDirtyRegions();
            try
            {
                Thread.sleep( 500/10 );
            }
            catch( InterruptedException ie )
            {
            }
        }
        screenIconLocation = new Point( 0, 0 );
        screenIcon = null;
	RepaintManager.currentManager( this ).addDirtyRegion( this, oldX*cellSize-cellSize, oldY*cellSize-cellSize, oldX*cellSize+cellSize*3, oldY*cellSize+cellSize*3 );
	RepaintManager.currentManager( this ).paintDirtyRegions();
    }
    
    public void animateBufferedPushActions()
    {
	PushObjectAction pushAction;
	MapObject mapObject;
	int x=0, y=0, oldX, oldY, newX, newY;
	int maxFrame = 10;
	if( bufferedPushActions.size() > 0 )
	{
	    for( int f=0; f < maxFrame; f++ )
	    {
		for( int o=0; o < bufferedPushActions.size(); o++ )
		{
		    pushAction = (PushObjectAction)bufferedPushActions.elementAt( o );
		    mapObject = pushAction.getObject();
		    oldX = pushAction.getStartPoint().x;
		    oldY = pushAction.getStartPoint().y;
		    newX = pushAction.getEndPoint().x;
		    newY = pushAction.getEndPoint().y;
		    System.out.println( "Pushing from " + oldX + ", " + oldY + " to " + newX + ", " + newY );
		    System.out.println( oldX*cellSize + ", " + oldY*cellSize + " to " + newX*cellSize + ", " + newY*cellSize );
		    if( !pushAction.isBounce() )
		    {
			x = (int)( Math.sin(Math.PI/2.0*(double)f/(double)maxFrame )*(newX*cellSize-oldX*cellSize) );
			y = (int)( Math.sin(Math.PI/2.0*(double)f/(double)maxFrame )*(newY*cellSize-oldY*cellSize) );
			// These are needed because the object has already been moved when we get to actually animate it.
			// They compensate for the actual location
			x+=(newX-oldX)*-cellSize;
			y+=(newY-oldY)*-cellSize;
		    }
		    else
		    {
			x = (int)( 0.5*Math.sin(Math.PI*(double)f/(double)(maxFrame-1))*(newX*cellSize-oldX*cellSize) );
			y = (int)( 0.5*Math.sin(Math.PI*(double)f/(double)(maxFrame-1))*(newY*cellSize-oldY*cellSize) );
			if( pushAction.getActor() != null )
			{
			    Character character = pushAction.getActor();
			    CharacterSprite sprite = (CharacterSprite)characterSprites.get( pushAction.getActor() );
			    sprite.setDirection( Map.getDirection( pushAction.getStartPoint(), pushAction.getEndPoint(), character.getDirection() ) );
			    Point location = character.getLocation();
			    sprite.setLocation( (int)(location.x*cellSize+x*1.5), (int)(location.y*cellSize+y*1.5) );
			}
		    }
		    int minX=0, minY=0;
		    System.out.println( "Object from: " + mapObject.getTopLeft().x + ", " + mapObject.getTopLeft().y + " to " + mapObject.getBottomRight().x + ", " + mapObject.getBottomRight().y );
		    minX = mapObject.getTopLeft().x*cellSize;
		    minY = mapObject.getTopLeft().y*cellSize;
		    int width = cellSize*mapObject.getWidth();
		    int height = cellSize*mapObject.getHeight();
		    // Surround the area with space
		    minX-=cellSize;
		    minY-=cellSize;
		    width+=cellSize*2;
		    height+=cellSize*2;
		    int direction = Map.getDirection( pushAction.getStartPoint(), pushAction.getEndPoint(), 0 );
		    if( pushAction.isBounce() )
		    {
			if( direction == Map.SOUTH || direction == Map.SOUTH_EAST || direction == Map.SOUTH_WEST )
			{
			    //minY-=cellSize;
			    //height+=cellSize;
			}
			if( direction == Map.NORTH || direction == Map.NORTH_EAST || direction == Map.NORTH_WEST )
			{
			    height+=cellSize;
			}
		    }
		    else
		    {
			// Compensate for the object actually having been already moved
			if( direction == Map.WEST || direction == Map.NORTH_WEST || direction == Map.SOUTH_WEST )
			{
			    minX+=cellSize;
			    //width+=cellSize;
			}
			if( direction == Map.EAST || direction == Map.NORTH_EAST || direction == Map.SOUTH_EAST )
			{
			    minX-=cellSize;
			}
			if( direction == Map.NORTH || direction == Map.NORTH_EAST || direction == Map.NORTH_WEST )
			{
			    //minY+=cellSize;
			    //height+=cellSize;
			}
			if( direction == Map.SOUTH || direction == Map.SOUTH_EAST || direction == Map.SOUTH_WEST )
			{
			    //minY-=cellSize;
			    //height+=cellSize;
			}
		    }
		    System.out.println( minX + ", " + minY + ",  " + width + " x " + height );
		    //highlightArea = new Rectangle( minX, minY, width, height );
		    RepaintManager.currentManager( this ).addDirtyRegion( this, minX, minY, width, height );
		    movingObjects.put( mapObject, new Point( x, y ) );
		}
		RepaintManager.currentManager( this ).paintDirtyRegions();	
		try
		{
		    Thread.sleep( 500/10 );
		}
		catch( InterruptedException ie )
		{
		}		
	    }
	    bufferedPushActions.clear();
	    movingObjects.clear();
	}
    }
    
    public boolean animateBufferedAttackActions()
    {
//	animating = true;
        GameAction action;
        //System.out.println( "Animating attack actions: " + bufferedAttackActions.size() );
        for( int i=0; i < bufferedAttackActions.size(); i++ )
        {
	    //System.out.println( "Animating action: " + i );
            action = (GameAction)bufferedAttackActions.elementAt( i );
	    //System.out.println( action.getClass().getName() );
            if( action instanceof MeleeAction )
            {
                animateMeleeAction( (MeleeAction)action );
            }
            if( action instanceof ShootAction )
            {
		//System.out.println( "Going to animate shoot attack to " + Objective.pointToString( ((ShootAction)action).getEndPoint() ) );
                animateShootAction( (ShootAction)action );
            }
            if( action instanceof NewCorpseAction )
            {
		CharacterSprite sprite = (CharacterSprite)characterSprites.get( action.getActor() );
		if( !action.getActor().isAlive() )
		{
		    sprite.setVisible( false );
		}
                addCorpse( ((NewCorpseAction)action).getActor() );
		RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
            }
            if( action instanceof WoundAction )
            {
                CharacterSprite sprite = (CharacterSprite)characterSprites.get( action.getActor() );
		sprite.setAction( action );
                RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
            }
        }
        bufferedAttackActions.clear();
        RepaintManager.currentManager( this ).paintDirtyRegions();
//	animating = false;
        return true;
    }
    
    public boolean animateShootAction( ShootAction action )
    {
	//System.out.println( "Animating shoot attack to " + action.getEndPoint() );

        // Paint old dirt first
        RepaintManager.currentManager( this ).paintDirtyRegions();

        Character character = action.getActor();
        CharacterSprite sprite = (CharacterSprite)characterSprites.get( character );
	sprite.setAction( action );
        
        if( game.getState() == Game.MONSTER_MOVE )
        {
	    if( character.isHostile( game.getPlayerCharacter( 0 ) ) && !game.isCellVisibleToPlayer( character.getLocation() ) && !game.isCellVisibleToPlayer( action.getEndPoint() ) )
	    {
		return false;
	    }
	    
            if( !getVisibleRect().intersects( sprite.getBounds() ) )
            {
                centerOnCharacter( character );
                try
                {
                    Thread.sleep( 100 );
                }
                catch( InterruptedException ie )
                {
                }
            }
        }
        Weapon weapon = action.getWeapon();
        if( !weaponSprites.containsKey( weapon.getClass().getName() ) )
        {
            // Add the weapon sprite
            addSpriteForWeapon( weapon );
        }
        weaponEffect = (WeaponSprite)weaponSprites.get( weapon.getClass().getName() );
        weaponEffect.setWeapon( weapon );
        
        int x = character.getLocation().x*cellSize;
        int y = character.getLocation().y*cellSize;
        sprite.setLocation( x, y );
        sprite.setDirection( character.getDirection() );

        if( playSounds )
        {
            AudioClip sound = weaponEffect.getFireSoundEffect();
            if( sound != null )
                sound.play();
        }

        for( int i = 0; i < sprite.getNumberOfRangedAttackFrames(); i++ )
        {
            sprite.setFrame( i );
            try
            {
                RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
                Thread.sleep( 50*slowifyRate );
            }
            catch( InterruptedException ie )
            {
            }
            RepaintManager.currentManager( this ).paintDirtyRegions();
        }
        
        //RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );

        for( weaponFrame = 0; weaponFrame < weaponEffect.getFrames(); weaponFrame++ )
        {
            weaponEffect.prepareFrame( character.getLocation(), action.getEndPoint(), weaponFrame, this );
            try
            {
                Thread.sleep( 50*slowifyRate );
            }
            catch( InterruptedException ie )
            {
            }
            RepaintManager.currentManager( this ).paintDirtyRegions();
        }

        //RepaintManager.currentManager( this ).addDirtyRegion( this, action.getEndPoint().x*cellSize-cellSize, action.getEndPoint().y*cellSize-cellSize, cellSize*3, cellSize*3 );
        //RepaintManager.currentManager( this ).paintDirtyRegions();

        if( !action.getHit() && playSounds )
        {
            AudioClip sound = weaponEffect.getMissSoundEffect();
            if( sound != null )
                sound.play();
        }
        weaponEffect = null;
        weaponFrame = -1;
	slowifyRate = 1;
        RepaintManager.currentManager( this ).addDirtyRegion( this, action.getEndPoint().x*cellSize-cellSize, action.getEndPoint().y*cellSize-cellSize, cellSize*3, cellSize*3 );
        return true;
    }
    
    public boolean animateUseAction( UseAction action )
    {
        Character character = action.getActor();
        CharacterSprite sprite = (CharacterSprite)characterSprites.get( character );
	sprite.setAction( action );
        
        int x = character.getLocation().x*cellSize;
        int y = character.getLocation().y*cellSize;
        sprite.setLocation( x, y );
        sprite.setDirection( character.getDirection() );
        
        RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
        RepaintManager.currentManager( this ).paintDirtyRegions();
        return true;
    }
    
    
    public void animateString( String string )
    {
	animating = true;
	// Clear possible old message
	Rectangle rect = getVisibleRect();
	RepaintManager.currentManager( this ).addDirtyRegion( this, rect.x, rect.y+rect.height-(messageImageBounds.y+messageImageBounds.height+10), messageImageBounds.width+10, messageImageBounds.y+messageImageBounds.height+20 );

	createString( string );
	messageImageBounds.y=0;

	for( int i = 0; i < 10; i++ )
	{
	    RepaintManager.currentManager( this ).addDirtyRegion( this, rect.x, rect.y+rect.height-(messageImageBounds.y+messageImageBounds.height+10), messageImageBounds.width+10, messageImageBounds.y+messageImageBounds.height+20 );
	    messageImageBounds.y=( (int)(i/10.0*(messageImageBounds.height) ) );
	    RepaintManager.currentManager( this ).paintDirtyRegions();
	    try
	    {
		Thread.sleep( 50 );
	    }
	    catch( InterruptedException ie )
	    {
	    }	    
	}
	animating = false;
    }
    
    public void animateBufferedDoorActions()
    {
        if( bufferedDoorActions.size() > 0 )
        {
            if( playSounds )
            {
                doorClip.play();
            }
            checkMonsterVisibility();
            Point door;
            for( doorFrame = closedDoorIndex; doorFrame <= openDoorIndex; doorFrame++ )
            {
                for( int i = 0; i < bufferedDoorActions.size(); i++ )
                {
                    door = ((DoorAction)bufferedDoorActions.elementAt( i )).getDoor1();
                    RepaintManager.currentManager( this ).addDirtyRegion( this, (door.x-1)*cellSize, (door.y-1)*cellSize, cellSize*3, cellSize*3 );
                }
                try
                {
                    Thread.sleep( 50 );
                    //System.out.println( "tick tock " +doorFrame );
                }
                catch( InterruptedException ie )
                {
                }
                RepaintManager.currentManager( this ).paintDirtyRegions();
                //paintImmediately( (cell.x-1)*cellSize, (cell.y-1)*cellSize, cellSize*3, cellSize*3 );
            }
            bufferedDoorActions.clear();
            doorFrame = -1;
        }
    }
    
    public boolean isAnimating()
    {
	return animating;
    }
    
    public void drawString( String string )
    {
	createString( string );
	repaint();
    }
    
    public void clearMessageString()
    {
	if( messageImage != null )
	{
	    Rectangle rect = getVisibleRect();
	    RepaintManager.currentManager( this ).addDirtyRegion( this, rect.x, rect.y+rect.height-(messageImageBounds.y+10), messageImageBounds.width+10, messageImageBounds.height+20 );
	    messageImage = null;
	    RepaintManager.currentManager( this ).paintDirtyRegions();
	}
    }
    
    private void createString( String string )
    {
	Font font = getFont();
	FontMetrics fm = getFontMetrics( font );
	StringTokenizer tokenizer = new StringTokenizer( string, "\n" );
	Vector strings = new Vector();
	int lines = tokenizer.countTokens();
	int width = 0;
	int aWidth;
	String aString;
	for( int i=0; i < lines; i++ )
	{
	    aString = tokenizer.nextToken();
	    strings.add( aString );
	    aWidth = SwingUtilities.computeStringWidth( fm, aString );
	    if( aWidth > width )
	    {
		width = aWidth;
	    }
	}
	int lineHeight = fm.getLeading() + fm.getMaxAscent() + fm.getMaxDescent();
	int height = lineHeight*lines;
	messageImage = new BufferedImage( width+10, height+10, BufferedImage.TYPE_INT_ARGB_PRE );
	System.out.println( "Drawing string of " + width + " x " + height + " with " + lines + " lines" );
	Graphics2D g2 = messageImage.createGraphics();
	g2.setFont( font );
	g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );
	g2.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON );
	FontRenderContext frc = g2.getFontRenderContext();
//	g2.setColor( Color.red );
//	g2.fillRect( 0, 0, width+10, height+10 );
	GlyphVector gv;
	Shape shape;
	g2.translate( 5, 0 );
	for( int i=0; i < strings.size(); i++ )
	{
	    aString = (String)strings.elementAt( i );
	    gv = font.createGlyphVector( frc, aString );
	    shape = gv.getOutline();
	    g2.translate( 0, lineHeight );  // Consecutive translations are cumulative
	    Stroke stroke = new BasicStroke( 2f );
	    g2.setStroke( stroke );
	    g2.setColor( Color.black );
	    g2.draw( shape );
	    stroke = new BasicStroke( 0f );
	    g2.setStroke( stroke );
	    g2.setColor( Color.white );
	    g2.fill( shape );
	}
	g2.dispose();
	messageImageBounds = new Rectangle( 0, 0, width+10, height+10 );
    }

    public Point getTargetReticleLocation()
    {
	return targetReticleLocation;
    }
    
    public void setTargetReticleLocation( Point point )
    {
	RepaintManager.currentManager( this ).addDirtyRegion( this, targetReticleLocation.x*cellSize, targetReticleLocation.y*cellSize, cellSize, cellSize );
	targetReticleLocation.x = point.x;
	targetReticleLocation.y = point.y;
	RepaintManager.currentManager( this ).addDirtyRegion( this, targetReticleLocation.x*cellSize, targetReticleLocation.y*cellSize, cellSize, cellSize );
	RepaintManager.currentManager( this ).paintDirtyRegions();
    }

    public void addStain( ImageStain stain )
    {
        stainSprites.add( stain );
        Point point = new Point( stain.getLocation() );
        stain.setVisible( game.isCellVisibleToPlayer( point ) );
        RepaintManager.currentManager( this ).addDirtyRegion( this, stain.getBounds().x-cellSize, stain.getBounds().y-cellSize, stain.getBounds().width+cellSize*2, stain.getBounds().height+cellSize*2 );
    }
        
    public void addCorpse( Character character )
    {
        //System.out.println( "Adding corpse" );
        ImageStain stain;
        Point location;
        CharacterSprite sprite = (CharacterSprite)characterSprites.get( character );
        location = character.getLocation();
        // We re-set the sprite's position to take care of those pesky size-changing sprites
        sprite.setLocation( location.x*cellSize, location.y*cellSize );
	sprite.setVisible( false );
        stain = new ImageStain( sprite );
        stain.setLocation( new Point( location ) );
        //System.out.println( "Corpse location: " + character.getLocation().toString() );
        stain.setBounds( sprite.getLocation().x, sprite.getLocation().y, sprite.getBounds().width, sprite.getBounds().height );
        //System.out.println( "Sprite location: " + sprite.getLocation().toString() );
        //System.out.println( "Stain bounds: " + sprite.getBounds().x + ", " + sprite.getBounds().y + ", " + sprite.getBounds().width + ", " + sprite.getBounds().height );
	addStain( stain );
    }
    
    public void addObjectStains( MapObject mapObject )
    {
	//System.out.println( "Adding stains for " + mapObject.getName() );
	MapObjectCell [] cells = mapObject.getCells();
	if( cells != null )
	{
	    for( int j=0; j < cells.length; j++ )
	    {
		int x = cells[j].getX();
		int y = cells[j].getY();
		
		int xLocation = x*cellSize;
		int yLocation = y*cellSize;
		int objIndex = cells[j].getType();
		if( cells[j].isDestroyed() )
		{
		    objIndex = cells[j].getDestroyedType();
		    if( objIndex < 0 )
		    {
			objIndex = cells[j].getType();
		    }
		}
		ImageStain stain = new ImageStain( (Image)objectCells.elementAt( objIndex ) );
		stain.setBounds( xLocation, yLocation, cellSize, cellSize );
		stain.setVisible( true );
		stainSprites.add( stain );
		RepaintManager.currentManager( this ).addDirtyRegion( this, stain.getBounds().x, stain.getBounds().y, stain.getBounds().width, stain.getBounds().height );
	    }
	}
	
    }
    
    public Point getCell( Point point )
    {
	int x=0;
	int y=0;

	x = point.x/cellSize;
	y = point.y/cellSize;
	if( x < 0 ) x = 0;
	if( x >= map.getHorizontalSize() ) x = map.getHorizontalSize()-1;
	if( y < 0 ) y = 0;
	if( y >= map.getVerticalSize() ) y = map.getVerticalSize()-1;

	return new Point( x, y );
    }

    public Point getCellForAttack( Point point )
    {
        int x=0;
        int y=0;

        x = point.x/cellSize;
        y = point.y/cellSize;
        if( x < 0 ) x = 0;
        if( x >= map.getHorizontalSize() ) x = map.getHorizontalSize()-1;
        if( y < 0 ) y = 0;
        if( y >= map.getVerticalSize() ) y = map.getVerticalSize()-1;

        Iterator iterator = characterSprites.entrySet().iterator();
        CharacterSprite sprite;
        java.util.Map.Entry entry;
        int newX=x;
        int newY=y;
        while( iterator.hasNext() )
        {
            entry = (java.util.Map.Entry)iterator.next();
            sprite = (CharacterSprite)entry.getValue();
            if( sprite.getVisualBounds().contains( point ) )
            {
                //System.out.println( "Clicked on " + entry.getKey().getClass().getName() );
                Character character = (Character)entry.getKey();
                if( character.isAlive() && sprite.isVisible() )
                {
                    newX = character.getLocation().x;
                    newY = character.getLocation().y;
                    if( newX == x && newY == y )
                    {
                        //It's spot on, we won't find a better match
                        //System.out.println( "Spot on" );
                        break;
                    }
                }
            }
        }

        return new Point( newX, newY );
    }
    

    public int getCellSize()
    {
        return cellSize;
    }
        
    public void update( Observable o, Object arg )
    {
	highlightArea = null;
        if( o == game )
        {
            if( arg instanceof DoorAction )
            {
                DoorAction action = (DoorAction)arg;
                bufferedDoorActions.add( action );
		if( action.getActor() != null )
		{
		    getSpriteFor( action.getActor() ).setDirection( action.getActor().getDirection() );
		}
            }
            if( arg instanceof MoveAction )
            {
		bufferedMoveActions.add( (MoveAction)arg );
            }
            if( arg instanceof UseAction )
            {
                UseAction action = (UseAction)arg;
                getSpriteFor( action.getActor() ).setDirection( action.getActor().getDirection() );
                //System.out.println( "GameMapPanel: Got useAction" );
            }
            if( arg instanceof HealAction )
            {
                animateHealAction( (HealAction)arg );
            }
	    if( arg instanceof NoAPAction )
	    {
		long time = System.currentTimeMillis();
		if( game.getState() == Game.PLAYER_MOVE && time-lastMessageTime > 1000 )
                {
		    animateNoAPAction( (NoAPAction)arg );
		    lastMessageTime = time;
		}
	    }
            if( arg instanceof ShootAction )
            {
		//System.out.println( "MapPanel got shoot action to " + Objective.pointToString( ((ShootAction)arg).getEndPoint() ) );
		bufferedAttackActions.add( (ShootAction)arg ); 
	    }
            if( arg instanceof WoundAction )
            {
                bufferedAttackActions.add( (WoundAction)arg );
            }            
            if( arg instanceof NewCorpseAction )
            {
                bufferedAttackActions.add( (NewCorpseAction)arg );
            }
            if( arg instanceof MeleeAction )
            {
                bufferedAttackActions.add( (MeleeAction)arg );
            }
	    if( arg instanceof WaitAction )
	    {
		WaitAction wAction = (WaitAction)arg;
		CharacterSprite sprite = getSpriteFor( wAction.getActor() );
		sprite.setFrame( 0 );
		sprite.setAction( wAction );
	    }
	    if( arg instanceof ObjectDestroyedEvent )
	    {
		addObjectStains( ((ObjectDestroyedEvent)arg).getObject() );
	    }
	    if( arg instanceof PushObjectAction )
	    {
		PushObjectAction pushAction = (PushObjectAction)arg;
		Point start = pushAction.getStartPoint();
		Point end = pushAction.getEndPoint();
		int xOffset = (end.x-start.x)*-cellSize;
		int yOffset = (end.y-start.y)*-cellSize;
		movingObjects.put( pushAction.getObject(), new Point( xOffset, yOffset ) );
		bufferedPushActions.add( pushAction );
	    }
            if( arg instanceof MonsterBirth )
            {
                addSpriteForCharacter( ((MonsterBirth)arg).getActor() );
                checkMonsterVisibility();
            }
            if( arg instanceof MapEvent )
            {
                checkMonsterVisibility();
                repaint();
            }
            if( arg instanceof Message )
            {
		String message = ((Message)arg).getMessage();
		previousMessages.add( message );
		currentMessageIndex = 0;
		lastMessageTime = System.currentTimeMillis();
		if( ((Message)arg).isAnnoying() )
		{
		    JOptionPane.showMessageDialog( getParent(), message, "Message", JOptionPane.INFORMATION_MESSAGE, null );
		}
		else
		{
		    animateString( message );
		}
            }
	    if( arg instanceof SyncEvent )
	    {
		animating = true;
		SyncEvent sync = (SyncEvent)arg;
		if( sync.getSyncDoors() )
		{
		    animateBufferedDoorActions();
		}
		if( sync.getSyncMove() )
		{
		    animateBufferedMoveActions();
		    animateBufferedPushActions();
		}
		if( sync.getSyncAttacks() )
		{
		    animateBufferedAttackActions();
		}
		if( game.getState() == Game.PLAYER_MOVE )
		{
		    checkMonsterVisibility();
		}	
		animating = false;
	    }
        }
    }
    
    public void setupNewTurn()
    {
	System.out.println( "New turn" );
	frame = 0;
	CharacterSprite sprite;
	for( int i=0; i < game.getPlayerCharacters().size(); i++ )
	{
	    sprite = (CharacterSprite)characterSprites.get( game.getPlayerCharacter( i ) );
	    sprite.setLocation( sprite.getCharacter().getLocation().x*cellSize, sprite.getCharacter().getLocation().y*cellSize );
	}
	for( int i=0; i < game.getMonsters().size(); i++ )
	{
	    sprite = (CharacterSprite)characterSprites.get( game.getMonster( i ) );
	    sprite.setLocation( sprite.getCharacter().getLocation().x*cellSize, sprite.getCharacter().getLocation().y*cellSize );
            sprite.setVisible( game.isCellVisibleToPlayer( sprite.getCharacter().getLocation() ) && sprite.getCharacter().isAlive() );
	}
    }
    
    public void showNextMessage()
    {
	currentMessageIndex--;
	if( currentMessageIndex < 1 )
	{
	    currentMessageIndex = 1;
	}
	if( currentMessageIndex <= previousMessages.size() )
	{
	    String message = currentMessageIndex + ": " + (String)previousMessages.elementAt( previousMessages.size()-currentMessageIndex );
	    animateString( message );
	}
    }
    
    public void showPreviousMessage()
    {
	currentMessageIndex++;
	if( currentMessageIndex > previousMessages.size() )
	{
	    currentMessageIndex = previousMessages.size();
	}
	if( currentMessageIndex <= previousMessages.size() )
	{
	    String message = currentMessageIndex + ": " + (String)previousMessages.elementAt( previousMessages.size()-currentMessageIndex );
	    animateString( message );
	}
    }
    
    public void checkMonsterVisibility()
    {
        CharacterSprite sprite;
        ImageStain stain;
        Point point;
        boolean oldVisibility;
    
	for( int i=0; i < game.getMonsters().size(); i++ )
	{
	    sprite = getSpriteFor( game.getMonster( i ) );
            oldVisibility = sprite.isVisible();
            sprite.setVisible( game.isCellVisibleToPlayer( sprite.getCharacter().getLocation() ) && sprite.getCharacter().isAlive() );
            //sprite.setLocation( sprite.getCharacter().getLocation().x*cellSize, sprite.getCharacter().getLocation().y*cellSize );
            if( oldVisibility || sprite.isVisible() )
            {
                // Sprite is or was visible, add dirty region
                RepaintManager.currentManager( this ).addDirtyRegion( this, sprite.getLocation().x-cellSize, sprite.getLocation().y-cellSize, sprite.getBounds().width+cellSize*2, sprite.getBounds().height+cellSize*2 );
            }
	}
        for( int i=0; i < stainSprites.size(); i++ )
        {
            stain = (ImageStain)stainSprites.elementAt( i );
            if( !stain.isVisible() )
            {
                point = new Point( stain.getLocation() );
		stain.setVisible( game.isCellVisibleToPlayer( point ) );
                RepaintManager.currentManager( this ).addDirtyRegion( this, stain.getLocation().x*cellSize-cellSize, stain.getLocation().y*cellSize-cellSize, stain.getBounds().width+cellSize*2, stain.getBounds().height+cellSize*2 );
            }
        }
    }
    
    public void addSpriteForCharacter( Character character )
    {
        String characterClass;
        CharacterSprite newSprite;
        
        try
        {
            characterClass = character.getClassName();
	    if( characterClass.endsWith( GenericMonster.GENERIC_MONSTER_KEY ) )
	    {
		newSprite = new GenericMonsterSprite();
		((GenericMonsterSprite)newSprite).setType( character.getType() );
	    }
	    else
	    {
		newSprite = (CharacterSprite)Class.forName( "TheGame."+characterClass+"Sprite" ).newInstance();
	    }
            newSprite.setCharacter( character );
            newSprite.setLocation( character.getLocation().x*cellSize, character.getLocation().y*cellSize );
            newSprite.setDirection( character.getDirection() );
	    newSprite.setAction( character.getLastAction() );
            characterSprites.put( character, newSprite );
            System.out.println( "Added sprite for: " + characterClass );
        }
        catch( Exception e )
        {
            e.printStackTrace();
            System.exit( 1 );
        }

    }

    public void addSpriteForWeapon( Weapon weapon )
    {
        String weaponClass;
        WeaponSprite newSprite;
        try
        {
            weaponClass = weapon.getClass().getName();
            if( !weaponSprites.containsKey( weaponClass ) )
            {
                newSprite = (WeaponSprite)Class.forName( weaponClass+"Sprite" ).newInstance();
                weaponSprites.put( weaponClass, newSprite );
                System.out.println( "Added sprite for weapon: " + weaponClass );
            }
        }
        catch( Exception e )
        {
            e.printStackTrace();
            System.exit( 1 );
        }
        
    }
    
    public CharacterSprite getSpriteFor( Character character )
    {
        return (CharacterSprite)characterSprites.get( character );
    }
    
    public void scrollRectToVisible( Rectangle aRect )
    {
	super.scrollRectToVisible( aRect );
	if( System.currentTimeMillis()-lastMessageTime > 1000 )
	{
	    clearMessageString();
	}
    }
    
    public void setDrawPlayerHint( boolean aBool )
    {
	drawPlayerHint = aBool;
    }
    
    public void setDrawShootingHints( boolean drawShootingHints )
    {
	this.drawShootingHints = drawShootingHints;
    }
    
    public void centerOnCharacter( Character character )
    {
        Rectangle visRect = getVisibleRect();
        int x = visRect.width/cellSize;
        int y = visRect.height/cellSize;
        scrollRectToVisible( new Rectangle( (character.getLocation().x-x/2)*cellSize, (character.getLocation().y-y/2)*cellSize, x*cellSize, y*cellSize ) );
    }


    private void loadFloorsFromMap()
    {
	Vector floorVector = map.getFloorSets();
	if( floorVector.size() > 0 )
	{
	    for( int i = 0; i < floorVector.size(); i++ )
	    {
		ResourceManager.cropImage( (String)floorVector.elementAt( i ), floorCells, cellSize );
	    }
	}	
    }

    private void loadWallsFromMap()
    {
	Vector wallVector = map.getWallSets();
	if( wallVector.size() > 0 )
	{
	    for( int i = 0; i < wallVector.size(); i++ )
	    {
		ResourceManager.cropImage( (String)wallVector.elementAt( i ), wallCells, cellSize );
	    }
	}	
    }

    private void loadDoorsFromMap()
    {
	System.out.println( "Loading door sets" );
	Vector doorVector = map.getDoorSets();
	Vector doorImages = new Vector();
	if( doorVector.size() > 0 )
	{
	    for( int i = 0; i < doorVector.size(); i++ )
	    {
		ResourceManager.cropImage( (String)doorVector.elementAt( i ), doorImages, cellSize );
	    }
	}
	doorCells = new Image[4][doorImages.size()/4];
	System.out.println( doorImages.size()/4 + " door frame(s) found" );
	openDoorIndex = (doorImages.size()/4)-1;
	System.out.println( "Open door is at " + openDoorIndex );
	closedDoorIndex = 0;
	for( int i=0; i < 4; i++ )
	{
	    for( int j=0; j < doorImages.size()/4;j++ )
	    {
		doorCells[i][j] = (Image)doorImages.elementAt( i+j*4 );
	    }
	}	
    }

    private void loadObjectsFromMap()
    {
        Vector objectVector = map.getObjectSets();
        if( objectVector.size() > 0 )
        {
            for( int i = 0; i < objectVector.size(); i++ )
            {
                ResourceManager.cropImage( (String)objectVector.elementAt( i ), objectCells, cellSize );
            }
        }
    }

    private void loadIcons()
    {
        Vector images = new Vector();
        ResourceManager.cropImage( "HealthIcons.png", images, cellSize );
        healthIcons = new Image[4];
        for( int i=0; i < 4; i++ )
        {
            healthIcons[i] = (Image)images.elementAt( i );
        }  
	noAPIcon = ResourceManager.loadImage( "NoAPIcon.png" );
    }
    
    public void setMap( Map map )
    {
	this.map = map;
	Dimension mySize = new Dimension ( map.getHorizontalSize()*cellSize, map.getVerticalSize()*cellSize );
	setMinimumSize( mySize );
	setPreferredSize( mySize );
	setMaximumSize( mySize );
	floorCells = new Vector();
	wallCells = new Vector();
	objectCells = new Vector();
	loadFloorsFromMap();
	loadWallsFromMap();
	loadDoorsFromMap();
	loadObjectsFromMap();
	repaint();
    }
    
    public void setSlowify()
    {
	slowifyRate = 30;
    }
    
    public Vector getFloorCells()
    {
	return floorCells;
    }
    
    public Vector getWallCells()
    {
	return wallCells;
    }
    
    public Vector getObjectCells()
    {
	return objectCells;
    }
    
/*    public Vector getDoorCells()
    {
	return doorCells;
    }*/
    
    public Dimension getPreferredScrollableViewportSize()
    {
	return viewportSize;
    }
    
    public int getScrollableUnitIncrement( Rectangle visibleRect, int orientation, int direction ) 
    {
	//get the current position
	int currentPosition = 0;
	if( orientation == SwingConstants.HORIZONTAL )
	    currentPosition = visibleRect.x;
	else
	    currentPosition = visibleRect.y;
    
	//return the number of pixels between currentPosition
	//and the nearest tick mark in the indicated direction
	if ( direction < 0 ) 
	{
	    int newPosition = currentPosition - (currentPosition / cellSize) * cellSize;
	    return (newPosition == 0) ? cellSize : newPosition;
	} 
	else 
	{
	    return ((currentPosition / cellSize) + 1) * cellSize - currentPosition;
	}
    }
    
    public int getScrollableBlockIncrement( Rectangle visibleRect, int orientation, int direction ) 
    {
	int amount = 0;
	if( orientation == SwingConstants.HORIZONTAL )
	{
	    amount = getScrollableUnitIncrement( visibleRect, orientation, direction );
	    if( amount < cellSize )
		amount = amount+cellSize;
	    return visibleRect.width - amount;
	}
	else
	{
	    amount = getScrollableUnitIncrement( visibleRect, orientation, direction );
	    if( amount < cellSize )
		amount = amount+cellSize;
	    return visibleRect.height - amount;
	}
    }
    
    public boolean getScrollableTracksViewportWidth()
    {
	return false;
    }
    
    public boolean getScrollableTracksViewportHeight()
    {
	return false;
    }

    
}

