// Map.java
// Started on June 16th, 2001
// Copyright (c) Virtu Halttunen

package TheGame;

import java.awt.Point;
import java.util.*;
import java.util.zip.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;
import java.io.*;
import java.awt.geom.*;

public class Map
{
    public static final int NORTH = 0;
    public static final int NORTH_EAST = 1;
    public static final int EAST = 2;
    public static final int SOUTH_EAST = 3;
    public static final int SOUTH = 4;
    public static final int SOUTH_WEST = 5;
    public static final int WEST = 6;
    public static final int NORTH_WEST = 7;

    public static final int LIGHT_LEVELS = 32;
    
    private MapCell[][] cellMap;
    private Vector doors;
    private Vector objects;
    private HashMap rooms;
    private short[][] utilityMap;
    private int horizontalSize;
    private int verticalSize;
    private Vector floorSetNames;
    private Vector wallSetNames;
    private Vector objectSetNames;
    private Vector doorSetNames;
    private String name;
    private String fileName;
    private Game game;
    private boolean doorsLocked = false;
    
    public Map( int horizontalSize, int verticalSize )
    {
	this.horizontalSize = horizontalSize;
	this.verticalSize = verticalSize;
	cellMap = new MapCell[ horizontalSize ][ verticalSize ];
        doors = new Vector();
        objects = new Vector();
	rooms = new HashMap();
	utilityMap = new short[ horizontalSize ][ verticalSize ];
	floorSetNames = new Vector();
	wallSetNames = new Vector();
	objectSetNames = new Vector();
	doorSetNames = new Vector();
    }
    
    public void setGame( Game game )
    {
	this.game = game;
    }
    
    public void setCell( int x, int y, MapCell cell )
    {
        cellMap[x][y] = cell;
    }

    public MapCell getCell( int x, int y )
    {
	//System.out.println( "Map: getCellType("+x+","+y+")" );
	return cellMap[x][y];
    }
    
    public MapCell getCell( Point point )
    {
	//System.out.println( "Map: getCellType() " + point.toString() );
	return cellMap[ point.x ][ point.y ];
    }

    public boolean isBlockedByInhabitant( Point point, int minBlockingSize, boolean ignoreObjects )
    {
        return isBlockedByInhabitant( point.x, point.y, minBlockingSize, ignoreObjects );
    }
    
    public boolean isBlockedByInhabitant( int x, int y, int minBlockingSize )
    {
        return isBlockedByInhabitant( x, y, minBlockingSize, false );
    }
    
    // Used for line of fire checking, hence isBlockingSight for objects
    public boolean isBlockedByInhabitant( int x, int y, int minBlockingSize, boolean ignoreObjects )
    {
	boolean blocked = false;
	Point point = new Point( x, y );
	Character character;
        MapObject mapObject;
	for( int i=0; i<game.getMonsters().size(); i++ )
	{
	    character = game.getMonster(i);
	    if( character.getLocation().equals( point ) && character.getSize() > minBlockingSize && character.isAlive() )
	    {
		blocked = true;
		break;
	    }
	}
 	for( int i=0; i<game.getPlayerCharacters().size(); i++ )
	{
	    character = game.getPlayerCharacter(i);
	    if( character.getLocation().equals( point ) && character.getSize() > minBlockingSize && character.isAlive() )
	    {
		blocked = true;
		break;
	    }
	}
        for( int i=0; i<game.getNumberOfObjectives(); i++ )
        {
            if( ( game.getObjective(i).getSize() > minBlockingSize ) && game.getObjective(i).isLocatedIn( x, y ) )
            {
                blocked = true;
                break;
            }
        }
        if( cellMap[x][y].hasObjects() && !ignoreObjects )
        {
            for( int i = 0; i < getNumberOfObjects(); i++ )
            {
                mapObject = getObject( i );
                if( mapObject.isLocatedIn( x, y ) && mapObject.isBlockingSight( new Point( x, y ) ) )
                {
                    blocked = true;
                    break;
                }
            }
        }
	return blocked;
   }

    public boolean isViewable( int x, int y, boolean ignoreObjects )
    {
        MapObject mapObject;
        if( x < 0 ) return false;
        if( x >= horizontalSize ) return false;
        if( y < 0 ) return false;
        if( y >= verticalSize ) return false;
        if( cellMap[x][y] == null )
            return false;
        if( cellMap[x][y].getWall() > -1 )
        {
            return false;
        }
        else
        {
            if( cellMap[x][y].hasDoor() )
            {
                if( getDoor( x, y ).isOpen() )
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            if( cellMap[x][y].hasObjects() && !ignoreObjects )
            {
                for( int i = 0; i < getNumberOfObjects(); i++ )
                {
                    mapObject = getObject( i );
                    if( mapObject.isLocatedIn( x, y ) && mapObject.isBlockingSight( new Point( x, y ) ) )
                    {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    public boolean isViewable( int x, int y )
    {
        return isViewable( x, y, false );
    }
    
    public boolean isPassable( int x, int y )
    {
        MapObject mapObject;
        if( x < 0 ) return false;
        if( x >= horizontalSize ) return false;
        if( y < 0 ) return false;
        if( y >= verticalSize ) return false;
        if( cellMap[x][y] == null )
	{
            return false;
	}
        if( cellMap[x][y].getFloor() <= -1 )
        {
            return false;
        }
	if( cellMap[x][y].getWall() > -1 )
	{
	    return false;
	}
	if( cellMap[x][y].hasObjects() )
	{
	    for( int i = 0; i < getNumberOfObjects(); i++ )
	    {
		mapObject = getObject( i );
		if( mapObject.isLocatedIn( x, y ) && mapObject.isBlockingMovement( new Point( x, y ) ) )
		{
		    return false;
		}
	    }
	}	    
	if( cellMap[x][y].hasDoor() )
	{
	    if( getDoor( x, y ).isOpen() )
	    {
		return true;
	    }
	    else
	    {
		return false;
	    }
	}
        return true;
    }
 
    public boolean isPassable( Point cell )
    {
        return isPassable( cell.x, cell.y );
    }
    
    public boolean isPassable( Point cell, int direction )
    {
	Point target = getNeighbor( cell, direction );
	if( target.x < 0 ) return false;
	if( target.x >= horizontalSize ) return false;
	if( target.y < 0 ) return false;
	if( target.y >= verticalSize ) return false;
	return isPassable( target );
    }
    
    public boolean isPassable( Point point, boolean canOpenDoors )
    {
        if( !canOpenDoors )
            return isPassable( point );
        boolean passable = isPassable( point );
        if( !passable )
        {
            if( isClosedDoor( point.x, point.y ) )
            {
                if( getDoor( point.x, point.y ).isLocked() )
                {
                    passable = false;
                }
                else
                {
                    passable = true;
                }
            }
        }
        return passable;
    }
    
    public boolean isPassable( int x, int y, boolean canOpenDoors )
    {
        if( !canOpenDoors )
            return isPassable( x, y );
        boolean passable = isPassable( x, y );
        if( !passable )
        {
            if( isClosedDoor( x, y ) )
            {
                if( getDoor( x, y ).isLocked() )
                {
                    passable = false;
                }
                else
                {
                    passable = true;
                }
            }
        }
        return passable;
    }
    
    public boolean isPassable( Point cell, int direction, boolean canOpenDoors )
    {
        if( !canOpenDoors )
            return isPassable( cell, direction );
        boolean passable = isPassable( cell, direction );
        passable = ( passable || isClosedDoor( cell.x, cell.y ) );
        return passable;
    }

    public boolean isClosedDoor( Point point )
    {
        if( cellMap[ point.x ][ point.y ] == null )
        {
            return false;
        }
        else
        {
            if( cellMap[ point.x ][ point.y ].hasDoor() )
            {
                Door door = getDoor( point.x, point.y );
                return door.isClosed();
            }
        }
        return false;
    }


    public boolean isClosedDoor( int x, int y )
    {
	if( x < 0 || x >= horizontalSize || y < 0 || y>= verticalSize )
	{
	    return false;
	}
        if( cellMap[ x ][ y ] == null )
        {
            return false;
        }
        else
        {
            if( cellMap[ x ][ y ].hasDoor() )
            {
                Door door = getDoor( x, y );
                return door.isClosed();
            }
        }
        return false;
    }

    public boolean isClosedDoor( Point cell, int direction )
    {
	Point target = getNeighbor( cell, direction );
	if( target.x < 0 ) return false;
	if( target.x >= horizontalSize ) return false;
	if( target.y < 0 ) return false;
	if( target.y >= verticalSize ) return false;
	return isClosedDoor( target.x, target.y );
    }
    
    public boolean isOpenDoor( Point point )
    {
        if( cellMap[ point.x ][ point.y ] == null )
        {
            return false;
        }
        else
        {
            if( cellMap[ point.x ][ point.y ].hasDoor() )
            {
                Door door = getDoor( point.x, point.y );
                return !door.isClosed();
            }
        }
        return false;
    }

    
    public boolean isOpenDoor( int x, int y )
    {
        if( cellMap[ x ][ y ] == null )
        {
            return false;
        }
        else
        {
            if( cellMap[ x ][ y ].hasDoor() )
            {
                Door door = getDoor( x, y );
                return !door.isClosed();
            }
        }
        return false;
    }

    public boolean isOpenDoor( Point cell, int direction )
    {
	Point target = getNeighbor( cell, direction );
	if( target.x < 0 ) return false;
	if( target.x >= horizontalSize ) return false;
	if( target.y < 0 ) return false;
	if( target.y >= verticalSize ) return false;
	return isOpenDoor( target.x, target.y );
    }
   
    public static Point getNeighbor( Point cell, int direction )
    {
	Point target = new Point( cell );
	switch( direction )
	{
	    case NORTH: target.y--; break;
	    case NORTH_EAST: target.x++; target.y--; break;
	    case EAST: target.x++; break;
	    case SOUTH_EAST: target.x++; target.y++; break;
	    case SOUTH: target.y++; break;
	    case SOUTH_WEST: target.x--; target.y++; break;
	    case WEST: target.x--; break;
	    case NORTH_WEST: target.x--; target.y--; break;
	}
	return target;
    }
    
    public static boolean isNeighbor( Point one, Point two )
    {
        if( Math.abs( one.x - two.x ) <= 1 && Math.abs( one.y - two.y ) <= 1 )
            return true;
        return false;
    }
    
    public void addDoor( Door door )
    {
        if( cellMap[ door.getX() ][ door.getY() ] != null )
        {
            if( cellMap[ door.getX() ][ door.getY() ].hasDoor() )
            {
                removeDoor( door.getX(), door.getY() );
            }
            cellMap[ door.getX() ][ door.getY() ].setDoor( true );
            doors.add( door );
        }
    }

    public void removeDoor( int x, int y )
    {
        if( cellMap[x][y].hasDoor() )
        {
            Door theDoor = getDoor( x, y );
            doors.remove( theDoor );
            cellMap[x][y].setWall( -1 );
        }
    }
    
    public Door getDoor( int x, int y )
    {
        Door theDoor = null;
        for( int i = 0; i < doors.size(); i++ )
        {
            if( ((Door)doors.elementAt( i )).isLocatedIn( x, y ) )
                theDoor = (Door)doors.elementAt( i );
        }
        return theDoor;
    }

    public void openDoor( Point point )
    {
        if( cellMap[point.x][point.y].hasDoor() )
        {
            Door door = getDoor( point.x, point.y );
            if( door.isClosed() && !door.isLocked() )
            {
                door.setOpen( true );
                openDoor( getNeighbor( point, door.getDirection() ) );
            }
        }
    }

    public void closeDoor( Point point )
    {
        if( cellMap[point.x][point.y].hasDoor() )
        {
            Door door = getDoor( point.x, point.y );
            if( !door.isClosed() && !door.isLocked() )
            {
                door.setOpen( false );
                closeDoor( getNeighbor( point, door.getDirection() ) );
            }
        }
    }

    public void lockAllDoors()
    {
        for( int i = 0; i < doors.size(); i++ )
        {
            ((Door)doors.elementAt(i)).setLocked( true );
        }
    }
    
    public void moveObject( MapObject mapObject, int direction )
    {
	MapObjectCell [] cells = mapObject.getCells();
	for( int i=0; i < cells.length; i++ )
	{
	    Point point = getNeighbor( new Point( cells[i].getX(), cells[i].getY() ), direction );
	    cells[i].setPosition( point.x, point.y );
	    MapCell mapCell = getCell( point );
	    mapCell.setObjects( false );
	}
	mapObject.calculateSize();
	// Reset the objects for map
	for( int o=0; o < objects.size(); o++ )
	{
	    MapObject theObject = (MapObject)objects.elementAt( o );
	    cells = theObject.getCells();
	    for( int c=0; c < cells.length; c++ )
	    {
		Point point = new Point( cells[c].getX(), cells[c].getY() );
		MapCell mapCell = getCell( point );
		mapCell.setObjects( true );
	    }
	}
    }
    
    public boolean canMoveObject( MapObject mapObject, int direction )
    {
	boolean canMove = true;
	
	//System.out.println( "Checking move for " + mapObject.getName() + "("+mapObject.getId() + ")");
	MapObjectCell [] cells = mapObject.getCells();
	for( int i=0; i < cells.length && canMove; i++ )
	{
	    Point check = getNeighbor( new Point( cells[i].getX(), cells[i].getY() ), direction );
	    if( check.x < 0 || check.x >= horizontalSize || check.y < 0 || check.y >= verticalSize )
	    {
		return false;
	    }
	    //System.out.println( "Checking neighbor for point (" + cells[i].getX() + ", "+  cells[i].getY() +")" );
	    MapCell mapCell = getCell( check );
	    if( mapCell == null )
	    {
		return false;
	    }
	    if( mapCell.getWall() > -1  )
	    {
		//System.out.println( "Blocked by wall at " + Objective.pointToString( check ) );
		canMove = false;
	    }
	    if( isClosedDoor( check ) )
	    {
		canMove = false;
	    }
	    if( mapCell.hasObjects() )
	    {
		for( int o=0; o < objects.size() && canMove; o++ )
		{
		    MapObject checkObject = (MapObject)objects.elementAt( o );
		    if( checkObject != mapObject && checkObject.isLocatedIn( check ) && !checkObject.isDestroyed() )
		    {
			//System.out.println( "Blocked by " + checkObject.getName() + "("+checkObject.getId() + ")");
			canMove = false;
		    }
		}
	    }
	    for( int o=0; o < game.getNumberOfObjectives() && canMove; o++ )
	    {
		Objective objective = game.getObjective( o );
		if( objective.isLocatedIn( check ) && objective.getSize() > 50 )
		{
		    canMove = false;
		}
	    }
	}
	return canMove;
    }
    
    public static int getDirection( Point from, Point to, int originalDirection )
    {
	int direction = originalDirection;
	if( !from.equals( to ) )
	{
	    // Uses Arc2D to save the trouble of calculating the angle between the points
	    Arc2D.Double helper = new Arc2D.Double( from.x-1, from.y-1, 2, 2, 0.0, 0.0, Arc2D.OPEN );
	    helper.setAngles( new Point( from.x, from.y-1 ), to );
	    int angle = (int)helper.getAngleExtent();
	    if( angle < 23 ) direction = NORTH;
	    if( angle >= 23 && angle < 68 ) direction = NORTH_WEST;
	    if( angle >= 68 && angle < 113 ) direction = WEST;
	    if( angle >= 113 && angle < 158 ) direction = SOUTH_WEST;
	    if( angle >= 158 && angle < 203 ) direction = SOUTH;
	    if( angle >= 203 && angle < 248 ) direction = SOUTH_EAST;
	    if( angle >= 248 && angle < 293 ) direction = EAST;
	    if( angle >= 293 && angle < 338 ) direction = NORTH_EAST;
	    if( angle >= 338 ) direction = NORTH;
	}
	
	return direction;
    }
   
    public boolean hasLineOfSight( Point start, Point end )
    {
        if( start.x < 0 || start.x >= getHorizontalSize() ) return false;
        if( start.y < 0 || start.y >= getVerticalSize() ) return false;
        if( end.x < 0 || end.x >= getHorizontalSize() ) return false;
        if( end.y < 0 || end.y >= getVerticalSize() ) return false;
	// Uses Bresenham's line drawing algorithm with a small change
	int dX = Math.abs( end.x-start.x );	// Difference in coordinates
	int dY = Math.abs( end.y-start.y );
	int checkX = start.x;
	int checkY = start.y;	
	
	int xIncr, yIncr;
	if( start.x > end.x ) 
	    xIncr=-1;
	else
	    xIncr=1;
	    
	if( start.y > end.y )
	    yIncr=-1;
	else
	    yIncr=1;
			    
	if( dX >= dY )
	{
	    int dPr = dY*2;		// Amount to increment decision to right
	    int dPru = dPr - (dX*2); 	// Amount to increment up
	    int p = dPr - dX;		// decision start value
	    
	    while( dX > 0 )
	    {
		if( p > 0 )	// Is the next one up and right 
		{
		    checkX+=xIncr;
		    checkY+=yIncr;
		    p+=dPru;
		}
		else		// Or just right
		{
		    checkX+=xIncr;
		    p+=dPr;
		}
		dX--;
		// Check the cell
		if( !isViewable( checkX, checkY ) )
		{
		    return false;
		}
	    }
	}
	else
	{
	    int dPr = dX*2;
	    int dPru = dPr - (dY*2);
	    int p = dPr-dY;
	    
	    while( dY>0 )
	    {

		if( p > 0 )
		{
		    checkX+=xIncr;
		    checkY+=yIncr;
		    p+=dPru;
		}
		else
		{
		    checkY+=yIncr;
		    p+=dPr;
		}
		dY--;
		//Check the cell
		if( !isViewable( checkX, checkY ) )
		{
		    return false;
		}
	    }
	}
	return true;
    }

    public Point getLineOfFire( Point start, Point end )
    {
        return getLineOfFire( start, end, 50 );
    }

    
    public Point getLineOfFire( Point start, Point end, int minBlockingSize )
    {
	// Uses Bresenham's line drawing algorithm with a small change to stop characters blocking their own LOR
	int dX = Math.abs( end.x-start.x );	// Difference in coordinates
	int dY = Math.abs( end.y-start.y );
	int checkX = start.x;
	int checkY = start.y;
	
	int xIncr, yIncr;
	if( start.x > end.x ) 
	    xIncr=-1;
	else
	    xIncr=1;
	    
	if( start.y > end.y )
	    yIncr=-1;
	else
	    yIncr=1;
			    
	if( dX >= dY )
	{
	    int dPr = dY*2;		// Amount to increment decision to right
	    int dPru = dPr - (dX*2); 	// Amount to increment up
	    int p = dPr - dX;		// decision start value
	    
	    while( dX > 0 )
	    {
		if( p > 0 )	// Is the next one up and right 
		{
		    checkX+=xIncr;
		    checkY+=yIncr;
		    p+=dPru;
		}
		else		// Or just right
		{
		    checkX+=xIncr;
		    p+=dPr;
		}
		dX--;
		// Check the cell
		if( !isViewable( checkX, checkY ) || isBlockedByInhabitant( checkX, checkY, minBlockingSize ) )
		{
		    return new Point( checkX, checkY );
		}
	    }
	}
	else
	{
	    int dPr = dX*2;
	    int dPru = dPr - (dY*2);
	    int p = dPr-dY;
	    
	    while( dY>0 )
	    {
		if( p > 0 )
		{
		    checkX+=xIncr;
		    checkY+=yIncr;
		    p+=dPru;
		}
		else
		{
		    checkY+=yIncr;
		    p+=dPr;
		}
		dY--;
		//Check the cell
		if( !isViewable( checkX, checkY ) || isBlockedByInhabitant( checkX, checkY, minBlockingSize ) )
		{
		    return new Point( checkX, checkY );
		}
	    }
	}
	return end;
    }

   
    public void resetUtilityMap()
    {
	for( int i=0; i < horizontalSize; i++ )
	    for( int j=0; j < verticalSize; j++ )
		utilityMap[i][j]=0;
    }
    
/*    public Vector findShortestPathFromPoint( Point start, Point end )
    {
        return findShortestPath( start, end, false );
    }*/
    
    public Vector findShortestPathFromPoint( Point start, Point end, boolean canOpenDoors )
    {
	Vector path = new Vector();
	resetUtilityMap();
	utilityMap[start.x][start.y]=(short)1;
	LinkedMapNodeList search = new LinkedMapNodeList();
	MapNode node;
	MapNode neighbor;
	Point point;
        Point npoint;
	int direction;

	node = new MapNode( start, 8, false );
	if( node.getX() == end.x && node.getY() == end.y )
	{
	    path.add( start );
	    return path;
	}
	while( ! ((node.getX() == end.x) && (node.getY() == end.y) ) )
	{
	    do
	    {
		direction = node.nextDirection();
		//System.out.println( "Checking: " + node.getX() + "," + node.getY() +" "+ direction );
		node.setChecked( direction );
                npoint = getNeighbor( node.getPoint(), direction );
		if( isPassable( npoint, canOpenDoors ) && ( (utilityMap[ npoint.x][ npoint.y] == 0 )  || (utilityMap[ npoint.x ][ npoint.y ] >  utilityMap[ node.getX() ][ node.getY() ]+1 ) ) )
		{
		    //System.out.println( "Adding point to search" );
		    neighbor = new MapNode( new Point( npoint ), 8, false  );
		    //neighbor.setChecked( (direction+4)%8 );	// Serves no purpose yet, quite on the contrary, breaks the search
		    search.push( neighbor );
                    if( (direction % 2 ) == 1 )
                        utilityMap[ neighbor.getX() ][ neighbor.getY() ] = (short)(utilityMap[ node.getX()][node.getY()]+3);
                    else
                        utilityMap[ neighbor.getX() ][ neighbor.getY() ] = (short)(utilityMap[ node.getX()][node.getY()]+1);
		} 
	    } while( !node.checkedAll() );
	    node = search.pull();
	    
	    if( node == null )
	    {
		System.out.println( "No path found!" );
		return path;
	    }
	}
	path.addElement( node );
	while( ! ( (node.getX() == start.x) && (node.getY() == start.y) ) )
	{
	    point = node.getPoint();
	    node = new MapNode( findPrevious( point ), 1, true );
            //System.out.println( "Path: " + node.getPoint().x +","+node.getPoint().y );
	    path.insertElementAt( node, 0 );
	}
	return path;
    }

    public Vector findShortestPath( Character character, Point end, boolean canOpenDoors )
    {
	long startTime = System.currentTimeMillis();
        Point start = character.getLocation();
        Vector path = new Vector();
        resetUtilityMap();
        utilityMap[start.x][start.y]=(short)1;
        LinkedMapNodeList search = new LinkedMapNodeList();
        MapNode node;
        MapNode neighbor;
        Point point;
        Point npoint;
        int direction;
	Character [] visibleCharacters = game.getVisibleCharacters( character );

        node = new MapNode( start, 8, false );
        if( node.getX() == end.x && node.getY() == end.y )
        {
            path.add( start );
            return path;
        }
        while( ! ((node.getX() == end.x) && (node.getY() == end.y) ) )
        {
            do
            {
                direction = node.nextDirection();
                //System.out.println( "Checking: " + node.getX() + "," + node.getY() +" "+ direction );
                node.setChecked( direction );
                npoint = getNeighbor( node.getPoint(), direction );
                if( (game.canCharacterMove( character, npoint, visibleCharacters, false, canOpenDoors ) || npoint.equals( end ) ) && ( (utilityMap[ npoint.x][ npoint.y] == 0 )  || (utilityMap[ npoint.x ][ npoint.y ] >  utilityMap[ node.getX() ][ node.getY() ]+1 ) ) )
                {
                    //System.out.println( "Adding point to search" );
                    neighbor = new MapNode( new Point( npoint ), 8, false  );
                    //neighbor.setChecked( (direction+4)%8 );	// Serves no purpose yet, quite on the contrary, breaks the search
                    search.push( neighbor );
		    short oldVal = utilityMap[ neighbor.getX() ][ neighbor.getY() ];
		    short newVal = 0;
		    if( (direction % 2 ) == 1 )
		    {
			newVal = (short)(utilityMap[ node.getX()][node.getY()]+3);
		    }
		    else
		    {
			newVal = (short)(utilityMap[ node.getX()][node.getY()]+1);
		    }
		    if( oldVal != 0 )
		    {
			newVal = (short)Math.min( oldVal, newVal );
		    }
		    utilityMap[ neighbor.getX() ][ neighbor.getY() ] = newVal;
		    /*
                    if( (direction % 2 ) == 1 )
                        utilityMap[ neighbor.getX() ][ neighbor.getY() ] = (short)(utilityMap[ node.getX()][node.getY()]+3);
                    else
                        utilityMap[ neighbor.getX() ][ neighbor.getY() ] = (short)(utilityMap[ node.getX()][node.getY()]+1);
		     */
                }
            } while( !node.checkedAll() );
            node = search.pull();

            if( node == null )
            {
                System.out.println( "No path found!" );
                return path;
            }
        }
        path.addElement( node );
        while( ! ( (node.getX() == start.x) && (node.getY() == start.y) ) )
        {
            point = node.getPoint();
            node = new MapNode( findPrevious( point ), 1, true );
            //System.out.println( "Path: " + node.getPoint().x +","+node.getPoint().y );
            path.insertElementAt( node, 0 );
        }
	long endTime = System.currentTimeMillis();
	//System.out.println( "Path finding took " + (endTime-startTime) + " ms, " + path.size() + " nodes" );
        return path;
    }
    
    // Find the shortest path through the room graph, then search to the hotspot a few rooms down the road
    public Vector findShortestPath( Character character, Point end, int APCutOff, boolean canOpenDoors )
    {
	Room startRoom = getClosestRoomForPoint( character.getLocation() );
	Room endRoom = getClosestRoomForPoint( end );
	if( startRoom == null || endRoom == null || startRoom == endRoom || startRoom.hasLinked( endRoom ) || character.getLocation().distance( end ) <= APCutOff*1.5 )
	{
	    return findShortestPath( character, end, canOpenDoors );
	}
	Vector path = new Vector();
	long startTime = System.currentTimeMillis();
	Stack roomPath = findPathToRoom( startRoom, endRoom, new Vector() );
	//System.out.println( "Stack size: " + roomPath.size() );
	if( roomPath.size() > 1 )
	{
	    roomPath.pop(); // Pop the starting room off the path
	}
	Room nextRoom;
	do
	{
	    nextRoom = (Room)roomPath.pop();
	} while( ( roomPath.size() > 0 ) && ( character.getLocation().distance( nextRoom.getHotSpot() ) < APCutOff ) );	
	Point newTarget = nextRoom.getHotSpot();
	if( nextRoom == endRoom )
	{
	    // Turns out we were close enough to search the actual end spot.
	    // Basically, we did this all for nothing. Oh well...
	    newTarget = end;
	}
	path = findShortestPath( character, newTarget, canOpenDoors );
	long endTime = System.currentTimeMillis();
	//System.out.println( "Path to next room found in " + (endTime-startTime) + " ms, " + path.size() + " nodes" );
	return path;
	
    }
    
    public Stack findPathToRoom( Room startRoom, Room endRoom, Vector visited )
    {
	Stack stack = new Stack();
	if( startRoom.hasLinked( endRoom ) )
	{
	    //System.out.println( "Found target room" );
	    stack.push( endRoom );
	    return stack;
	}
	Stack newStack = null;
	String [] linkedRoomIds = startRoom.getLinkedRoomIds();
	//System.out.println( "Target room not found, checking linked from " + startRoom.getName() );
	for( int i=0; i < linkedRoomIds.length; i++ )
	{
	    Room room = getRoom( linkedRoomIds[i] );
	    if( room != null && !visited.contains( room ) )
	    {
		Vector newVisited = new Vector( visited );
		newVisited.add( startRoom );
		newStack = findPathToRoom( room, endRoom, newVisited );
		//System.out.println( "Substack size: " + newStack.size() );
		if( ((newStack.size() > 0 ) && ( newStack.size() < stack.size() )) || ( stack.size() == 0 ) )
		{
		    //System.out.println( "Shorter than previous, replacing" );
		    stack = newStack;
		}
	    }
	}
	if( stack.size() > 0 )
	{
	    stack.push( startRoom );
	}
	return stack;
    }
    
    private Point findPrevious( Point point )
    {
	int level = utilityMap[point.x][point.y];
        int smallest;
	int direction;
	Point testPoint;
	Point previous = null;
        smallest = level;
	for( int i = 7; i>=0; i-- )
	{
	    testPoint = getNeighbor( point, i );
            if( testPoint.x >= 0 && testPoint.x < horizontalSize && testPoint.y >= 0 && testPoint.y < verticalSize )
	    {
		if( (utilityMap[testPoint.x][testPoint.y] <= smallest) && (utilityMap[testPoint.x][testPoint.y] > 0) )
		{
		    smallest = utilityMap[testPoint.x][testPoint.y];
		    previous = testPoint;
		}
            }
	}
        // Extra check for CommSentry and doors
        if( previous == null )
        {
            System.out.println( "Previous node on path not found! Point: " + point.toString() + " at level " + level);
            for( int i=0; i<verticalSize; i++ )
            {
                for( int j=0; j<horizontalSize; j++ )
                {
                    System.out.print( utilityMap[j][i] +" " );
                    if( utilityMap[j][i] < 10 )
                        System.out.print( " " );
                }
                System.out.print("\n");
            }
        }
	return previous;
    }

    public void setCellFloor( int x, int y, int type )
    {
        if( cellMap[x][y] == null )
        {
            cellMap[x][y] = new MapCell();
        }
        cellMap[x][y].setFloor( type );
    }

    public void addCellFloor( int x, int y, int type )
    {
        if( cellMap[x][y] == null )
        {
            cellMap[x][y] = new MapCell();
            cellMap[x][y].setFloor( type );
        }
        else
        {
            cellMap[x][y].addFloor( type );
        }
    }

    public void setCellWall( int x, int y, int type )
    {
        if( cellMap[x][y] == null )
        {
            cellMap[x][y] = new MapCell();
        }
        cellMap[x][y].setWall( type );
    }
    
    public byte getLight( int x, int y )
    {
        if( cellMap[x][y] != null )
            return cellMap[x][y].getLight();
        return 0;
    }

    public byte getLight( Point point )
    {
        if( cellMap[ point.x ][ point.y ] != null )
            return cellMap[ point.x ][ point.y ].getLight();
        return 0;
    }

    public void setLight( int x, int y, byte light )
    {
        if( cellMap[x][y] != null )
            cellMap[x][y].setLight( light );
    }
    
    public int getUtility( int x, int y )
    {
	return utilityMap[x][y];
    }
    
    public void setUtility( int x, int y, int value )
    {
	utilityMap[x][y] = (short)value;
    }
    
    public void addObject( MapObject object )
    {
	if( !objects.contains( object ) )
	{
	    objects.add( object );
	    for( int x=object.getTopLeft().x; x <=object.getBottomRight().x; x++ )
	    {
		for( int y=object.getTopLeft().y; y <= object.getBottomRight().y; y++ )
		{
		    if( object.isLocatedIn( x, y ) )
		    {
			cellMap[x][y].setObjects( true );
		    }
		}
	    }
	}
    }
    
    public void removeObjects( int x, int y )
    {
        if( cellMap[x][y] != null )
        {
            if( cellMap[x][y].hasObjects() )
            {
                for( int i=0; i < objects.size(); i++ )
                {
                    if( ((MapObject)objects.elementAt(i) ).isLocatedIn( x, y ) )
                    {
                        objects.removeElementAt( i );
                        i--;
                    }
                }
                cellMap[x][y].setObjects( false );
            }
        }
    }

    public int getNumberOfObjects()
    {
        return objects.size();
    }

    public MapObject getObject( int index )
    {
        return (MapObject)objects.elementAt( index );
    }
    
    public MapObject getObject( String id )
    {
	MapObject object = null;
	boolean found = false;
	for( int i=0; i < objects.size() && !found; i++ )
	{
	    if( ((MapObject)objects.elementAt( i )).getId().equals( id ) )
	    {
		object = (MapObject)objects.elementAt( i );
		found = true;
	    }
	}
	
	return object;
    }
    
    public MapObject getObjectForPoint( Point point )
    {
	if( cellMap[ point.x ][ point.y ] == null )
	{
	    return null;   
	}
	if( cellMap[ point.x ][ point.y ].hasObjects() )
	{
	    for( int i=0; i < objects.size(); i++ )
	    {
		MapObject object = getObject( i );
		if( object.isLocatedIn( point.x, point.y ) && object.getStrength() > 0 )
		{
		    return object;   
		}
	    }
	}
	return null;
    }
    
    public void addRoom( Room room )
    {
	if( !rooms.containsKey( room.getId() ) )
	{
	    rooms.put( room.getId(), room );
	    //System.out.println( "Added room: " + room.getName() + " (" + room.getId() +")" );
	}
    }
    
    public Room getRoom( String id )
    {
	return (Room)rooms.get( id );
    }
    
    public Room getRoomForPoint( Point point )
    {
	Room [] theRooms = getRooms();
	for( int i=0; i < theRooms.length; i++ )
	{
	    if( point.x >= theRooms[i].getTopLeft().x && point.x <= theRooms[i].getBottomRight().x &&
		point.y >= theRooms[i].getTopLeft().y && point.y <= theRooms[i].getBottomRight().y )
	    {
		// Found the room
		return theRooms[i];
	    }
	}
	return null;
    }
    
    public Room getClosestRoomForPoint( Point point )
    {
	Room room = null;
	room = getRoomForPoint( point );
	if( room != null )
	{
	    return room;
	}
	int distance = Integer.MAX_VALUE;
	Room [] theRooms = getRooms();
	for( int i=0; i < theRooms.length; i++ )
	{
	    int roomDistance = (int)(theRooms[i].getHotSpot().distance( point ));
	    if( roomDistance < distance )
	    {
		room = theRooms[i];
		distance = roomDistance;
	    }
	}
	return room;
    }
    
    public Room [] getRooms()
    {
	Collection collection = rooms.values();
	Room [] theRooms = new Room[ collection.size() ];
	
	Iterator iterator = collection.iterator();
	int i=0;
	while( iterator.hasNext() )	
	{
	    theRooms[ i++ ] = (Room)(iterator.next());
	}
	return theRooms;
    }
    
    public void removeRoom( Room room )
    {
	if( rooms.containsKey( room.getId() ) )
	{
	    String [] linkedRoomIds = room.getLinkedRoomIds();
	    Room otherRoom;
	    for( int i=0; i < linkedRoomIds.length; i++ )
	    {
		String id = linkedRoomIds[ i ];
		otherRoom = (Room)rooms.get( id );
		if( otherRoom != null )
		{
		    otherRoom.removeLinkedRoom( room );
		}
	    }
	    rooms.remove( room.getId() );
	}
    }    
    
    public void addWallSet( String setName )
    {
	wallSetNames.add( new String( setName ) );
    }
    
    public Vector getWallSets()
    {
	return wallSetNames;
    }
    
    public void addFloorSet( String setName )
    {
	floorSetNames.add( new String( setName ) );
    }
    
    public Vector getFloorSets()
    {
	return floorSetNames;
    }
    
    public void addDoorSet( String setName )
    {
	doorSetNames.add( new String( setName ) );
    }
    
    public Vector getDoorSets()
    {
	return doorSetNames;
    }
    
    public void addObjectSet( String setName )
    {
	objectSetNames.add( new String( setName ) );
    }
    
    public Vector getObjectSets()
    {
	return objectSetNames;
    }
    
    public void setName( String name )
    {
	this.name = new String( name );
    }
    
    public String getName()
    {
	return new String( name );
    }
    
    public String getFileName()
    {
	return fileName;
    }
    
    public int getHorizontalSize()
    {
	return horizontalSize;
    }
    
    public int getVerticalSize()
    {
	return verticalSize;
    }

    public static Map load( File file ) throws Exception
    {
	Map newMap = null;
	if( file.exists() && file.canRead() )
	{
	    DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
	    DOMImplementation builder = documentBuilder.getDOMImplementation();
	    
	    FileInputStream fis = new FileInputStream( file );
	    GZIPInputStream gis = new GZIPInputStream( fis );
	    
	    //System.out.println( "Parsing map file: " + file.getName() );
	    Document document = documentBuilder.parse( gis );
	    
	    gis.close();
	    fis.close();
	    
	    Element root = document.getDocumentElement();
	    String mapName = root.getAttribute( "name" );
	    System.out.println( "Map name: " + mapName );
	    int mapWidth = Integer.parseInt( root.getAttribute( "width" ) );
	    int mapHeight = Integer.parseInt( root.getAttribute( "height" ) );
	    newMap = new Map( mapWidth, mapHeight );
	    newMap.setName( mapName );
	    newMap.fileName = file.getName();
	    Element setsElement = XMLUtils.getNamedChild( root, "graphicssets" );
	    Element [] floorSets = XMLUtils.getNamedChildren( setsElement, "floorset" );
	    //System.out.println( "Reading floor sets" );
	    for( int i=0; i < floorSets.length; i++ )
	    {
		if( i != Integer.parseInt( floorSets[i].getAttribute( "index" ) ) )
		{
		    System.out.println( "Index mismatch!" );
		}
		String setName = floorSets[i].getAttribute( "file" );
		if( setName != null )
		{
		    newMap.addFloorSet( setName );
		}
	    }
	    Element [] wallSets = XMLUtils.getNamedChildren( setsElement, "wallset" );
	    //System.out.println( "Reading wall sets" );
	    for( int i=0; i < wallSets.length; i++ )
	    {
		if( i != Integer.parseInt( wallSets[i].getAttribute( "index" ) ) )
		{
		    System.out.println( "Index mismatch!" );
		}
		String setName = wallSets[i].getAttribute( "file" );
		if( setName != null )
		{
		    newMap.addWallSet( setName );
		}
	    }
	    Element [] objectSets = XMLUtils.getNamedChildren( setsElement, "objectset" );
	    //System.out.println( "Reading object sets" );
	    for( int i=0; i < objectSets.length; i++ )
	    {
		if( i != Integer.parseInt( objectSets[i].getAttribute( "index" ) ) )
		{
		    System.out.println( "Index mismatch!" );
		}
		String setName = objectSets[i].getAttribute( "file" );
		if( setName != null )
		{
		    newMap.addObjectSet( setName );
		}
	    }
	    Element [] doorSets = XMLUtils.getNamedChildren( setsElement, "doorset" );
	    //System.out.println( "Reading door sets" );
	    for( int i=0; i < doorSets.length; i++ )
	    {
		if( i != Integer.parseInt( doorSets[i].getAttribute( "index" ) ) )
		{
		    System.out.println( "Index mismatch!" );
		}
		String setName = doorSets[i].getAttribute( "file" );
		if( setName != null )
		{
		    newMap.addDoorSet( setName );
		}
	    }
	    // Load cells
	    Element cellsElement = XMLUtils.getNamedChild( root, "cells" );
	    Element [] cellElements = XMLUtils.getNamedChildren( cellsElement, "cell" );
	    //System.out.println( cellElements.length + " cells found" );
	    for( int c=0; c < cellElements.length; c++ )
	    {
		MapCell cell = new MapCell();
		int x = Integer.parseInt( cellElements[c].getAttribute( "x" ) );
		int y = Integer.parseInt( cellElements[c].getAttribute( "y" ) );
		byte light = Byte.parseByte( cellElements[c].getAttribute( "light" ) );
		cell.setLight( light );
		Element [] floorElements = XMLUtils.getNamedChildren( cellElements[c], "floor" );
		if( floorElements != null )
		{
		    //System.out.println( floorElements.length + " floors found for cell: " + x + ", " + y );
		    for( int f=0; f < floorElements.length; f++ )
		    {
			int index = Integer.parseInt( floorElements[f].getAttribute( "index" ) );
			if( index != f )
			{
			    System.out.println( "Index mismatch!" );
			}
			cell.addFloor( Integer.parseInt( floorElements[f].getAttribute( "tile" ) ) );
		    }
		}
		Element wallElement = XMLUtils.getNamedChild( cellElements[c], "wall" );
		if( wallElement != null )
		{
		    cell.setWall( Integer.parseInt( wallElement.getAttribute( "type" ) ) );
		}
		newMap.setCell( x, y, cell );
	    }
	    // Load doors
	    Element doorsElement = XMLUtils.getNamedChild( root, "doors" );
	    if( doorsElement != null )
	    {
		Element [] doorElements = XMLUtils.getNamedChildren( doorsElement, "door" );
		//System.out.println( doorElements.length + " doors found" );
		for( int d=0; d < doorElements.length; d++ )
		{
		    int x = Integer.parseInt( doorElements[d].getAttribute( "x" ) );
		    int y = Integer.parseInt( doorElements[d].getAttribute( "y" ) );
		    int direction = Integer.parseInt( doorElements[d].getAttribute( "direction" ) );
		    boolean locked = doorElements[d].getAttribute( "locked" ).trim().equalsIgnoreCase( "true" );
		    boolean open = doorElements[d].getAttribute( "open" ).trim().equalsIgnoreCase( "true" );
		    Door door = new Door( x, y, direction );
		    door.setLocked( locked );
		    door.setOpen( open );
		    newMap.addDoor( door );
		}
	    }
	    // Load objects
	    Element objectsElement = XMLUtils.getNamedChild( root, "objects" );
	    if( objectsElement != null )
	    {
		Element [] objectElements = XMLUtils.getNamedChildren( objectsElement, "object" );
		//System.out.println( objectElements.length + " objects found" );
		for( int o=0; o < objectElements.length; o++ )
		{
		    String objectId = objectElements[o].getAttribute( "id" );
		    String name = objectElements[o].getAttribute( "name" );
		    int strength = Integer.parseInt( objectElements[o].getAttribute( "strength" ) );
		    int moveCost = Integer.parseInt( objectElements[o].getAttribute( "movecost" ) );
		    MapObject mapObject = new MapObject( objectId, name, strength, moveCost );
		    /// Load object cells
		    Element [] objectCellElements = XMLUtils.getNamedChildren( objectElements[o], "objectcell" );
		    if( objectCellElements != null )
		    {
			for( int c=0; c < objectCellElements.length; c++ )
			{
			    int x = Integer.parseInt( objectCellElements[c].getAttribute( "x" ) );
			    int y = Integer.parseInt( objectCellElements[c].getAttribute( "y" ) );
			    int type = Integer.parseInt( objectCellElements[c].getAttribute( "type" ) );
			    int destroyedType = Integer.parseInt( objectCellElements[c].getAttribute( "destroyedtype" ) );
			    int hitPoints = Integer.parseInt( objectCellElements[c].getAttribute( "hitpoints" ) );
			    boolean blocksMovement = objectCellElements[c].getAttribute( "blocksmovement" ).trim().equalsIgnoreCase( "true" );
			    boolean blocksSight = objectCellElements[c].getAttribute( "blocksvision" ).trim().equalsIgnoreCase( "true" );
			    MapObjectCell objectCell = new MapObjectCell( x, y, type, hitPoints, blocksMovement, blocksSight );
			    objectCell.setDestroyedType( destroyedType );
			    mapObject.addObjectCell( objectCell );
			}
		    }
		    mapObject.setStrength( strength );
		    //System.out.println( "Loaded object " + name + " (" + objectId + "), " + objectCellElements.length + " components" );
		    newMap.addObject( mapObject );
		}
	    }
	    // Load rooms
	    Element roomsElement = XMLUtils.getNamedChild( root, "rooms" );
	    if( roomsElement != null )
	    {
		Element [] roomElements = XMLUtils.getNamedChildren( roomsElement, "room" );
		for( int r=0; r < roomElements.length; r++ )
		{
		    String roomId = roomElements[r].getAttribute( "id" );
		    String roomName = roomElements[r].getAttribute( "name" );
		    int minX = Integer.parseInt( roomElements[r].getAttribute( "minx" ) );
		    int maxX = Integer.parseInt( roomElements[r].getAttribute( "maxx" ) );
		    int minY = Integer.parseInt( roomElements[r].getAttribute( "miny" ) );
		    int maxY = Integer.parseInt( roomElements[r].getAttribute( "maxy" ) );
		    int hotSpotX = Integer.parseInt( roomElements[r].getAttribute( "hotspotx" ) );
		    int hotSpotY = Integer.parseInt( roomElements[r].getAttribute( "hotspoty" ) );
		    Room room = new Room( roomId, roomName, minX, minY, maxX, maxY );
		    room.setHotSpot( new Point( hotSpotX, hotSpotY ) );
		    Element [] linkElements = XMLUtils.getNamedChildren( roomElements[r], "link" );
		    if( linkElements != null )
		    {
			for( int l=0; l < linkElements.length; l++ )
			{
			    room.addLink( linkElements[l].getAttribute( "target" ) );
			}
		    }		    
		    newMap.addRoom( room );
		}
	    }
	    //System.out.println( "Loaded map" );
	}
	return newMap;
    }
}
